/*******************************************************************************
********************************************************************************
****                                                                        ****
****            Copyright (C) Ciprico Incorporated 1986, 1988               ****
****                                                                        ****
****                         Ciprico Incorporated                           ****
****                           2955 Xenium Lane                             ****
****			     Plymouth, MN  55441			    ****
****                            (612) 559-2034                              ****
****                                                                        ****
****	Module Name:	rfutil.c                                            ****
****	Package:	Rimfire 3200/3400 Sun OS driver for Sun 3/160       ****
****	Date:		$Date:   13 Feb 1989 13:47:32  $		    ****
****	Module Rev:	$Revision:   1.16  $				    ****
****	                                                                    ****
****	Subroutines:    main, help, opendevice, readlabel, writelabel,      ****
****	                doformat, debug, dolabel, showlabel, domap,         ****
****	                dogeometry, showgeometry, dopartitions, newconf,    ****
****	                askpartitions, setpartitions, getpartitions,        ****
****	                showstatistics, doslip, readtid, doopen, readdef    ****
****	                                                                    ****
****	Description:                                                        ****
****		This program is a utility to work hand in hand with the     ****
****		Rimfire 3200 SMD disk controller driver and the Rimfire	    ****
****		3400 ESDI disk controller.  It is intended for use with     ****
****		the Sun VME workstations.  It supports labelling,	    ****
****		formatting, track and sector mapping.  It allows access to  ****
****		some of the data structures used in the Rimfire 3200/3400   ****
****		device driver for configuring and partitioning purposes.    ****
****	                                                                    ****
********************************************************************************
*******************************************************************************/

/*******************************************************************************
*									       *
*				Revision History			       *
*									       *
********************************************************************************

    Revision    Date		Author
	Description of Change
--------------------------------------------------------------------------------

    1.1		04/16/86	M. J. Tobias
	Initial Release

    1.2		07/16/86	R. W. Lee
	Update include sector slipping and board statistics

    1.3		11/12/86	D. A. Dickey
	Incorporate Rimfire 3400 additions
    1.3X	05/20/87	D. A. Dickey
	Adapt ioctl error return for DKIOMAP to conform to
	driver changes.

    1.4		05/21/87	J. E. Lind
	Improve user-friendliness through automatic features and
	improved screen appearance, add copyright notice and titling.
	Read defect list from drive and write to file.

    1.5		09/21/87	J. E. Lind	Annotated
	a) Added positive feedback to domap and doslip.
	b) Made "register" attribute conditional due to compiler bugs in
	   some systems.

    1.6		11/11/87	J. E. Lind	Annotated
	Added support for ESDI read defect list.

    1.9		87/12/15 - 88/1/3	J. E. Lind	Annotated
	a) Fixed automatic mapping of bad tracks (was a no-op).
	b) Replaced the 'y' command with the 'a' command.
	c) Added the 'e' command
	d) Added error returns to the format and read defect ioctl calls.
	e) Made error checking on "read label" more rigorous.
	f) In doconf, changed the note to say "changeable in label"
	   rather than "changeable in geometry" since the geometry command
	   does not permit editing.
	g) Updated many SMD "canned labels"

    1.10	88/1/28-88/3/14		J. E. Lind	Annotated
	a) Print out error code as well as error messages for Rimfire errors
	   in all cases.
	b) Display the name of the file which cannot be opened in setpartitions
	   to make it more abundantly clear what is happening.
	c) Replaced the reminder notes in writelabel() with checks for valid
	   geometry, label, and partitions.
	d) Changed verification automatic mapping to explicitly try to slip
	   the sector first, and if that is not possible, to map the whole
	   track for efficiency purposes.  Also added ioctl call to actually
	   perform the map track function, which was missing from 1.9 (!)
	e) Changed the error messages which erroneously perported to come from
	   DKIOCMAP to DKIOCFMT.

    1.11	88/4/27         J. D. Flatt
       No change

    1.12	88/5/27 - 88/7/5	J. E. Lind
       a) Made the creation of the defect file non-destructive.
       b) Ask for the track at which to begin reading the defect list.
	  This is only done for SMDs, since ESDIs don't have the defect
	  list scattered all over the disk surface.
       c) Break up verify and format commands into tracks so that we
	  don't kill the performance for other drives on the same controller.
       d) Add dummy values for new SunOS4 label definitions until such time
	  as we support them actively.
       e) Added and corrected default label entries.

    1.13	88/7/27		J. E. Lind
       a) Added missing leading quote to Rodime entry is SunOS4 labels.

    1.14	88/8/8		J. E. Lind
       a) Moved the dummy values for the new SunOS4 label to the correct
	  position within the structure.

    $Revision:   1.16  $	$Date:   13 Feb 1989 13:47:32  $	$Author:   JLind  $
	a)  Allowed the default drive to come from the command line.

*******************************************************************************/

#ifndef lint
static char	Sccsid[] = "@(#)rfutil.c  75182013 $Revision:   1.16  $ $Date:   13 Feb 1989 13:47:32  $ Copyright 1986, 1989 Ciprico, Inc.";
#endif

/* 1.5b */
#define CREGBUG	(1)

#ifdef CREGBUG
# define register
#endif

#define VERSION		"Version 1.16"
#define COPYRIGHT	"Copyright 1986, 1989 Ciprico, Inc."
#define TITLE		"R F U T I L"

#include <stdio.h>
#include <ctype.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/file.h>
#include <sys/stat.h>
#include "/sys/sun/dkio.h"
#include "/sys/sun/dklabel.h"
#define	RFSTATS	1
#define MAXSEC 300
#include <sys/time.h>
#include <sys/errno.h>
#include "rfreg.h"
#include "rfioctl.h"
#include "rferr.h"

#include <curses.h>

#define	DEFDEVICE	"/dev/rrf0a"	/* The device to open at startup */
#define LCBIT		(040)		/* bit to set for lower case	*/

#define	RF3200		(0)		/* controller type codes */
#define	RF3400		(1)

static int	disk = -1;	/* disk device file descriptor (-1 = none) */
static int	ctype = -1;	/* controller type (-1=NONE, 0=3200, 1=3400 */
static char	diskname[100];
static int	disksize = 0;
static char	line[256];
static daddr_t	badblocks[2048];
static int	badblki = 0;
static int	maxbadblks = sizeof (badblocks) / sizeof (daddr_t);
static WINDOW	*verwin;
static struct	dk_label label;	/* current label			*/
static int	lbcur = 0;	/* TRUE indicates that label matches disk */
static int	glabel = 0;		/* 0 indicates invalid label	*/

char	inp_char();
long lseek();

extern char	*sys_errlist[];
extern int	errno;

/* this is the list of error numbers returned by opendevice.  The numbers
   are biased and negative, such that

   openerrlist[0] = error # -1
   openerrlist[1] = error # -2
   ...
   openerrlist[n] = error # -(n+1)
   */
char *openerrlist[] = {
	"Cannot open file",
	"File not character device",
	"Cannot read partitions/label",
	"Cannot stat",
	"Wrong controller or driver too old"
	};

/* this is the name displayed for the type of controller associated with the
 * currently open disk.
 */
static char *cname[] = {
	"no controller",
	"SMD",
	"ESDI",
	"SCSI"
	};

char *rindex(), *index();

#define CHR_CR	('\r')
#define CHR_NL	('\n')
#define CHR_ESC	('\033')
#define CHR_FF	('\014')
#define CHR_RPT ('\022')

/* This is the number of bits used in the "configure disk parameter block"
 * flag byte, and the name associated with each.
 */

#define NCONFF	(2)

static char *cnffname[NCONFF] = {
    "SSP",
    "EAD"
    };


/*
 * This structure contains the labels used by SUN 4.2 to define the
 * configuration and partitions of a particular disk drive.
 * To add a label for a new disk or disk type, duplicate an existing
 * entry and change the appropriate fields of the structure.
 */

#ifdef SunOS3
static struct dk_label esdilabels[] = {
	/*    apc     gap1    gap2    intrlv   ncyl   acyl    nhead   nsect   bhead   ppart */
	{ "Edit Current Label", "",
		0,	0,	0,	0,	0,	0,	0,	0,	0,	0,
		{ 0, 0 },	{ 0, 0 },	{ 0, 0 },	{ 0, 0 },
		{ 0, 0 },	{ 0, 0 },	{ 0, 0 },	{ 0, 0 },
		DKL_MAGIC, 0
	},
	{ "CDC-Wren II cyl 900 alt 17 hd 9 sec 16 apc 9", "",
		9,	0,	0,	1,	900,	17,	9,	16,	0,	0,
		{ 0, 16128 },	{ 112, 32256 },	{ 0, 129600 },	{ 0, 0 },
		{ 0, 0 },	{ 0, 0 },	{ 336, 81216 },	{ 0, 0 },
		DKL_MAGIC, 0
	},
	{ "CDC-Wren III cyl 960 alt 9 hd 9 sec 33 apc 9", "",
		9,      0,      0,      1,      960,   9 ,     9,      33,   
0,	0,
		{ 0, 16929 },   { 57, 33858 },  { 0, 285120 },  { 0, 0 },
		{ 0, 0 },       { 0, 0 },       { 171, 234333 },       { 0, 0 },
		DKL_MAGIC, 0
	},
	{ "Fujitsu 2246E cyl 818 alt 5 hd 10 sec 31", "",
		10,	0,	0,	1,	818,	5,	10,	31,	0,	0,
		{ 0, 15500 },	{ 50, 31000 },	{ 0, 253580 },	{ 0, 0 },
		{ 0, 0 },	{ 0, 0 },	{150, 207080 },	{ 0, 0 },
		DKL_MAGIC, 0
	},
	{ "Fujitsu 2249E cyl 1238 alt 5 hd 15 sec 34 apc 15", "",
		15,	0,	0,	1,     1238,	5,	15,	34,	0,	0,
		{ 0, 20400 },	{ 40, 40800 },	{ 0, 631380 },	{ 0, 0 },
		{ 0, 0 },	{ 0, 0 },	{120, 570180 },	{ 0, 0 },
		DKL_MAGIC, 0
	},
	{ "Maxtor EXT-4380 cyl 1200 alt 21 hd 15 sec 33 apc 15", "",
		15,	0,	0,	1,	1200,	21,    15,	33,	0,	0,
		{ 0, 16335 },	{33, 32670 },	{ 0, 594000 },	{ 0, 0 },
		{ 0, 0 },	{ 0, 0 },	{ 99, 544995 }, 	{ 0, 0 },
		DKL_MAGIC, 0
	},
	{ "Maxtor XT8760E cyl 1622 alt 10 hd 15 sec 50","",
		15,	0,	0,	1,	1622,	10,	15,	50,	0,	0,
		{ 0, 18750 },	{ 25, 37500 },	{ 0, 1216500 },	{ 0, 0 },
		{ 0, 0 },	{ 0, 0 }, 	{ 75, 1160250 },{ 0, 0}
	},
	{ "MICROPOLIS 1355 cyl 1000 alt 22 hd 8 sec 34 apc 8", "",
		8,	0,	0,	1,	1000,	22,	8,	34,	0,	0,
		{ 0, 16592 },	{ 61, 33184 },	{ 0, 272000 },	{ 0, 0 },
		{ 0, 0 },	{ 0, 0 },	{ 183, 222224 },	{ 0, 0 },
		DKL_MAGIC, 0
	},  
};

/* 1.9g */
static struct dk_label smdlabels[] = {
	/*    apc     gap1    gap2    intrlv   ncyl   acyl    nhead   nsect   bhead   ppart */
	{ "Edit Current Label", "",
		0,	0,	0,	0,	0,	0,	0,	0,	0,	0,
		{ 0, 0 },	{ 0, 0 },	{ 0, 0 },	{ 0, 0 },
		{ 0, 0 },	{ 0, 0 },	{ 0, 0 },	{ 0, 0 },
		DKL_MAGIC, 0
	},
	{ "Century Data-AMS571 cyl 820 alt 3 hd 19 sec 32", "",
		0,	0,	0,	1,	820,	3,	19,	32,	0,	0,
		{ 0, 16416 },	{ 27, 33440 },	{ 0, 498560 },	{ 0, 0 },
		{ 0, 0 },	{ 0, 0 },	{ 82, 448704 },	{ 0, 0 },
		DKL_MAGIC, 0
	},
	{ "Century Data-C2800 cyl 843 alt 2 hd 24 sec 67", "",
		24,	16,	16,	1,	843,	2,	24,	67,	0,	0,
		{ 0, 17688 },	{ 11, 35376 },	{ 0, 1355544 },	{ 0, 0 },
		{ 0, 0 },	{ 0, 0 },	{ 33, 1302480 },{ 0, 0 },
		DKL_MAGIC, 0
	},
        { "CDC-FSD515 cyl 707 alt 4 hd 24 sec 49 apc 24","",
	      24,	21,	15,	1,	707,	4,	24,	49,	0, 0,
		{ 0, 16464 },	{ 14, 32928 },	{ 0, 831432 },	{ 0, 0 },
		{ 0, 0 },	{ 0 , 0 },	{ 42, 782040 },	{ 0, 0 },
		DKL_MAGIC, 0
	},
	{ "CDC-9720-368 cyl 1215 alt 2 hd 10 sec 50 apc 10", "",
	       10,	15,	15,	1,     1215,   2,	10,	50,	0, 0,
		{ 0, 16500 },   { 33, 33000 },  { 0, 607500 }, { 0, 0 },
		{ 0, 0 }, 	{ 0, 0 }, 	{ 99, 558000 }, { 0, 0},
		DKL_MAGIC, 0
	},
	{ "CDC-9720-500 cyl 1215 alt 2 hd 10 sec 68 apc 10", "",
	       10,	15,	15,	1,     1215,   2,	10,	68,	0, 0,
		{ 0, 17000 },   { 25, 34000 },  { 0, 826200 }, { 0, 0 },
		{ 0, 0 }, 	{ 0, 0 }, 	{ 75, 775200 }, { 0, 0},
		DKL_MAGIC, 0
	},
	{ "CDC-9720-736 cyl 1632 alt 3 hd 15 sec 50 apc 15", "",
	       15,	15,	15,	1,     1632,   3,	15,	50,	0, 0,
		{ 0, 16500 },   { 22, 33000 },  { 0, 1224000 }, { 0, 0 },
		{ 0, 0 }, 	{ 0, 0 }, 	{ 66, 1174500 }, { 0, 0},
		DKL_MAGIC, 0
	},
	{ "CDC-9720-850 cyl 1378 alt 3 hd 15 sec 68 apc 15", "",
	       15,	15,	15,	1,     1378,   3,	15,	68,	0, 0,
		{ 0, 17340 },   { 17, 34680 },  { 0, 1405560 }, { 0, 0 },
		{ 0, 0 }, 	{ 0, 0 }, 	{ 51, 1353540 }, { 0, 0},
		DKL_MAGIC, 0
	},
	{ "CDC-9720-1230 Sabre-V cyl 1633 alt 2 hd 15 sec 83 apc 15", "",
	       15,	15,	15,	1,     1633,   2,	15,	83,	0, 0,
		{ 0, 17430 },   { 14, 34860 },  { 0, 2033085 }, { 0, 0 },
		{ 0, 0 }, 	{ 0, 0 }, 	{ 42, 1980795 }, { 0, 0},
		DKL_MAGIC, 0
	},
	{ "CDC-9766 cyl 820 alt 3 hd 19 sec 32 apc 19", "",
	       19,	21,	15,	1,	820,	3,	19,	32,	0,	0,
		{ 0, 16416 },	{ 27, 33440 },	{ 0, 498560 },	{ 0, 0 },
		{ 0, 0 },	{ 0, 0 },	{ 82, 448704},	{ 0, 0 },
		DKL_MAGIC, 0
	},
	{ "CDC-9772-850 XMD-II cyl 1062 alt 2 hd 16 sec 82 apc 16", "",
	       16,	19,	15,	1,     1062,   2,	16,	82,	0, 0,
		{ 0, 17056 },   { 13, 34112 },  { 0, 1393344 }, { 0, 0 },
		{ 0, 0 }, 	{ 0, 0 }, 	{ 39, 1342166 }, { 0, 0},
		DKL_MAGIC, 0
	},
	{ "CDC-9773-1350 cyl 1418 alt 2 hd 19 sec 84 apc 19", "",
	       19,	15,	15,	1,     1418,   2,	19,	84,	0, 0,
		{ 0, 17556 },   { 11, 33516 },  { 0, 2263128 }, { 0, 0 },
		{ 0, 0 }, 	{ 0, 0 }, 	{ 32, 2212056 }, { 0, 0},
		DKL_MAGIC, 0
	},
	{ "Fujitsu-M2312K cyl 586 alt 3 hd 7 sec 33 apc 7", "",
	        7,	15,	15,	1,	586,	3,	7,	33,	0,	0,
		{ 0, 16401 },	{ 71, 32802 },	{ 0, 135366 },	{ 0, 0 },
		{ 0, 0 },	{ 0, 0 },	{ 213, 86163 },	{ 0, 0 },
		DKL_MAGIC, 0
	},
	{ "Fujitsu-M2322 cyl 821 alt 2 hd 10 sec 33 apc 10", "",
	        10,	15,	17,	1,	821,	2,	10,	33,	0,	0,
		{ 0, 16500 },	{ 50, 33000 },	{ 0, 270930 },	{ 0, 0 },
		{ 0, 0 },	{ 0, 0 },	{ 150, 221430 },{ 0, 0 },
		DKL_MAGIC, 0
	},
	{ "Fujitsu-2333 Swallow cyl 821 alt 2 hd 10 sec 67 apc 10", "",
	      10,	16,	19,	1,	821,	2,	10,	67,	0, 0,
		{ 0, 17420 },   { 26, 32830 },  { 0, 550070 },  { 0, 0 },
		{ 0, 0 }, 	{ 0, 0 }, 	{ 75, 499820 }, { 0, 0 },
	        DKL_MAGIC, 0
	},
	{ "Fujitsu-M2344 Swallow II cyl 622 alt 2 hd 27 sec 67 apc 27", "",
	       27,     16,     19,	1,	622,	2,	27,	67,	0,	0,
		{ 0, 21708 },	{ 12, 34371 },	{ 0, 1125198},	{ 0, 0 },
		{ 0, 0 },	{ 0, 0 },	{ 31, 1069119},	{ 0, 0 },
		DKL_MAGIC, 0
	},
	{ "Fujitsu-M2351 Eagle cyl 840 alt 2 hd 20 sec 45 apc 20", "",
	       20,	18,	20,	1,	840,	2,	20,	45,	0,	0,
		{ 0, 17100 },	{ 19, 34200 },	{ 0, 756000 },	{ 0, 0 },
		{ 0, 0 },	{ 0, 0 },	{ 57, 704700 },	{ 0, 0 },
		DKL_MAGIC, 0
	},
	{ "Fujitsu-M2361A Eagle XP cyl 840 alt 2 hd 20 sec 66 apc 20", "",
               20,	18,	20,	1,	840,	2,	20,	66,	0, 0,
		{ 0, 17160 },   { 13, 34320 },  { 0, 1108800 },  { 0, 0 },
		{ 0, 0 },       { 0, 0 },       { 39, 1057320 }, { 0, 0 },
		DKL_MAGIC, 0
	},
	{ "Fujitsu-M2372 Swallow III cyl 743 alt 2 hd 27 sec 67 apc 27", "",
	       27,     20,     20,	1,	743,	2,	27,	67,	0,	0,
		{ 0, 21708 },	{ 12, 34371 },	{ 0, 1344087},	{ 0, 0 },
		{ 0, 0 },	{ 0, 0 },	{ 31, 1288008},	{ 0, 0 },
		DKL_MAGIC, 0
	},
	{ "Fujitsu-M2382 Swallow IV cyl 743 alt 2 hd 27 sec 83 apc 27 /SSP", "",
	       27,     19,     20,	1,	743,	2,	27,	83,	0,	0,
		{ 0, 22410 },	{ 10, 33615 },	{ 0, 1665063},	{ 0, 0 },
		{ 0, 0 },	{ 0, 0 },	{ 25, 1609038},	{ 0, 0 },
		DKL_MAGIC, 0
	},
	{ "Hitachi DK815-10 cyl 1735 alt 2 hd 15 sec 67 apc 15 /SSP /EAD", "",
	       15,	19,	19,	1,     1735,   2,	15,	67,	0, 0,
		{ 0, 17085 },   { 17, 34170 },  { 0, 1743675 }, { 0, 0 },
		{ 0, 0 }, 	{ 0, 0 }, 	{ 51, 1692420 }, { 0, 0},
		DKL_MAGIC, 0
	},
	{ "NEC-D2352 cyl 758 alt 2 hd 19 sec 60 apc 19", "",
	       19,     19,     14,	1,	758,	2,     19,	60,	0,	0,
		{ 0, 17100 },	{ 15, 34200 },	{ 0, 864120 },	{ 0, 0 },
		{ 0, 0 },	{ 0, 0 },	{ 45, 812820},	{ 0, 0 },
		DKL_MAGIC, 0
	},
	{ "NEC-D2362 cyl 848 alt 2 hd 23 sec 66 apc 23", "",
	       23,     19,     12,	1,	848,	2,     23,	66,	0,	0,
		{ 0, 16698 },	{ 11, 33396 },	{ 0, 1287264 },	{ 0, 0 },
		{ 0, 0 },	{ 0, 0 },	{ 33, 1237170},	{ 0, 0 },
		DKL_MAGIC, 0
	},
	{ "NEC-D2363 cyl 1022 alt 2 hd 19 sec 67 apc 19", "",
	       19,     16,     19,	1,	1022,	2,     19,	67,	0,	0,
		{ 0, 16549 },	{ 13, 33098 },	{ 0, 1301006},	{ 0, 0 },
		{ 0, 0 },	{ 0, 0 },	{ 39, 1251359},	{ 0, 0 },
		DKL_MAGIC, 0
	},
	{ "NEC-D2268H cyl 821 alt 2 hd 10 sec 67 apc 10", "",
	       10,	19,	14,	1,	821,	2,	10,	67,	0,	0,
		{ 0, 16750 },	{ 25, 33500 },{ 0, 550070 },	{ 0, 0 },
		{ 0, 0 },	{ 0, 0 },	{ 75, 499820 },{ 0, 0 },
		DKL_MAGIC, 0
	},
	{ "Rodime RO8074 cyl 1644 alt 2 hd 11 sec 63 apc 11 /SSP /EAD", "",
	       11,	19,	19,	1,     1644,   2,	11,	63,	0, 0,
		{ 0, 16632 },   { 24, 33264 },  { 0, 1139292 }, { 0, 0 },
		{ 0, 0 }, 	{ 0, 0 }, 	{ 72, 1089396 }, { 0, 0},
		DKL_MAGIC, 0
	},
	{ "Toshiba MK-288FC cyl 820 alt 3 hd 15 sec 67 apc 15", "",
		15,	19,	19,	1,	820,	3,	15,	67,	0,	0,
		{ 0, 17085 },	{ 17, 34170 },	{ 0, 824100 },	{ 0, 0 },
		{ 0, 0 },	{ 0, 0 },	{ 51, 772845 },	{ 0, 0 },
		DKL_MAGIC, 0
	},
	{ "Toshiba MK-388FA cyl 1160 alt 2 hd 15 sec 68 apc 15 /SSP", "",
		15,	18,	18,	1,	1160,	2,	15,	68,	0,	0,
		{ 0, 21420 },	{ 21, 33660 },	{ 0, 1183200 },	{ 0, 0 },
		{ 0, 0 },	{ 0, 0 },	{ 54, 1128120}, { 0, 0 },
		DKL_MAGIC, 0
	},
};
#endif /* SunOS3 */

#ifdef SunOS4 /* 1.12d */
static struct dk_label esdilabels[] = {
	/*    apc  gap1  gap2  intrlv   ncyl   acyl  nhead  nsect  bhead  ppart */
	{ "Edit Current Label", "", 0,0,
		0,	0,	0,	0,	0,	0,	0,	0,	0,	0,
		{ 0, 0 },	{ 0, 0 },	{ 0, 0 },	{ 0, 0 },
		{ 0, 0 },	{ 0, 0 },	{ 0, 0 },	{ 0, 0 },
		DKL_MAGIC, 0
	},
	{ "CDC-Wren II cyl 900 alt 17 hd 9 sec 16 apc 9", "", 0,0,
		9,	0,	0,	1,	900,	17,	9,	16,	0,	0,
		{ 0, 16128 },	{ 112, 32256 },	{ 0, 129600 },	{ 0, 0 },
		{ 0, 0 },	{ 0, 0 },	{ 336, 81216 },	{ 0, 0 },
		DKL_MAGIC, 0
	},
	{ "CDC-Wren III cyl 960 alt 9 hd 9 sec 33 apc 9", "", 0,0,
		9,      0,      0,      1,      960,   9 ,     9,      33,   
0,	0,
		{ 0, 16929 },   { 57, 33858 },  { 0, 285120 },  { 0, 0 },
		{ 0, 0 },       { 0, 0 },       { 171, 234333 },       { 0, 0 },
		DKL_MAGIC, 0
	},
	{ "Fujitsu 2246E cyl 818 alt 5 hd 10 sec 31", "", 0,0,
		10,	0,	0,	1,	818,	5,	10,	31,	0,	0,
		{ 0, 15500 },	{ 50, 31000 },	{ 0, 253580 },	{ 0, 0 },
		{ 0, 0 },	{ 0, 0 },	{150, 207080 },	{ 0, 0 },
		DKL_MAGIC, 0
	},
	{ "Fujitsu 2249E cyl 1238 alt 5 hd 15 sec 34", "", 0,0,
		15,	0,	0,	1,	1238,	5,	15,	34,	0,	0,
		{ 0, 20400 },	{ 40, 40800 },	{ 0, 631380 },	{ 0, 0 },
		{ 0, 0 },	{ 0, 0 },	{120, 570180 },	{ 0, 0 },
		DKL_MAGIC, 0
	},
	{ "Maxtor EXT-4380 cyl 1200 alt 21 hd 15 sec 33 apc 15", "", 0,0,
		15,	0,	0,	1,	1200,	21,    15,	33,	0,	0,
		{ 0, 16335 },	{33, 32670 },	{ 0, 594000 },	{ 0, 0 },
		{ 0, 0 },	{ 0, 0 },	{ 99, 544995 }, 	{ 0, 0 },
		DKL_MAGIC, 0
	},
	{ "Maxtor XT8760E cyl 1622 alt 10 hd 15 sec 50", "", 0,0,
		15,	0,	0,	1,	1622,	10,	15,	50,	0,	0,
		{ 0, 18750 },	{ 25, 37500 },	{ 0, 1216500 },	{ 0, 0 },
		{ 0, 0 },	{ 0, 0 }, 	{ 75, 1160250 },{ 0, 0}
	},
	{ "MICROPOLIS 1355 cyl 1000 alt 22 hd 8 sec 34 apc 8", "", 0,0,
		8,	0,	0,	1,	1000,	22,	8,	34,	0,	0,
		{ 0, 16592 },	{ 61, 33184 },	{ 0, 272000 },	{ 0, 0 },
		{ 0, 0 },	{ 0, 0 },	{ 183, 222224 },	{ 0, 0 },
		DKL_MAGIC, 0
	},  
};

/* 1.9g */
static struct dk_label smdlabels[] = {
	/*    apc     gap1    gap2    intrlv   ncyl   acyl    nhead   nsect   bhead   ppart */
	{ "Edit Current Label", "", 0,0,
		0,	0,	0,	0,	0,	0,	0,	0,	0,	0,
		{ 0, 0 },	{ 0, 0 },	{ 0, 0 },	{ 0, 0 },
		{ 0, 0 },	{ 0, 0 },	{ 0, 0 },	{ 0, 0 },
		DKL_MAGIC, 0
	},
	{ "Century Data-AMS571 cyl 820 alt 3 hd 19 sec 32", "", 0,0,
		0,	0,	0,	1,	820,	3,	19,	32,	0,	0,
		{ 0, 16416 },	{ 27, 33440 },	{ 0, 498560 },	{ 0, 0 },
		{ 0, 0 },	{ 0, 0 },	{ 82, 448704 },	{ 0, 0 },
		DKL_MAGIC, 0
	},
	{ "Century Data-C2800 cyl 843 alt 2 hd 24 sec 67", "", 0,0,
		24,	16,	16,	1,	843,	2,	24,	67,	0,	0,
		{ 0, 17688 },	{ 11, 35376 },	{ 0, 1355544 },	{ 0, 0 },
		{ 0, 0 },	{ 0, 0 },	{ 33, 1302480 },{ 0, 0 },
		DKL_MAGIC, 0
	},
        { "CDC-FSD515 cyl 707 alt 4 hd 24 sec 49 apc 24", "", 0,0,
	      24,	21,	15,	1,	707,	4,	24,	49,	0, 0,
		{ 0, 16464 },	{ 14, 32928 },	{ 0, 831432 },	{ 0, 0 },
		{ 0, 0 },	{ 0 , 0 },	{ 42, 782040 },	{ 0, 0 },
		DKL_MAGIC, 0
	},
	{ "CDC-9720-368 cyl 1215 alt 2 hd 10 sec 50 apc 10", "", 0,0,
	       10,	15,	15,	1,     1215,   2,	10,	50,	0, 0,
		{ 0, 16500 },   { 33, 33000 },  { 0, 607500 }, { 0, 0 },
		{ 0, 0 }, 	{ 0, 0 }, 	{ 99, 558000 }, { 0, 0},
		DKL_MAGIC, 0
	},
	{ "CDC-9720-500 cyl 1215 alt 2 hd 10 sec 68 apc 10", "", 0,0,
	       10,	15,	15,	1,     1215,   2,	10,	68,	0, 0,
		{ 0, 17000 },   { 25, 34000 },  { 0, 826200 }, { 0, 0 },
		{ 0, 0 }, 	{ 0, 0 }, 	{ 75, 775200 }, { 0, 0},
		DKL_MAGIC, 0
	},
	{ "CDC-9720-736 cyl 1632 alt 3 hd 15 sec 50 apc 15", "", 0,0,
	       15,	15,	15,	1,     1632,   3,	15,	50,	0, 0,
		{ 0, 16500 },   { 22, 33000 },  { 0, 1224000 }, { 0, 0 },
		{ 0, 0 }, 	{ 0, 0 }, 	{ 66, 1174500 }, { 0, 0},
		DKL_MAGIC, 0
	},
	{ "CDC-9720-850 cyl 1378 alt 3 hd 15 sec 68 apc 15", "", 0,0,
	       15,	15,	15,	1,     1378,   3,	15,	68,	0, 0,
		{ 0, 17340 },   { 17, 34680 },  { 0, 1405560 }, { 0, 0 },
		{ 0, 0 }, 	{ 0, 0 }, 	{ 51, 1353540 }, { 0, 0},
		DKL_MAGIC, 0
	},
	{ "CDC-9720-1230 Sabre-V cyl 1633 alt 2 hd 15 sec 83 apc 15", "", 0,0,
	       15,	15,	15,	1,     1633,   2,	15,	83,	0, 0,
		{ 0, 17430 },   { 14, 34860 },  { 0, 2033085 }, { 0, 0 },
		{ 0, 0 }, 	{ 0, 0 }, 	{ 42, 1980795 }, { 0, 0},
		DKL_MAGIC, 0
	},
	{ "CDC-9766 cyl 820 alt 3 hd 19 sec 32 apc 19", "", 0,0,
	       19,	21,	15,	1,	820,	3,	19,	32,	0,	0,
		{ 0, 16416 },	{ 27, 33440 },	{ 0, 498560 },	{ 0, 0 },
		{ 0, 0 },	{ 0, 0 },	{ 82, 448704},	{ 0, 0 },
		DKL_MAGIC, 0
	},
	{ "CDC-9772-850 XMD-II cyl 1062 alt 2 hd 16 sec 82 apc 16", "", 0,0,
	       16,	19,	15,	1,     1062,   2,	16,	82,	0, 0,
		{ 0, 17056 },   { 13, 34112 },  { 0, 1393344 }, { 0, 0 },
		{ 0, 0 }, 	{ 0, 0 }, 	{ 39, 1342166 }, { 0, 0},
		DKL_MAGIC, 0
	},
	{ "CDC-9773-1230 cyl 1418 alt 2 hd 19 sec 84 apc 19", "", 0,0,
	       19,	15,	15,	1,     1418,   2,	19,	84,	0, 0,
		{ 0, 17556 },   { 11, 33516 },  { 0, 2263128 }, { 0, 0 },
		{ 0, 0 }, 	{ 0, 0 }, 	{ 32, 2212056 }, { 0, 0},
		DKL_MAGIC, 0
	},
	{ "Fujitsu-M2312K cyl 586 alt 3 hd 7 sec 33 apc 7", "", 0,0,
	        7,	15,	15,	1,	586,	3,	7,	33,	0,	0,
		{ 0, 16401 },	{ 71, 32802 },	{ 0, 135366 },	{ 0, 0 },
		{ 0, 0 },	{ 0, 0 },	{ 213, 86163 },	{ 0, 0 },
		DKL_MAGIC, 0
	},
	{ "Fujitsu-M2322K cyl 821 alt 2 hd 10 sec 33 apc 10", "", 0,0,
	        10,	15,	17,	1,	821,	2,	10,	33,	0,	0,
		{ 0, 16500 },	{ 50, 33000 },	{ 0, 270930 },	{ 0, 0 },
		{ 0, 0 },	{ 0, 0 },	{150, 221430},	{ 0, 0 },
		DKL_MAGIC, 0
	},
	{ "Fujitsu-2333 Swallow cyl 821 alt 2 hd 10 sec 67 apc 10", "", 0,0,
	      10,	16,	19,	1,	821,	2,	10,	67,	0, 0,
		{ 0, 17420 },   { 26, 32830 },  { 0, 550070 },  { 0, 0 },
		{ 0, 0 }, 	{ 0, 0 }, 	{ 75, 499820 }, { 0, 0 },
	        DKL_MAGIC, 0
	},
	{ "Fujitsu-M2344 Swallow II cyl 622 alt 2 hd 27 sec 67 apc 27", "", 0,0,
	       27,     16,     19,	1,	622,	2,	27,	67,	0,	0,
		{ 0, 21708 },	{ 12, 34371 },	{ 0, 1125198},	{ 0, 0 },
		{ 0, 0 },	{ 0, 0 },	{ 31, 1069119},	{ 0, 0 },
		DKL_MAGIC, 0
	},
	{ "Fujitsu-M2351 Eagle cyl 840 alt 2 hd 20 sec 45 apc 20", "", 0,0,
	       20,	18,	20,	1,	840,	2,	20,	45,	0,	0,
		{ 0, 17100 },	{ 19, 34200 },	{ 0, 756000 },	{ 0, 0 },
		{ 0, 0 },	{ 0, 0 },	{ 57, 704700 },	{ 0, 0 },
		DKL_MAGIC, 0
	},
	{ "Fujitsu-M2361A Eagle XP cyl 840 alt 2 hd 20 sec 66 apc 20", "", 0,0,
               20,	18,	20,	1,	840,	2,	20,	66,	0, 0,
		{ 0, 17160 },   { 13, 34320 },  { 0, 1108800 },  { 0, 0 },
		{ 0, 0 },       { 0, 0 },       { 39, 1057320 }, { 0, 0 },
		DKL_MAGIC, 0
	},
	{ "Fujitsu-M2372 Swallow III cyl 743 alt 2 hd 27 sec 67 apc 27", "", 0,0,
	       27,     20,     20,	1,	743,	2,	27,	67,	0,	0,
		{ 0, 21708 },	{ 12, 34371 },	{ 0, 1344087},	{ 0, 0 },
		{ 0, 0 },	{ 0, 0 },	{ 31, 1288008},	{ 0, 0 },
		DKL_MAGIC, 0
	},
	{ "Fujitsu-M2382 Swallow IV cyl 743 alt 2 hd 27 sec 83 apc 27 /SSP", "", 0,0,
	       27,     19,     20,	1,	743,	2,	27,	83,	0,	0,
		{ 0, 22410 },	{ 10, 33615 },	{ 0, 1665063},	{ 0, 0 },
		{ 0, 0 },	{ 0, 0 },	{ 25, 1609038},	{ 0, 0 },
		DKL_MAGIC, 0
	},
	{ "Hitachi DK815-10 cyl 1735 alt 2 hd 15 sec 67 apc 15 /SSP /EAD", "", 0,0,
	       15,	19,	19,	1,     1735,   2,	15,	67,	0, 0,
		{ 0, 17085 },   { 17, 34170 },  { 0, 1743675 }, { 0, 0 },
		{ 0, 0 }, 	{ 0, 0 }, 	{ 51, 1692420 }, { 0, 0},
		DKL_MAGIC, 0
	},
	{ "NEC-D2268H cyl 821 alt 2 hd 10 sec 67 apc 10", "", 0,0,
	       10,	19,	14,	1,	821,	2,	10,	67,	0,	0,
		{ 0, 16750 },	{ 25, 33500 },{ 0, 550070 },	{ 0, 0 },
		{ 0, 0 },	{ 0, 0 },	{ 75, 499820 },{ 0, 0 },
		DKL_MAGIC, 0
	},
	{ "NEC-D2352 cyl 758 alt 2 hd 19 sec 60 apc 19", "", 0,0,
	       19,     19,     14,	1,	758,	2,     19,	60,	0,	0,
		{ 0, 17100 },	{ 15, 34200 },	{ 0, 864120 },	{ 0, 0 },
		{ 0, 0 },	{ 0, 0 },	{ 45, 812820},	{ 0, 0 },
		DKL_MAGIC, 0
	},
	{ "NEC-D2362 cyl 848 alt 2 hd 23 sec 66 apc 23", "", 0,0,
	       23,     19,     12,	1,	848,	2,     23,	66,	0,	0,
		{ 0, 16698 },	{ 11, 33396 },	{ 0, 1287264 },	{ 0, 0 },
		{ 0, 0 },	{ 0, 0 },	{ 33, 1237170},	{ 0, 0 },
		DKL_MAGIC, 0
	},
	{ "NEC-D2363 cyl 1022 alt 2 hd 19 sec 67 apc 19", "", 0,0,
	       19,     16,     19,	1,	1022,	2,     19,	67,	0,	0,
		{ 0, 16549 },	{ 13, 33098 },	{ 0, 1301006},	{ 0, 0 },
		{ 0, 0 },	{ 0, 0 },	{ 39, 1251359},	{ 0, 0 },
		DKL_MAGIC, 0
	},
	{ "Rodime RO8074 cyl 1644 alt 2 hd 11 sec 63 apc 11 /SSP /EAD", "", 0,0,
	       11,	19,	19,	1,     1644,   2,	11,	63,	0, 0,
		{ 0, 16632 },   { 24, 33264 },  { 0, 1139292 }, { 0, 0 },
		{ 0, 0 }, 	{ 0, 0 }, 	{ 72, 1089396 }, { 0, 0},
		DKL_MAGIC, 0
	},
	{ "Toshiba MK-288FC cyl 820 alt 3 hd 15 sec 67 apc 15", "", 0,0,
		15,	19,	19,	1,	820,	3,	15,	67,	0,	0,
		{ 0, 17085 },	{ 17, 34170 },	{ 0, 824100 },	{ 0, 0 },
		{ 0, 0 },	{ 0, 0 },	{ 51, 772845 },	{ 0, 0 },
		DKL_MAGIC, 0
	},
	{ "Toshiba MK-388FA cyl 1160 alt 2 hd 15 sec 68 apc 15 /SSP", "", 0,0,
		15,	18,	18,	1,	1160,	2,	15,	68,	0,	0,
		{ 0, 21420 },	{ 21, 33660 },	{ 0, 1183200 },	{ 0, 0 },
		{ 0, 0 },	{ 0, 0 },	{ 54, 1128120 }, { 0, 0 },
		DKL_MAGIC, 0
	},
};
#endif /* SunOS4 1.12d */


static struct	dk_label *labels[2] = { smdlabels, esdilabels };
static int	nlabels[2] = { sizeof (smdlabels) / sizeof (struct dk_label),
			       sizeof (esdilabels)/ sizeof (struct dk_label) };

/*
 * This array defines strings which are used as labels when modifying the
 * disk configuration portion of the label.  They should be ordered exactly
 * as they are in the label structure.
 */
static char *fieldnames[] = {		/* starting with dkl_apc */
	"# of alt/cylinder",
	"size of gap1",		"size of gap2",		"interleave factor",
	"# of data cylinders",	"# of alt cylinders",	"# of heads",
	"# of sectors/track",	"label location",	"physical partition #",
	NULL
};

/*******************************************************************************
*									       *
*	Subroutine:		main                                           *
*									       *
*	Calling Sequence:	rfutil                                         *
*									       *
*	Called by:		shell                                          *
*									       *
*	Calling Parameters:	None                                           *
*				                                               *
*	Local Variables:	None                                           *
*				                                               *
*	Calls Subroutines:	                                   	       *
*           	help, doformat, dogeometry, dolabel, domap, doopen,            *
*           	dopartitions, readlabel, showlabel, writelabel, showstatistics *
*               doverify.						       *
*				                                               *
*	Public/Global Variables:None					       *
*				                                               *
*	Description:							       *
*	        This loops reading commands from the user and invokes a        *
*        	subroutine to act as required.                                 *
*		                                                               *
*******************************************************************************/

main(argc,argv)

int argc;
char *argv[];
{
	/* 1.15a */
	if (argc > 1)
		util_init(argv[1]);
	else
		util_init(DEFDEVICE);

	rflogo();
	do {
		help();
		prompt();
		line[0] = inp_char(stdscr);
		if (isupper(line[0]))
		    line[0] = tolower(line[0]);
		wmove(stdscr, 3, 0);
		wclrtoeol(stdscr);
		wmove(stdscr, 1, 0);
		wclrtoeol(stdscr);
		wmove(stdscr, 1, 0);
		waddstr(stdscr, "Command: ");
		switch (line[0]) {
		case 'b':
			waddstr(stdscr, "Debug control");
			debug();
			break;
		case 'c':
			addstr("Examine Controller Identification");
			doident();
			break;
		case 'd':
			waddstr(stdscr, "Examine/Modify IO Control groups");
			doiocg();
			break;
		case 'f':
			waddstr(stdscr, "Format disk");
			doformat();
			break;
		case 'g':
			waddstr(stdscr, "Examine disk geometry");
			dogeometry();
			break;
		case 'i':
			waddstr(stdscr, "Slip sectors");
			doslip();
			break;
		case 'l':
			waddstr(stdscr, "Choose label");
			dolabel();
			break;
		case 'm':
			waddstr(stdscr, "Map sectors/tracks");
			domap();
			break;
		case 'n':
			waddstr(stdscr, "New configuration");
			doconf();
			break;
		case 'o':
			waddstr(stdscr, "Open a new device");
			doopen();
			break;
		case 'p':
			waddstr(stdscr, "Examine/modify partitions");
			dopartitions();
			break;
		case 'q':
			waddstr(stdscr, "Quit");
			finish(0);
		case 'r':
			waddstr(stdscr, "Read label");
			readlabel();
			break;
		case 's':
			waddstr(stdscr, "Show label");
			showlabel();
			break;
		case 't':
			waddstr(stdscr, "Show statistics");
			showstatistics();
			break;
		case 'v':
			waddstr(stdscr, "Verify disk format");
			doverify();
			break;
		case 'w':
			waddstr(stdscr, "Write label to disk");
			writelabel();
			break;
		case 'a':  /* 1.9b */
			waddstr(stdscr, "Add automatic defect-handling list");
			abdefect();
			break;
                case 'z':
			waddstr(stdscr, "Read the defect map");
			readdef();
			break;
		case 'e':  /* 1.9c */
			addstr("Enquire disk characteristics");
			enqdc();
			break;
		case 'x': /* 1.16e */
			addstr("Examine Track IDs");
			doxtrack();
			break;
		default:
			wmove(stdscr, 3, 0);
			namechar(line);
			printw("Illegal command - '%s'",line);
		}
		wmove(stdscr, 1, 0);
		winsch(stdscr, ' ');
		winsch(stdscr, 't');
		winsch(stdscr, 's');
		winsch(stdscr, 'a');
		winsch(stdscr, 'L');
	} while (TRUE);
	/*NOTREACHED*/
}

char
inp_char(win)
register WINDOW *win;
{
	register char ch;

	wrefresh(win);
	while( ((ch = wgetch(win) &0x7f) == CHR_FF) || (ch == CHR_RPT) ) {
	    touchwin(curscr);
	    wrefresh(curscr);
	    }
	return (ch);
}

/* make a readable string out of an unknown character*/
namechar(str)
char *str;
{
    char c;

    c = *str;

    if (iscntrl(c)) {
	*(str++) = '^';
	*str = (c != 0177) ? (c | 0x60) : '^';
	}
    *(str+1) = '\0';
    }
	    
clriwin() {
	register int i;
	register int cnt;

	cnt = (11 < LINES ? 11 : LINES) - 4;
	for (i = 4; cnt; cnt--) {
		wmove(stdscr, i++, 0);
		wclrtoeol(stdscr);
	}
}

clrowin() {
	register int i;

	for (i = 11; i < LINES; i++) {
		wmove(stdscr, i, 0);
		wclrtoeol(stdscr);
	}
}

input(win, buf)
register WINDOW *win;
register char *buf;
{
	register char ch;
	register int cnt = 0;
	int cur_x, cur_y;

	getyx(win, cur_y, cur_x);

	while (TRUE) {
		wrefresh(win);
		ch = inp_char(win);
		if (ch == _tty.sg_kill) {
			cnt = 0;
			wmove(win, cur_y, cur_x);
			wclrtoeol(win);
		} else if (ch == _tty.sg_erase && cnt) {
			int nowcur_y, nowcur_x;
			getyx(win, nowcur_y, nowcur_x);
			--nowcur_x;
			--cnt;
			wmove(win, nowcur_y, nowcur_x);
			wclrtoeol(win);
		} else if (ch == '\n') {
			buf[cnt] = '\0';
			break;
		} else if (isprint(ch)) {
			buf[cnt++] = ch;
			if (waddch(win, ch) == ERR) {
				buf[cnt] = '\0';
				return(ERR);
			}
		}
	}
	return(OK);
} /* input */

finish(val)
register int val;
{
	wmove(stdscr, LINES-1, 0);
	wrefresh(stdscr);
	delwin(verwin);
	endwin();
	printf("\n");

	exit(val);
}


prompt() {
	char sbuf[32];

	cen_line(2, cname[ctype+1]);
	sprintf(sbuf,"Size: %d",disksize);
	move(2,COLS-strlen(sbuf)); addstr(sbuf);
	move( 2, 0);
	printw("Device: %s > ", diskname);
}

pak() {
	cen_line(LINES-1,"Press Any Key");
	inp_char(stdscr);
	}

util_init(device)
char *device;
{
	int util_abort();
	int oerr;

	if (initscr() == ERR) {
		error("Can't init curses.");
		exit(-1);
	}

	cbreak();
	noecho();

	signal(SIGQUIT,util_abort);
	signal(SIGINT,util_abort);

	if ((oerr = opendevice(device)) < 0) {
		char ebuf[128];

		move(2,0);
		printw("\n%s ",openerrlist[(-1)-oerr]);
		refresh();
		endwin();
		sprintf(ebuf,"can't open \"%s\"",device); /* 1.15a */
		error(ebuf);
		exit(-1);
	}

	init_verwin();
}

util_abort(sig)
int sig;
{
    WINDOW *abwin;
    WINDOW *savewin;
    int y,x;
    char c;

    /* eliminate the possibility of repeating this signal until we
       have processed it */
    signal(sig,SIG_IGN);

    /* remember current cursor co-ordinates */
    getyx (curscr,y,x);

    if ((abwin = newwin(0, 0, 0, 0)) == ERR) {
	endwin();
	exit(43);
	}

    touchwin(abwin); wrefresh(abwin);
    wmove(abwin,10,16);
    waddstr(abwin,"RFUTIL: really abort? (y/n): ");
    while (((c = inp_char(abwin) | LCBIT) != 'y') &&
		(c != 'n'))
	;
    if (c == 'y') {
	endwin();
	exit(1);
	}

    wmove(stdscr,y,x);
    touchwin(stdscr);
    wrefresh(stdscr);

    delwin(abwin);
    signal(sig,util_abort);

    } /* util_abort */

init_verwin() {
	if ((verwin = newwin(0, 0, 11, COLS-40)) == ERR) {
		closedevice();
		endwin();
		printf("\n");
		error("Can't init verify window.");
		exit(-1);
	}
	scrollok(verwin, TRUE);
}

error(str)
register char *str;
{
	fprintf(stderr, "%s\n", str);
	fflush(stderr);
}

rflogo()
{
    cen_line(0,TITLE);
    move(1,0); addstr(VERSION);
    move(1,COLS-strlen(COPYRIGHT)); addstr(COPYRIGHT);

    } /* rflogo */

/* cen_line:
 Center a message on stdscrn at the given line.  Assumes no new-lines,
 returns, tabs, or other unusual characters in the message */

cen_line(line,message)
int line;
char message[];
{
    move(line,0); clrtoeol();
    move(line,(COLS-strlen(message))/2);
    addstr(message);

    } /* cen_line */

mkdefname(devname,fname)
char *devname;		/* device name to start with (assumed /dev/something) */
char *fname;		/* file name built from device name		*/
{
	strcpy(fname,rindex(devname,'/')+1);
	strcat(fname,".dfl");
	} /* mkdefname */

/*******************************************************************************
*									       *
*	Subroutine:		help                                           *
*									       *
*	Calling Sequence:	help()                                         *
*									       *
*	Called by:		main                                           *
*									       *
*	Calling Parameters:	None                                           *
*				                                               *
*	Local Variables:	None                                           *
*				                                               *
*	Calls Subroutines:	None                               	       *
*				                                               *
*	Public/Global Variables:None					       *
*				                                               *
*	Description:							       *
*              This subroutine prints a short description of the               *
*              commands available.                                             *
*		                                                               *
*******************************************************************************/

help() {
	clrowin();
	clriwin();
	move(8, 0);
	addstr("Commands available are:\n\n");

	addstr("o - Open a disk device\n");
	addstr("s - Show the current label\n");
	addstr("l - Create or modify a label\n");
	addstr("r - Read label\n");
	addstr("w - Write label to the disk\n");
	addstr("v - Verify format\n");
	addstr("f - Format the drive\n");
	addstr("g - Examine disk geometry\n");
	addstr("p - Examine and modify partitions\n");
	addstr("x - Examine Track IDs\n"); /* 1.16e */
	move(10, COLS/2);
	addstr("z - Read the defect map\n");
	move(11, COLS/2);   /* 1.9b */
	addstr("a - Add automatic defect-handling list\n");
	move(12, COLS/2);
	addstr("b - Debug control\n");
	move(13, COLS/2);
	addstr("i - Slip sector\n");
	move(14, COLS/2);
	addstr("m - Map sector or track\n");
	move(15, COLS/2);
	addstr("n - Change configuration\n");
	move(16, COLS/2);
	addstr("t - Show Rimfire 3200/3400 statistics\n");
	move(17, COLS/2);
	addstr("d - Examine IO Control Groups\n");
	move(18, COLS/2);
	addstr("c - Examine Controller identification\n");
	move(19, COLS/2);
	addstr("e - Enquire Disk Characteristics\n");
	cen_line(22, "q - quit      ");
} /* help */

/*******************************************************************************
*									       *
*	Subroutine:		doopen                                         *
*									       *
*	Calling Sequence:	doopen()                                       *
*									       *
*	Called by:		main                                           *
*									       *
*	Calling Parameters:	None                                           *
*				                                               *
*	Local Variables:	None                                           *
*				                                               *
*	Calls Subroutines:	opendevice                             	       *
*				                                               *
*	Public/Global Variables: 					       *
*				                                               *
*                              line - buffer for reading user input            *
*                              disk - file descriptor of device file opened.   *
*				                                               *
*	Description:							       *
*              This subroutine reads a line from the user which it then        *
*              attempts to open as device file.  If a device file is already   *
*              open, it is closed.  If the open fails, "disk" is set to -1.    *
*		                                                               *
*******************************************************************************/

doopen()
{
	int oerr;		/* open error code		*/

	move(4, 0); clrtobot();

	/* ask for device names until 1) we get a good one, 2) they
	   press return to give up */
	do {
		move(4, 0);
		addstr("New Device: "); clrtoeol();
		line[0] = '\0';
		input(stdscr, line);
		if (line[0] && (line[0] != CHR_ESC)) {
			if ((oerr = opendevice(line)) < 0) {
				wmove(stdscr, 6, 0);
				printw("%s: %s",line,openerrlist[(-1)-oerr]);
				}
			}
		else
			line[0] = '\0';
		} while (line[0] && (disk < 0));

	} /* doopen */


closedevice()
{
	close(disk);
	glabel = 0;
	ctype = disk = -1;
	disksize = 0;
	strcpy(diskname,"(None)");

	} /* closedevice */

opendevice(dname)
register char *dname;
{
	struct stat statbuf;
	int part;		/* partition number	*/

	if (disk != -1) {
		closedevice();
	}

	if ((disk = open(dname, O_RDWR)) > 0) {
		/* found a file by that name that we can open --
		   check to make sure it is a device file */

		if(fstat(disk,&statbuf)) {
		    closedevice();
		    return(-4);
		    }
		if ((statbuf.st_mode & S_IFBLK) != S_IFCHR) {
		    /* not a character device type file --
		       better close it and report an error */
		    closedevice();
		    return(-2);
		    }

		/* get the controller type */
		/* NOTE: IF THIS CALL FAILS, THE DRIVER MAY BE TOO OLD
		   FOR USE WITH THIS UTILITY! */
		if ( (ioctl(disk, DKIOCGBTYP, &ctype) < 0) ||
			(ctype < 0) || (ctype > 1 )) {
		    /* Not a RF controller! */
		    closedevice();
		    return(-5);
		    }

		/* make the name official */
		strcpy(diskname, dname);


		if (getpartitions(label.dkl_map) < 0) {
			closedevice();
			return(-3);
			}
		part = diskname[strlen(diskname) - 1] - 'a';
		disksize = label.dkl_map[part].dkl_nblk;
		}
	else {
	    /* never opened the file */
	    disk = -1;
	    disksize = 0;
	    }

	return(disk);
}

/*******************************************************************************
*									       *
*	Subroutine:		readlabel                                      *
*									       *
*	Calling Sequence:	readlabel()                                    *
*									       *
*	Called by:		main                                           *
*									       *
*	Calling Parameters:	None                                           *
*				                                               *
*	Local Variables:	None                                           *
*				                                               *
*	Calls Subroutines:	None                               	       *
*				                                               *
*	Public/Global Variables:     					       *
*                   disk - file descriptor of device file to read label from   *
*                   label - disk label structure to read label into            *
*				                                               *
*	Description:							       *
*              This subroutine attempts to read the label from the device file *
*              currently open.                                                 *
*		                                                               *
*******************************************************************************/

readlabel()
{
	wmove(stdscr, 5, 0);
	wclrtoeol(stdscr);
	wmove(stdscr, 6, 0);
	wclrtoeol(stdscr);
	wmove(stdscr, 5, 0);
	if (disk == -1) {
		waddstr(stdscr, "Please use 'o' to open a disk device file");
		pak();
		return;
	}
	lseek(disk, 0L, 0);
	/* 1.9e */
	if (read(disk, (char *)&label, sizeof(struct dk_label)) != sizeof(struct dk_label)) {
		wprintw(stdscr, "Read: %s", sys_errlist[errno]);
		pak();
		return;
	}
	if (label.dkl_magic != DKL_MAGIC) {
		glabel = 0;
		printw("Warning: Bad magic number (0x%x) in label",
			label.dkl_magic);
		pak();
		}
	else {
		glabel = 1;
		showlabel();
		}

} /* readlabel */

/*******************************************************************************
*									       *
*	Subroutine:		writelabel                                     *
*									       *
*	Calling Sequence:	writelabel()                                   *
*									       *
*	Called by:		main                                           *
*									       *
*	Calling Parameters:	None                                           *
*				                                               *
*	Local Variables:	None                                           *
*				                                               *
*	Calls Subroutines:	None                               	       *
*				                                               *
*	Public/Global Variables:    					       *
*                     disk - file descriptor of device file to write label to  *
*                     label - disk label structure to write label from         *
*				                                               *
*	Description:							       *
*              This subroutine attempts to write the current label to          *
*              the device file currently open.                                 *
*		                                                               *
*******************************************************************************/

writelabel()
{
	struct dk_geom geometry;

	clriwin();
	clrowin();
	wmove(stdscr, 5, 0);
	if (disk == -1) {
		waddstr(stdscr, "Please use 'o' to open a disk device file");
		pak();
		return;
	}

	/* 1.10c */
	if (!glabel) {
		addstr("You must prepare a label with the 'l' command.\n");
		pak();
		return;
	}

	if (ioctl(disk, DKIOCGGEOM, &geometry) < 0) {
		addstr("Can't get the drive geometry from the kernel.\n");
		printw("DKIOCGGEOM: %s", sys_errlist[errno]);
		pak();
		return;
	}

	if (geometry.dkg_ncyl == 0 || geometry.dkg_acyl == 0
		|| geometry.dkg_nhead == 0 || geometry.dkg_nsect == 0) {
		addstr("\nYou must set the geometry with the 'g' command.\n");
		pak();
		return;
	}

	if (!disksize) {
		addstr("You must set the partitions with the 'p' command.\n");
		pak();
		return;
	}

	waddstr(stdscr,"Ready to write label, enter 'y' to proceed: ");
	if ((inp_char(stdscr)| LCBIT) != 'y') {
		waddstr(stdscr, "No");
		return;
	}
	waddstr(stdscr, "Write");
	complbck(&label);
	lseek(disk, 0L, 0);
	if (write(disk, (char *)&label, sizeof(struct dk_label)) < 0) {
		wmove(stdscr, 8, 0);
		wprintw(stdscr, "Write: %s", sys_errlist[errno]);
		pak();
		return;
	}
	return;
}

/*******************************************************************************
*									       *
*	Subroutine:		doformat                                       *
*									       *
*	Calling Sequence:	doformat()                                     *
*									       *
*	Called by:		main                                           *
*									       *
*	Calling Parameters:	None                                           *
*				                                               *
*	Local Variables:	                                               *
*                         lp - pointer to the current label                    *
*                         scyl, ecyl, shd, ehd -  starting and ending cylinder *
*                                                 and head.                    *
*                         lastcyl - last cylinder requested                    *
*				                                               *
*	Calls Subroutines:  ioctl - to pass the format command to the driver   *
*				                                               *
*	Public/Global Variables:    					       *
*                     disk - file descriptor of disk device to format          *
*                     label - the current label for disk configuration info    *
*                     line - buffer for user input                             *
*				                                               *
*	Description:							       *
*               This subroutine asks the user what portion of the disk to      *
*               format and then proceeds to format the specified area on a     *
*               cylinder by cylinder basis.                                    *
*		                                                               *
*******************************************************************************/

doformat()
{
	struct dk_label *lp = &label;
	struct dk_geom geometry;
	int scyl, ecyl, shd, ehd;
	int sblock, eblock, inptype;
	RETCONF retconf;
	extern char *index();
	FILE *dflf;		/* defect list file			*/
	char dflname[256];	/* defect list file name		*/
	int physsec;		/* physical sector size			*/
	char ch;

	clriwin();
	clrowin();
	wmove(stdscr, 5, 0);
	if (disk == -1) {
		addstr("You must open the disk first.");
		pak();
		return;
	}

	if (!glabel) {
		addstr("You must prepare a label with the 'l' command.\n");
		pak();
		return;
	}

	if (!disksize) {
		addstr("You must set the partitions with the 'p' command.\n");
		pak();
		return;
	}

	if (ioctl(disk, DKIOCGGEOM, &geometry) < 0) {
		addstr("Can't get the drive geometry from the kernel.\n");
		printw("DKIOCGGEOM: %s", sys_errlist[errno]);
		pak();
		return;
	}

	if (geometry.dkg_ncyl == 0 || geometry.dkg_acyl == 0
		|| geometry.dkg_nhead == 0 || geometry.dkg_nsect == 0) {
		addstr("Bad disk geometry in kernel:");
		showgeometry(&geometry);
		addstr("\nYou must set the geometry with the 'g' command.\n");
		pak();
		return;
	}

	if (ioctl(disk, DKIOCGSMD, &retconf) < 0) {
		waddstr(stdscr, "Can't get the SMD parameters from the kernel.\n");
		wprintw(stdscr, "DKIOCGSMD: %s", sys_errlist[errno]);
		pak();
		return;
	}

	showgeometry(&geometry);

	wmove(stdscr, 9, COLS/2);
	waddstr(stdscr, "SMD Parameters\n");
	wmove(stdscr, 10, COLS/2);
	wprintw(stdscr, "Bytes/Sec    : %d\n", retconf.bytsec);
	wmove(stdscr, 11, COLS/2);
	wprintw(stdscr, "Cyls/Disk    : %d\n", retconf.cyldsk);
	wmove(stdscr, 12, COLS/2);
	wprintw(stdscr, "NSpares      : %d\n", retconf.nspares);
	wmove(stdscr, 13, COLS/2);
	wprintw(stdscr, "Sec/Trk      : %d\n", retconf.sectrk);
	wmove(stdscr, 14, COLS/2);
	wprintw(stdscr, "Head/Cyl     : %d\n", retconf.headcyl);
	wmove(stdscr, 15, COLS/2);
	wprintw(stdscr, "Basehead     : %d\n", retconf.basehead);
	wmove(stdscr, 16, COLS/2);
	wprintw(stdscr, "Flags        : %d\n", retconf.flags );
	wmove(stdscr, 17, COLS/2);
	wprintw(stdscr, "Sec/Disk     : %d\n", retconf.secdsk);
	wmove(stdscr, 18, COLS/2);
	wprintw(stdscr, "Interleave   : %d\n", retconf.interleave);
	wmove(stdscr, 19, COLS/2);
	wprintw(stdscr, "Headskew     : %d\n", retconf.headskew);
	wmove(stdscr, 20, COLS/2);
	wprintw(stdscr, "Cylskew      : %d\n", retconf.cylskew);
	wmove(stdscr, 21, COLS/2);
	wprintw(stdscr, "Recovery     : %d\n", retconf.recovery);
	wmove(stdscr, 22, COLS/2);
	wprintw(stdscr, "ID Preamble  : %d\n", retconf.idpre);
	wmove(stdscr, 23, COLS/2);
	wprintw(stdscr, "Data Preamble: %d\n", retconf.datapre);

	mkdefname(diskname,dflname);

	move(5, 0);

#ifndef STANDALONE
	if ((dflf = fopen(dflname,"r")) != NULL) {
	    
	    printw("Opened \"%s\" for defect list.  OK? (y/n) ",dflname);

	    if ((inp_char(stdscr) | LCBIT) != 'y') {
		/* They don't want that one -- close it up */
		fclose(dflf);  dflf = NULL;
		}
	    } /* opened a file for the defect list */
	
	else {
	    /* couldn't open default defect list file */
	    printw("Cannot open \"%s\" for defect list.",dflname);
	    }
	
	while (dflf == NULL) {
	    move(6,0);
	    addstr("Enter a file name for the defect list, or press RETURN for none:\n");

	    dflname[0] = '\0';
	    input(stdscr,dflname);
	    if (!dflname[0] || (dflname[0] == CHR_ESC)) {
		/* no defect file desired */
		break;
		}

	    if ((dflf = fopen(dflname,"r")) == NULL) {
		printw("\nCannot open %s: %s",dflname,sys_errlist[errno]);
		}

	    } /* while file not open and file desired */

#else
	if ( (dflf = fopen(dflname,"r")) == NULL) {
		addstr("No defect list");
	} else {
		addstr("Using defect list");
	}
#endif

	move(6,0); clrtoeol();

	if (dflf != NULL) {
	    register int minsec, maxsec;

	    /* The constant 33 below is derived from the controller overhead */
	    minsec = retconf.bytsec + retconf.datapre + retconf.idpre + 33;
	    maxsec = minsec + retconf.bytsec;
	    physsec = -1;
	    do {
		move(7,0); clrtoeol();
		printw("Enter the size in bytes of the physical sector(between %d and %d): ",
			minsec,maxsec);
		input(stdscr, line);
		sscanf(line,"%d",&physsec);
		} while ((physsec < minsec) || (physsec > maxsec));
	    }

	scyl = shd = 0;
	ecyl = lp->dkl_ncyl + lp->dkl_acyl - 1;
	ehd = lp->dkl_nhead - 1;
	sblock = 0;
	eblock = retconf.secdsk - 1;

	wmove(stdscr, 7, 0);
	wclrtoeol(stdscr);
	waddstr(stdscr, "Enter 'y' to format the whole disk: ");
	if ((inp_char(stdscr) | LCBIT) != 'y') {
		wmove(stdscr, 11, 0);
		wclrtoeol(stdscr);
		do {
			if (!getcyltrack("Starting", &scyl, &shd, &sblock, &inptype)) {
				if (dflf != NULL) fclose(dflf);
				return;
				}
			wmove(stdscr, 11, 0);
			waddstr(stdscr, "The starting block must start on a head boundary.\n");
		} while (sblock % lp->dkl_nsect);
		wmove(stdscr, 11, 0);
		wclrtoeol(stdscr);
		do {
			if (!getcyltrack("Ending", &ecyl, &ehd, &eblock, &inptype)) {
				if (dflf != NULL) fclose(dflf);
				return;
				}
			wmove(stdscr, 11, 0);
			waddstr(stdscr, "The ending block must end on the last sector of a track.\n");
			if (inptype == 0)
				eblock += lp->dkl_nsect - 1;
		} while ((eblock % lp->dkl_nsect) != (lp->dkl_nsect - 1));
		wmove(stdscr, 11, 0);
		wclrtoeol(stdscr);
		if (sblock > eblock) {
			if (dflf != NULL) fclose(dflf);
			wmove(stdscr, 5, 0);
			waddstr(stdscr, "The start location must be before the end location.");
			pak();
			return;
		}
	}
	clriwin();
	wmove(stdscr, 5, 0);
	wprintw(stdscr, "Formatting %d/%d/%d to %d/%d/%d.\n",
		scyl, shd, 0, ecyl, ehd, lp->dkl_nsect-1);
	wprintw(stdscr, "Blocks %u to %u\nAre you sure? ", sblock, eblock);
	if ((inp_char(stdscr) | LCBIT) != 'y') {
		if (dflf != NULL) fclose(dflf);
		return;
	}
	waddstr(stdscr, "Yes");
	wmove(stdscr, 10, 0);
	waddstr(stdscr, "Formatting:\n");
	clrowin();
	if (fmt_to_cyl(&sblock, &eblock) && fmt_cyls(&sblock, &eblock) &&
		fmt_to_hd(&sblock, &eblock)) {
		wmove(stdscr, 10, 12);
		waddstr(stdscr, "Done");

	/* the disk is formatted, now time for automatic defect handling */

	if (dflf != NULL) {
	    register int i;
	    int cyl, hd, pos, len;
	    int acyl;

	    wclear(verwin);
	    move(10,0); clrtobot(); 
	    addstr("Processing Defect List"); refresh();
	    while (fscanf(dflf," %d %d: %d,%d", &cyl, &hd, &pos, &len) == 4) {
		if ( ((cyl > scyl) || ((cyl == scyl) && (hd >= shd))) &&
		     ((cyl < ecyl) || ((cyl == ecyl) && (hd <= ehd))) &&
		      (cyl < (lp->dkl_ncyl - 1)) ) {
		    for_map(&retconf,physsec, cyl, hd, pos, (len + 7)/8 );

		    } /* if the defect is within the formatted area */

		} /* while not end of defect list */

	    fclose(dflf);

	    } /* if there are defects to map */

	/* Ask about writing out the label -- probably want to 
	 * if the format covered cylinder 0 head 0.
	 */
	if ((scyl + shd) == 0) {
		move(6,0);
		addstr("Do you wish to write out the label at this time? (y/n): ");
		while (((ch = inp_char(stdscr) | LCBIT) != 'y') &&
			    (ch != 'n'))
		    ;
		if (ch == 'y') {
		    waddstr(stdscr, "Write");
		    complbck(&label);
		    lseek(disk, 0L, 0);
		    if (write(disk, (char *)&label, sizeof(struct dk_label)) < 0) {
			    wmove(stdscr, 8, 0);
			    wprintw(stdscr, "Write: %s", sys_errlist[errno]);
			    pak();
			    }
		    }
		}

		} /* format completed successfully */
	pak();

} /* doformat */

for_map(retconf,physsec,cyl,hd,badbyte,len)
RETCONF *retconf;
int physsec,cyl,hd,badbyte;
{
	struct dk_mapr map;
	struct dk_slip slip;
	int badblk;

	map.dkm_nblk = compsec(retconf,physsec,cyl,hd,badbyte,len,&badblk);
	map.dkm_fblk = badblk;
	map.dkm_tblk = -1;
	map.dkm_recover = 0;

	/* if the error is "harmless," ignore it. */
	if (!map.dkm_nblk) return;

	if (map.dkm_nblk == 1) {
	    /* try to slip the sector */
	    slip.dks_fblk = map.dkm_fblk;
	    slip.dks_recover = 0;
	    slip.dks_error = 0;

	    /* pass pointer to the driver for the dirty work */
	    if (ioctl(disk, DKIOCSLP, &slip) < 0) {
		    wprintw(stdscr, "DKIOCSLP: %s", sys_errlist[errno]);
	    }

	    /* check for error returned by the driver */
	    else if (slip.dks_error == 0) {
		    wprintw(verwin,"Block %d slipped on track.\n",
			slip.dks_fblk);
		    wrefresh(verwin);
		    return;
	    }
	}

	/* map the track or sector */
	ioctl(disk, DKIOCMAP, &map);
	if (map.dkm_error != 0) {
		wprintw(stdscr, "DKIOCMAP: %02x, - %s", map.dkm_error,
			rfmsg[map.dkm_error]); /* 1.10a */
		pak();
		return;
	}
	if (map.dkm_nblk == 1) {
		wprintw(verwin,"Block %d mapped to alt.\n", map.dkm_fblk);
	} 
	else {
		wprintw(verwin,"Track @ block %d mapped to alt.\n",
			map.dkm_fblk);
	}
	wrefresh(verwin);

} /* for_map */

/* 1.16d */
/* This routine computes the first block and number of blocks to map, based on
 * the physical sector size, the device geometry, the physical byte number,
 * and the length (in bytes) of the error.  If the error occurs in a "bad"
 * place, the entire track must be mapped.  If the error is long enough to
 * cross a sector boundry the routine notes that.  The return value is 0 for
 * a "harmless" error (one in the "short" sector of an SMD drive), or the
 * number of sectors to map.  Errors within the head scatter and those within
 * gap 3 are no longer considered "harmless".  Currently, one of two values
 * will be returned for the number of sectors to map in the case that mapping
 * must be done.  They are 1, or the number of sectors per track.  This is
 * because an error long enough to affect two sectors must take out the
 * critical bytes of the trailing sector.
 *
 * Error returns are negative.
 */

#define HDSCAT	(8)
#define CRITBYT	(HDSCAT+1+6+2+3)

int
compsec(retconf, physsec, cylnum, hdnum, badbyte, badlen, first)
RETCONF *retconf;
int physsec, cylnum, hdnum, badbyte, badlen, *first;
{
        int	idb;		/* number of bytes in ID field and preamble */
	int	leadb;		/* number of bytes into sector of error */
	int	badsect;	/* physical sector number containing error*/
	long	offset;		/* number of blocks before track
				   containing the error	*/
        TRAKID trakid;		/* structure for IOCTL call passing sectoring
				   info */

	offset = retconf->sectrk * (retconf->headcyl*cylnum + hdnum);

	/* 1.9a If the length is given as 0, it means bad track */
	if ( !badlen ){
	    *first = offset;
	    return(retconf->sectrk);
	    }

	/* compute the physical sector number and the byte offset
	   into the physical sector */
	badsect = badbyte/physsec;
	leadb = badbyte % physsec;

	/* compute the "critical" bytes in the sector -- those comprising
	   the sector id and preamble */
	idb = CRITBYT  + retconf->idpre;

	/* 1.16d -- removed code for other "harmless" errors.

	/* If the address area of either sector is unreliable it means bad
	 * track 1.16d */
	if ( ( ((leadb + badlen) >= HDSCAT) && (leadb <= idb) ) ||
		((leadb + badlen) > (physsec + HDSCAT))  ){
	    *first = offset;
	    return(retconf->sectrk);
	    }

	/* prepare for the ioctl request */
	trakid.cyl=cylnum;
	trakid.sec=badsect;
	trakid.head=hdnum;

	if (ioctl(disk, DKIOCPHYSID, &trakid) < 0) { /* 1.16e */
	    return(-1);
	    }

	*first = trakid.sec + offset;

        /* CHECKING CONDITIONS OF SECTOR */

	if ( trakid.sec==SEC_SPR && trakid.flag==FLG_NRM && trakid.alt==SEC_SLP) {
	    /* spare sector -- must map track */
	    *first = offset;
	    return(retconf->sectrk);
	    }
	else if ( (trakid.sec==SEC_SHR && trakid.flag==FLG_NRM &&
			trakid.alt==SEC_SLP) ||
	          (trakid.flag==FLG_DSC) ||
	          (trakid.sec==SEC_SLP && trakid.flag==FLG_NRM &&
			trakid.alt==SEC_SLP) ) {
	    /* short sector, or already mapped or slipped  -- no problem */
	    return(0);
	    }

	/* normal sector -- slip it or map it at the controller's choice */
	return(1);

    }  /* compsec */

getcyltrack(which, cylp, hdp, bp, kind)
char *which;
register int *cylp, *hdp, *bp, *kind;
{
	struct dk_label *lp = &label;
	int cyl, head, block;

	clriwin();

	for (;;) {
		move(5, 0); clrtoeol();
		addstr("'q' to quit\n");

		clrtoeol(); /* 1.16f */
		printw("Enter %s Block or Cylinder/Head: ", which);
		input(stdscr, line);
		if ((line[0] | LCBIT) == 'q' || line[0] == CHR_ESC)
			return(FALSE);

		move(8,0); clrtoeol();

		if (sscanf(line, "%d/%d", &cyl, &head) == 2) {
			block = cyl * lp->dkl_nhead * lp->dkl_nsect + 
				head * lp->dkl_nsect;
			*cylp = cyl;
			*hdp = head;
			*bp = block;
			*kind = 0;
			return(TRUE);
			} /* cyl/head entered */
		else if (sscanf(line, "%d", &block) == 1) {
			*kind = 1;
			if (block == -1) {
				*cylp = 0;
				*hdp = 0;
				*bp = -1;
				return(TRUE);
				}
			*cylp = (block / (lp->dkl_nhead * lp->dkl_nsect));
			*hdp = (block % (lp->dkl_nhead * lp->dkl_nsect)) /
					lp->dkl_nsect;
			*bp = block;
			return(TRUE);
			} /* block number entered */
		else {
			wprintw(stdscr, "Illegal input: %s", line);
		    }
	} /* forever */

} /* getcyltrack */

getbnchs(which, bp)
char *which;
register int *bp;
{
	struct dk_label *lp = &label;
	int cyl, head, block, sector;

	clriwin();

	for (;;) {
		move(5, 0); clrtoeol();
		addstr("'q' to quit\n");

		clrtoeol();
		printw("Enter %s Block or Cylinder/Head/Sector: ", which);
		input(stdscr, line);
		if ((line[0] | LCBIT) == 'q' || line[0] == CHR_ESC)
			return(FALSE);

		move(8,0); clrtoeol();

		if (sscanf(line, "%d/%d/%d", &cyl, &head, &sector) == 3) {
			*bp = cyl * lp->dkl_nhead * lp->dkl_nsect + 
				head * lp->dkl_nsect + sector;
			return(TRUE);
			} /* cyl/head entered */
		else if (sscanf(line, "%d", &block) == 1) {
			*bp = block;
			return(TRUE);
			} /* block number entered */
		else {
			wprintw(stdscr, "Illegal input: %s", line);
		    }
	} /* forever */

} /* getbnchs */

fmt_to_cyl(sbp, ebp)
register int *sbp, *ebp;
{
	struct dk_label *lp = &label;
	register int shd, ehd, scyl, ecyl;
	struct dk_fmt fmt;

	if (*sbp > *ebp)
		return(FALSE);

	shd = (*sbp % (lp->dkl_nhead * lp->dkl_nsect) / lp->dkl_nsect);
	ehd = (*ebp % (lp->dkl_nhead * lp->dkl_nsect) / lp->dkl_nsect);
	scyl = *sbp / (lp->dkl_nhead * lp->dkl_nsect);
	ecyl = *ebp / (lp->dkl_nhead * lp->dkl_nsect);
	if (scyl != ecyl)
		ehd = lp->dkl_nhead - 1;

	if (shd != 0) {
		while (shd <= ehd) {
			wmove(stdscr, 12, 0);
			wprintw(stdscr, "Track %4d/%d", scyl, shd);
			wrefresh(stdscr);
			fmt.dkf_blkno = scyl * lp->dkl_nhead * lp->dkl_nsect
					+ shd * lp->dkl_nsect;
			fmt.dkf_nblk = lp->dkl_nsect;
			/* 1.9d */
			ioctl(disk, DKIOCFMT, &fmt);
			if (fmt.dkf_error) {
				move(12, 20);
				printw("DKIOCFMT: %02x - %s @ %d",  /* 1.10e */
					fmt.dkf_error, rfmsg[fmt.dkf_error],
					fmt.dkf_badblk);
				return(FALSE);
			}
			shd++;
			*sbp += lp->dkl_nsect;
		}
	}
	return(TRUE);
} /* fmt_to_cyl */

fmt_cyls(sbp, ebp)
register int *sbp, *ebp;
{
	struct dk_label *lp = &label;
	register int shd, ehd, scyl, ecyl;
	struct dk_fmt fmt;

	if (*sbp > *ebp)
		return(TRUE);

	shd = (*sbp % (lp->dkl_nhead * lp->dkl_nsect) / lp->dkl_nsect);
	ehd = (*ebp % (lp->dkl_nhead * lp->dkl_nsect) / lp->dkl_nsect);
	scyl = *sbp / (lp->dkl_nhead * lp->dkl_nsect);
	ecyl = *ebp / (lp->dkl_nhead * lp->dkl_nsect);
	if (ehd != (lp->dkl_nhead - 1))
		ecyl--;

	while (scyl <= ecyl) {
		wmove(stdscr, 13, 0);
		wprintw(stdscr, "Cylinder %4d", scyl);
		wrefresh(stdscr);
		fmt.dkf_blkno = scyl * lp->dkl_nhead * lp->dkl_nsect;
		fmt.dkf_nblk = lp->dkl_nhead * lp->dkl_nsect;
		/* 1.9d */
		ioctl(disk, DKIOCFMT, &fmt);
		if (fmt.dkf_error) {
			move(12, 20);
			printw("DKIOCFMT: %02x - %s @ %d",  /* 1.10e */
				fmt.dkf_error, rfmsg[fmt.dkf_error],
				fmt.dkf_badblk);
			return(FALSE);
		}
		scyl++;
		*sbp += (lp->dkl_nhead * lp->dkl_nsect);
	}
	return(TRUE);
} /* fmt_cyls */

fmt_to_hd(sbp, ebp)
register int *sbp, *ebp;
{
	struct dk_label *lp = &label;
	register int shd, ehd, scyl, ecyl;
	struct dk_fmt fmt;

	if (*sbp > *ebp)
		return(TRUE);

	shd = (*sbp % (lp->dkl_nhead * lp->dkl_nsect) / lp->dkl_nsect);
	ehd = (*ebp % (lp->dkl_nhead * lp->dkl_nsect) / lp->dkl_nsect);
	scyl = *sbp / (lp->dkl_nhead * lp->dkl_nsect);
	ecyl = *ebp / (lp->dkl_nhead * lp->dkl_nsect);

	while (shd <= ehd) {
		wmove(stdscr, 14, 0);
		wprintw(stdscr, "Track %4d/%d", scyl, shd);
		wrefresh(stdscr);
		fmt.dkf_blkno = scyl * lp->dkl_nhead * lp->dkl_nsect
				+ shd * lp->dkl_nsect;
		fmt.dkf_nblk = lp->dkl_nsect;
		/* 1.9d */
		ioctl(disk, DKIOCFMT, &fmt);
		if (fmt.dkf_error) {
			move(12, 20);
			printw("DKIOCFMT: %02x - %s @ %d",  /* 1.10e */
				fmt.dkf_error, rfmsg[fmt.dkf_error],
				fmt.dkf_badblk);
			return(FALSE);
		}
		shd++;
		*sbp += lp->dkl_nsect;
	}
	return(TRUE);
} /* fmt_to_hd */

/*******************************************************************************
*									       *
*	Subroutine:		doverify                                       *
*									       *
*	Calling Sequence:	doverify()                                     *
*									       *
*	Called by:		main                                           *
*									       *
*	Calling Parameters:	None                                           *
*				                                               *
*	Local Variables:	                                               *
*                         lp - pointer to the current label                    *
*                         vfy - disk verify structure to pass to the driver    *
*                         scyl, ecyl, shd, ehd -  starting and ending cylinder *
*                                                 and head.                    *
*                         lastcyl - last cylinder requested                    *

*					                                       *
*	Calls Subroutines:  ioctl - to pass the format command to the driver   *
*				                                               *
*	Public/Global Variables:    					       *
*                     disk - file descriptor of disk device to format          *
*                     label - the current label for disk configuration info    *
*                     line - buffer for user input                             *
*				                                               *
*	Description:							       *
*               This subroutine asks the user what portion of the disk to      *
*               verify and then proceeds to verify the specified area on a     *
*               cylinder by cylinder basis.                                    *
*		                                                               *
*******************************************************************************/

doverify()
{
	struct dk_label *lp = &label;
	int scyl, ecyl, shd, ehd;
	int domaps = FALSE;
	int numtimes, sblock, eblock, inptype;
	extern char *index();

	clriwin();
	clrowin();
	wmove(stdscr, 5, 0);
	if (disk == -1) {
		waddstr(stdscr, "You must open the disk first.");
		pak();
		return;
	}
	if (!glabel) {
		addstr("You must read the label using the 'r' command, or\n");
		addstr("set the label using the 'l' command.");
		pak();
		return;
		}

	move(7, 0);
	addstr("How many passes do you want to make? ");
	input(stdscr, line);
	if (line[0] == '\0' || (line[0] | LCBIT) == 'q')
		return;
	numtimes = atoi(line);
	if (numtimes < 1)
		numtimes = 1;
	wmove(stdscr, 8, 0);
	wprintw(stdscr, "OK, doing %d verify passes.\n", numtimes);
	wmove(stdscr, 9, 0);
	waddstr(stdscr, "Do you want to map bad sectors as they are discovered? ");
	if ((inp_char(stdscr) | LCBIT) != 'y') {
		waddstr(stdscr, "No");
		domaps = FALSE;
	} else {
		waddstr(stdscr, "Yes");
		domaps = TRUE;
	}
	wmove(stdscr, 10, 0);
	waddstr(stdscr, "Ready to verify disk, enter 'y' to proceed: ");
	if ((inp_char(stdscr) | LCBIT) != 'y') {
		waddstr(stdscr, "No");
		return;
	}

	scyl = shd = 0;
	ecyl = lp->dkl_ncyl + lp->dkl_acyl - 1;
	ehd = lp->dkl_nhead - 1;
	sblock = 0;
	eblock = lp->dkl_nsect * lp->dkl_nhead * (ecyl + 1) - 1;

	clriwin();
	wmove(stdscr, 5, 0);
	waddstr(stdscr, "Enter 'y' to verify the whole disk: ");
	if ((inp_char(stdscr) | LCBIT) != 'y') {
		wmove(stdscr, 11, 0);
		wclrtoeol(stdscr);
		do {
			if (getcyltrack("Starting", &scyl, &shd, &sblock, &inptype) != TRUE) {
				pak();
				return;
				}
			wmove(stdscr, 11, 0);
			waddstr(stdscr, "The starting block must start on a head boundary.\n");
		} while ((sblock % lp->dkl_nsect));
		wmove(stdscr, 11, 0);
		wclrtoeol(stdscr);
		do {
			if (getcyltrack("Ending", &ecyl, &ehd, &eblock, &inptype) != TRUE) {
				pak();
				return;
				}
			wmove(stdscr, 11, 0);
			waddstr(stdscr, "The ending block must end on the last sector of a track.\n");
			if (inptype == 0)
				eblock += lp->dkl_nsect - 1;
		} while ((eblock % lp->dkl_nsect) != (lp->dkl_nsect - 1));
		wmove(stdscr, 11, 0);
		wclrtoeol(stdscr);
		if (sblock > eblock) {
			wmove(stdscr, 5, 0);
			waddstr(stdscr, "The start location must be before the end location.");
			pak();
			return;
		}
	}
	clriwin();
	wmove(stdscr, 5, 0);
	wprintw(stdscr, "Verifying %d/%d/%d to %d/%d/%d.\n",
		scyl, shd, 0, ecyl, ehd, lp->dkl_nsect-1);
	wprintw(stdscr, "Blocks %u to %u\nAre you sure? ", sblock, eblock);
	if ((inp_char(stdscr) | LCBIT) != 'y') {
		waddstr(stdscr, "No");
		return;
	}
	waddstr(stdscr, "Yes");
	wmove(stdscr, 10, 0);
	waddstr(stdscr, "Verifying:\n");
	clrowin();
	wclear(verwin);
	if (ver_to_cyl(&sblock, &eblock, domaps, numtimes) &&
		ver_cyls(&sblock, &eblock, domaps, numtimes) &&
		ver_to_hd(&sblock, &eblock, domaps, numtimes)) {
		wmove(stdscr, 10, 12);
		waddstr(stdscr, "Done");
	}
	pak();
}
ver_map(ver)
struct dk_vfy *ver;
{
	struct dk_label *lp = &label;
	struct dk_mapr map;
	struct dk_slip slip;

	if (ver->dkv_error == RFE_ECC) {
		/* first try to slip the sector  1.10d */
		slip.dks_fblk = ver->dkv_badblk;
		slip.dks_recover = 1;
		slip.dks_error = 0;
		ioctl(disk, DKIOCSLP, &slip);
		if (!slip.dks_error) {
		    wprintw(verwin, "Block %d slipped on track.\n",
			    slip.dks_fblk);
		    }
		else {
		    /* better to map the track than map a single sector, since
		     * there is already at least one error on the track, or
		     * we wouldn't be out of spare sectors, and mapping a track
		     * gives faster performance.  1.10d
		     */
		    if (slip.dks_error == RFE_NOSPARE)
			wprintw(verwin, "Out of spares --  mapping track.\n");
		    else {
			if (slip.dks_error != RFE_ECCFAIL)
			    wprintw(verwin, "Error %02x - %s\n", slip.dks_error,
				    rfmsg[slip.dks_error]);
			wprintw(verwin, "Need to map track.\n");
			}
		    map.dkm_fblk = (ver->dkv_badblk / lp->dkl_nsect);
		    map.dkm_fblk *= lp->dkl_nsect;
		    map.dkm_tblk = -1;
		    map.dkm_nblk = lp->dkl_nsect;
		    map.dkm_recover = 2;
		    ioctl(disk, DKIOCMAP, &map);
		    if (map.dkm_error) {
		        wprintw(stdscr, "DKIOCMAP: %02x - %s", map.dkm_error,
				rfmsg[map.dkm_error]); /* 1.10a */
			pak();
			return;
			}
		    wprintw(verwin, "Track @ block %d mapped to alt.\n",
			map.dkm_fblk);
		    } /* slip error was NOSPARE */

		} /* Error was ECC -- try map then slip */
	    else {
		wprintw(verwin, "Error %02x - %s\n", ver->dkv_error,
			rfmsg[ver->dkv_error]);
		wprintw(verwin, "Need to map track.\n");
		map.dkm_fblk = (ver->dkv_badblk / lp->dkl_nsect);
		map.dkm_fblk *= lp->dkl_nsect;
		map.dkm_tblk = -1;
		map.dkm_nblk = lp->dkl_nsect;
		map.dkm_recover = 2;
		ioctl(disk, DKIOCMAP, &map); /* 1.10d */
		if (map.dkm_error) {
		    wprintw(stdscr, "DKIOCMAP: %02x - %s", map.dkm_error,
			    rfmsg[map.dkm_error]); /* 1.10a */
		    pak();
		    return;
		    }
		wprintw(verwin, "Track @ block %d mapped to alt.\n", map.dkm_fblk);
		} /* error other than ECC -- map track */
	    wrefresh(verwin);
    } /* ver_map */

ver_to_cyl(sbp, ebp, domaps, numtimes)
int *sbp, *ebp, domaps, numtimes;
{
	struct dk_label *lp = &label;
	register int shd, ehd, scyl, ecyl;
	int i;
	struct dk_vfy ver;

	if (*sbp > *ebp)
		return(FALSE);

	shd = (*sbp % (lp->dkl_nhead * lp->dkl_nsect) / lp->dkl_nsect);
	ehd = (*ebp % (lp->dkl_nhead * lp->dkl_nsect) / lp->dkl_nsect);
	scyl = *sbp / (lp->dkl_nhead * lp->dkl_nsect);
	ecyl = *ebp / (lp->dkl_nhead * lp->dkl_nsect);
	if (scyl != ecyl)
		ehd = lp->dkl_nhead - 1;

	if (shd != 0) {
		while (shd <= ehd) {
			wmove(stdscr, 12, 0);
			wprintw(stdscr, "Track %4d/%d", scyl, shd);
			wrefresh(stdscr);
			for (i = numtimes; i; i--) {
				ver.dkv_blkno = scyl * lp->dkl_nhead *
							lp->dkl_nsect
						+ shd * lp->dkl_nsect;
				ver.dkv_nblk = lp->dkl_nsect;
				ver.dkv_error = 0;
				ver.dkv_badblk = 0;
				ioctl(disk, DKIOCVFY, &ver);
				if (ver.dkv_error != 0) {
					wprintw(verwin,
					  "Verify failed: %d/%d/%d, block=%d\n",
					   scyl, shd,	/* 1.16h */
					   ver.dkv_badblk % lp->dkl_nsect,
					   ver.dkv_badblk);
					wrefresh(verwin);
					if (domaps)
						ver_map(&ver);
				}
			}
			shd++;
			*sbp += lp->dkl_nsect;
		}
	}
	return(TRUE);
}

ver_cyls(sbp, ebp, domaps, numtimes)
register int *sbp, *ebp, domaps, numtimes;
{
	struct dk_label *lp = &label;
	int ehd, scyl, ecyl;
	register int i;
	register int head;	/* current head to verify	*/
	struct dk_vfy ver;

	if (*sbp > *ebp)
		return(FALSE);

	ehd = (*ebp % (lp->dkl_nhead * lp->dkl_nsect) / lp->dkl_nsect);
	scyl = *sbp / (lp->dkl_nhead * lp->dkl_nsect);
	ecyl = *ebp / (lp->dkl_nhead * lp->dkl_nsect);
	if (ehd != (lp->dkl_nhead - 1))
		ecyl--;

	while (scyl <= ecyl) {
		wmove(stdscr, 13, 0);
		wprintw(stdscr, "Cylinder %4d", scyl);
		wrefresh(stdscr);
		for (i = numtimes; i; i--) {
			/* 1.12c */
			for (head = 0; head < lp->dkl_nhead; head++) {
			    ver.dkv_blkno = (scyl *lp->dkl_nhead + head)
				* lp->dkl_nsect;	/* 1.16h */
			    ver.dkv_nblk = lp->dkl_nsect;
			    ver.dkv_error = 0;
			    ver.dkv_badblk = 0;
			    ioctl(disk, DKIOCVFY, &ver);
			    if (ver.dkv_error != 0) {
				    wprintw(verwin,
				       "Verify failed: %d/%d/%d, block=%d\n",
					scyl, head, ver.dkv_badblk % /* 1.16h */
					lp->dkl_nsect, ver.dkv_badblk);
				    wrefresh(verwin);
				    if (domaps)
					    ver_map(&ver);
			    }
			}
		}
		scyl++;
		*sbp += (lp->dkl_nhead * lp->dkl_nsect);
	}
	return(TRUE);
}

ver_to_hd(sbp, ebp, domaps, numtimes)
int *sbp, *ebp, domaps, numtimes;
{
	struct dk_label *lp = &label;
	register int shd, ehd, scyl, ecyl;
	int i;
	struct dk_vfy ver;

	if (*sbp > *ebp)
		return(FALSE);

	shd = (*sbp % (lp->dkl_nhead * lp->dkl_nsect) / lp->dkl_nsect);
	ehd = (*ebp % (lp->dkl_nhead * lp->dkl_nsect) / lp->dkl_nsect);
	scyl = *sbp / (lp->dkl_nhead * lp->dkl_nsect);
	ecyl = *ebp / (lp->dkl_nhead * lp->dkl_nsect);

	while (shd <= ehd) {
		wmove(stdscr, 14, 0);
		wprintw(stdscr, "Track %4d/%d", scyl, shd);
		wrefresh(stdscr);
		for (i = numtimes; i; i--) {
			ver.dkv_blkno = scyl * lp->dkl_nhead * lp->dkl_nsect
					+ shd * lp->dkl_nsect;
			ver.dkv_nblk = lp->dkl_nsect;
			ver.dkv_error = 0;
			ver.dkv_badblk = 0;
			ioctl(disk, DKIOCVFY, &ver);
			if (ver.dkv_error != 0) {
				wprintw(verwin,
					"Verify failed: %d/%d/%d, block=%d\n",
					scyl, shd, ver.dkv_badblk % /* 1.16h */
					lp->dkl_nsect, ver.dkv_badblk);
				wrefresh(verwin);
				if (domaps)
					ver_map(&ver);
			}
		}
		shd++;
		*sbp += lp->dkl_nsect;
	}
	return(TRUE);
}

/*******************************************************************************
*									       *
*	Subroutine:		debug                                          *
*									       *
*	Calling Sequence:	debug()                                        *
*									       *
*	Called by:		main                                           *
*									       *
*	Calling Parameters:	None                                           *
*				                                               *
*	Local Variables:  dbg_var - the debug variable from the driver.        *
*				                                               *
*	Calls Subroutines: ioctl - to pass the ioctl request to the driver     *
*				                                               *
*	Public/Global Variables:    					       *
*               disk - file descriptor of disk device to query for the cache   *
*               line - buffer for user input                                   *
*				                                               *
*	Description:							       *
*               This routine prints out the debug variable and then            *
*               queries the user for a new value.                              *
*		                                                               *
*******************************************************************************/

debug()
{
	unsigned long dbg_var;

	clriwin();
	clrowin();
	wmove(stdscr, 5, 0);
	if (disk < 0) {
		waddstr(stdscr, "Please use 'o' to open a disk device file");
		pak();
		return;
	}

	if (ioctl(disk, DKIOCGDEBUG, &dbg_var) < 0) {
		wprintw(stdscr, "Get debug: %s", sys_errlist[errno]);
		pak();
		return;
	}

	addstr("Debug bit-wise values are:\n");
	addstr("\t0xFF -- general trace and pb/sb dump\n");


	printw("The current debug variable setting is: 0x%x\n", dbg_var);
	waddstr(stdscr, "Do you want to change it? ");
	if ((inp_char(stdscr) | LCBIT) != 'y') {
		waddstr(stdscr, "No");
		return;
	}
	waddstr(stdscr, "Yes\n\n");

	wprintw(stdscr, "Enter new value (in hex) : 0x");
	input(stdscr, line);
	sscanf(line, "%x", &dbg_var);

	if (ioctl(disk, DKIOCSDEBUG, &dbg_var) < 0) {
		printw("\nSet debug: %s\n", sys_errlist[errno]);
		addstr("Debug value probably not changed.");
		pak();
		return;
	}
	printw("\nDebug value changed to 0x%x\n", dbg_var);
	pak();
}

/*******************************************************************************
*									       *
*	Subroutine:		doxtrack				       *
*									       *
*	Calling Sequence:	doxtrack()				       *
*									       *
*	Called by:		main                                           *
*									       *
*	Calling Parameters:	None                                           *
*				                                               *
*	Local Variables:  ident - structure returning controller info.         *
*				                                               *
*	Calls Subroutines: ioctl - to pass the ioctl request to the driver     *
*				                                               *
*	Public/Global Variables:    					       *
*               disk - file descriptor of disk device to query for the cache   *
*				                                               *
*	Description:							       *
*               This routine requests the Identification information from the  *
*		controller and displays that information.		       *
*		                                                               *
*******************************************************************************/

doxtrack()
{
	TRAKID *tid;	/* structure returning identification info	*/
	struct dk_geom geometry; /* disk geometry			*/
	struct dk_geom *gp;	/* pointer to above			*/
	int spc;	/* sectors per cylinder				*/
	int cyl, hd;	/* disk address					*/
	int inptype;	/* parameter dummy				*/
	register int i;	/* loop control					*/
	int skew;	/* sector skew of this track			*/
	int tsect;	/* total full-sized sectors on a track		*/
	int trksize;	/* total number of IDs on a track		*/
	int block;	/* logical block number from start		*/

	clriwin();
	clrowin();
	move(5, 0);
	if (disk < 0) {
		addstr("Please use 'o' to open a disk device file");
		pak();
		return;
	}

	if (!glabel) {
		addstr("You must read the label using the 'r' command, or\n");
		addstr("set the label using the 'l' command.");
		pak();
		return;
		}

	gp = &geometry;
	geometry.dkg_ncyl = label.dkl_ncyl;
	geometry.dkg_nhead = label.dkl_nhead;
	geometry.dkg_nsect = label.dkl_nsect;
	geometry.dkg_apc = label.dkl_apc;
	geometry.dkg_extra[XTCFF] = label.dkl_asciilabel[POSCFF];

	tsect = gp->dkg_nsect + gp->dkg_apc/gp->dkg_nhead;
	trksize = sizeof(TRAKID)*(tsect + (gp->dkg_extra[XTCFF] & CONFSSP));

	/* get some memory to hold the ID's */
	if ((tid = (TRAKID *)malloc(trksize)) == NULL) {
		addstr("Cannot get memory to hold ID's");
		pak();
		return;
	}

	if (!getcyltrack("desired", &cyl, &hd, &block, &inptype)) {
		return;
		}

	spc = gp->dkg_nsect * gp->dkg_nhead;

	/* prepare for the ioctl request */
	tid[0].cyl=cyl;
	tid[0].head=hd;

	if (ioctl(disk, DKIOCIDTRAK, &tid) < 0) {
	    addstr("DKIOCIDTRAK: io error");
	    pak();
	    return;
	    }

	switch (tid[0].flag) {
	case FLG_DEF:
		printw("\nTrack is mapped to %d/%d (block %d)",
			tid[0].cyl,tid[0].head,tid[0].cyl*spc+
			tid[0].head*gp->dkg_nsect);
		break;

	case FLG_ALT:
		printw("\nTrack is alternate for %d/%d (block %d)",
			tid[0].cyl,tid[0].head,tid[0].cyl*spc+
			tid[0].head*gp->dkg_nsect);
		break;

	case FLG_BAD:
		addstr("\nTrack is bad");
		break;

	case FLG_NRM:
	case FLG_DSC:
	case FLG_ASC:
	case FLG_BSC:
	    /* find the first non-slipped sector */
	    for (i=0; tid[i].sec == SEC_SLP; i++)
		;
	    skew = (gp->dkg_nsect - tid[i].sec) % gp->dkg_nsect;
	    printw("\nTrack has a sector skew of %d\n",skew);
	    for (i=0; i < tsect; i++) {
		switch (tid[i].flag) {
		case FLG_NRM:
		    if (tid[i].sec == 0xFF) {
			printw("Sector %d from index is slipped on track\n",i);
			}
		    break;

		case FLG_DSC:
		    printw("Sector %d from index (logical %d) is mapped to %d/%d/%d (block %d)\n",
			i,(cyl*gp->dkg_nhead+hd)*gp->dkg_nsect+tid[i].sec,
			tid[i].cyl,tid[i].head,tid[i].alt,
			(tid[i].cyl*gp->dkg_nhead + tid[i].head)*gp->dkg_nsect
			+tid[i].alt);
		    break;

		case FLG_ASC:
		    printw("Sector %d from index (logical %d) is alt for %d/%d/%d (block %d)\n",
			i,(cyl*gp->dkg_nhead+hd)*gp->dkg_nsect+tid[i].sec,
			tid[i].cyl,tid[i].head,tid[i].alt,
			((tid[i].cyl*gp->dkg_nhead + tid[i].head)*gp->dkg_nsect)
			+tid[i].alt);
		    break;

		case FLG_BSC:
		    printw("Sector %d from index is bad\n",i);
		    break;

		default:
		    printw("Sector %d from index has unknown flag %x\n",
			i,tid[i].flag);
		    break;
		    }
		} /* for each sector on the track */
	    break;
	
	default:
		printw("\nUnknown flag 0x%x on track!!",tid[0].flag);
		break;

		}

	pak();
} /* doxtrack */

/*******************************************************************************
*									       *
*	Subroutine:		doident                                        *
*									       *
*	Calling Sequence:	doident()                                      *
*									       *
*	Called by:		main                                           *
*									       *
*	Calling Parameters:	None                                           *
*				                                               *
*	Local Variables:  ident - structure returning controller info.         *
*				                                               *
*	Calls Subroutines: ioctl - to pass the ioctl request to the driver     *
*				                                               *
*	Public/Global Variables:    					       *
*               disk - file descriptor of disk device to query for the cache   *
*				                                               *
*	Description:							       *
*               This routine requests the Identification information from the  *
*		controller and displays that information.		       *
*		                                                               *
*******************************************************************************/

doident()
{
	RETID	ident;	/* structure returning identification info	*/

	clriwin();
	clrowin();
	move(5, 0);
	if (disk < 0) {
		addstr("Please use 'o' to open a disk device file");
		pak();
		return;
	}

	if (ioctl(disk, DKIOCGBID, &ident) < 0) {
		printw("Examine ID: %s", sys_errlist[errno]);
		pak();
		return;
	}

	addstr("Controller identification values are:\n\n");
	printw("Firmware Revision Level: %d\nEngineering Revision Level: %d\n",
		ident.fwrev, ident.engrev);
	printw("Firmware Date: %d/%02d/%02d",ident.year,ident.month,ident.day);


	pak();

} /* doident */

/*******************************************************************************
*									       *
*	Subroutine:		dolabel                                        *
*									       *
*	Calling Sequence:	dolabel()                                      *
*									       *
*	Called by:		main                                           *
*									       *
*	Calling Parameters:	None                                           *
*				                                               *
*	Local Variables:                                                       *
*                         wp - used to step through configuration info         *
*                         lp - pointer to label                                *
*                         type - type of label chosen                          *
*                         i - looping                                          *
*			  y - screen line for displaying list of labels	       *
*				                                               *
*	Calls Subroutines:        	                               	       *
*                          ioctl - to pass the label command to the driver     *
*                          askpartitions - allow user to modify partition      *
*                                          tables in the label                 *
*                          showpartitions - display partitions in the label    *
*				                                               *
*	Public/Global Variables:    					       *
*             fieldnames - names of the disk configuration fields in the label *
*				                                               *
*	Description:							       *
*              This subroutine displays the available standard labels          *
*              and allows the user to select one or edit whichever label       *
*              is currently selected.                                          *
*		                                                               *
*******************************************************************************/

#define LABLLIST	7		/* line to start double-column list */
dolabel()
{
	register int i;
	register unsigned short *wp;
	struct dk_label *lp;
	char c;			/* single-character response buffer */
	int type;
	int	y;

	clriwin();
	clrowin();

	y = LABLLIST;
	for (i = 0, lp = labels[ctype]; i < nlabels[ctype]; i++, lp++) {
		if (y >= LINES - 1) {
			cen_line(LINES-1,"Press ESC or RETURN to quit or any other key to see more");
			if ( ((c = inp_char(stdscr)) == CHR_ESC) ||
				(c == CHR_NL) )
			    break;
			move(y = LABLLIST, 0);
			clrtobot();
		}
		move(y++, 0);
		printw("%2d %s", i, lp->dkl_asciilabel);
	}
choose:
	move(5, 0);
	waddstr(stdscr, "Please enter your choice: ");
	input(stdscr, line);
	if (line[0] == '\0' || (line[0] | LCBIT) == 'q')
		return;
	if ((type = atoi(line)) < 0 || type >= nlabels[ctype]) {
		wmove(stdscr, 5, 0);
		waddstr(stdscr, "Bad choice, please try again.");
		goto choose;
	}
	lp = &label;
	clriwin();
	clrowin();
	if (type == 0) {
label:
		wmove(stdscr, 6, 0);
		waddstr(stdscr, "Please enter an ascii label for the disk (less than 127 characters):\n");
		wprintw(stdscr, "Press RETURN to keep:\n<%s>\n<", lp->dkl_asciilabel);
		wclrtoeol(stdscr);
		wmove(stdscr, 9, 1);
		input(stdscr, line);
		waddch(stdscr, '>');
		wrefresh(stdscr);
		if (line[0] != '\0') {
			if (strlen(line) > 126) {
				wmove(stdscr, 10, 0);
				waddstr(stdscr, "Label is too long, please try again.");
				goto label;
			} else
				strcpy(lp->dkl_asciilabel, line);
		}

		newconf(lp);

		clrowin();
		wmove(stdscr, 11, 0);
		wp = &lp->dkl_apc;
		for (i = 0; fieldnames[i] != NULL; i++) {
			wmove(stdscr, 11 + i, 0);
			wprintw(stdscr, "%s is %d\n", fieldnames[i], *wp++);
		}
		showpartitions(lp->dkl_map);
		askpartitions(lp->dkl_map);
	} else {
		*lp = labels[ctype][type];
		wmove(stdscr, 10, 0);
		wprintw(stdscr, "<%s>\n", lp->dkl_asciilabel);
		wp = &lp->dkl_apc;
		for (i = 0; fieldnames[i] != NULL; i++) {
			wmove(stdscr, 11 + i, 0);
			wprintw(stdscr, "%s is %d\n", fieldnames[i], *wp++);
		}
		if (ctype == RF3200) {
			lp->dkl_asciilabel[127] = CONFSSP | 
				((lp->dkl_ncyl > 1023) ? CONFEAD : 0);
			for (i=0; i < NCONFF; i++)
			    printw("%s is %s\n",cnffname[i],
			   (lp->dkl_asciilabel[127] & (1 << i)) ? "ON" : "OFF");
		} /* RF3200 uses configuration bits */
		showpartitions(lp->dkl_map);
	}
	lp->dkl_magic = DKL_MAGIC;
	complbck(lp);
	glabel = 1;
	pak();
	return;
}

/* complbck -- compute new label checksum
 * calculates new checksum for the label copy in memory
 */
complbck(lp)
struct dk_label *lp;	/* pointer to mem. label */
{
    register unsigned short *wp;  /* pointer to words of the label	*/
    register int i;		/* loop control variable		*/

    lp->dkl_cksum = 0;
    for (i = 0, wp = (unsigned short *)lp; i < sizeof(struct dk_label)/2-1; i++)
	lp->dkl_cksum ^= *wp++;

    } /* complbck */

/*******************************************************************************
*									       *
*	Subroutine:		newconf                                        *
*									       *
*	Calling Sequence:	newconf(lp)                                    *
*									       *
*	Called by:		dolabel                                        *
*									       *
*	Calling Parameters:	lp - pointer to label structure                *
*				                                               *
*	Local Variables:                                                       *
*                         wp - used to step through configuration info         *
*                         i - looping                                          *
*				                                               *
*	Calls Subroutines:        	                               	       *
*                          ioctl - to pass the label command to the driver     *
*                          askpartitions - allow user to modify partition      *
*                                          tables in the label                 *
*                          showpartitions - display partitions in the label    *
*				                                               *
*	Public/Global Variables:    					       *
*             fieldnames - names of the disk configuration fields in the label *
*				                                               *
*	Description:							       *
*              This subroutine allow the user to change the current            *
*	       setting of disk label.                                          *
*		                                                               *
*******************************************************************************/

newconf(lp)
struct dk_label *lp;
{
	int	item, i,j;
	register unsigned short *wp;

	clriwin();
	clrowin();
	while (1) {
		wmove(stdscr, 11, 0);
		wp = &lp->dkl_apc;
		for (i = 0; fieldnames[i] != NULL; i++, wp++) {
			wmove(stdscr, 11 + i, 0);
			wprintw(stdscr, "%2d - %s <%d>\n", i + 1, fieldnames[i], *wp);
		}

		if (ctype == RF3200) {
			printw("%2d - Short Sector Present (SSP) <%s>\n",i+1,
			   (lp->dkl_asciilabel[127] & CONFSSP) ? "ON" : "OFF");
			printw("%2d - Extended Addressing (EAD) <%s>\n",i+2,
			   (lp->dkl_asciilabel[127] & CONFEAD) ? "ON" : "OFF");
		} /* RF3200 uses configuration flag bits */

		wmove(stdscr, 5, 0);
		wclrtoeol(stdscr);
		wmove(stdscr, 5, 0);
		wprintw(stdscr, "Enter item number to change, <CR> when done: ");
		input(stdscr, line);
		wmove(stdscr, 6, 0);
		wclrtoeol(stdscr);
		if (line[0] == '\0')
			break;
		item = atoi(line);

		wmove(stdscr, 8, 0);
		wclrtoeol(stdscr);
		move(9, 0);
		clrtoeol();
		wmove(stdscr, 7, 0);
		switch (item) {
		case 1:
#ifdef NSPARES
			wprintw(stdscr, " 1.  # of spare sectors per cylinder:   %u\n", lp->dkl_apc);
#else
			wprintw(stdscr, " 1.  # of spare sectors per cylinder:   %u\n", lp->dkl_pad[0]);
#endif
			waddstr(stdscr, "new value or <CR> for no change: ");
			input(stdscr, line);
			if (line[0] == '\0')
				break;
#ifdef NSPARES
			lp->dkl_apc = atoi(line);
#else NSPARES
			lp->dkl_pad[0] = atoi(line);
#endif NSPARES
			break;
		case 2:
			wprintw(stdscr, " 2.  Size of gap1:   %u\n", lp->dkl_gap1);
			waddstr(stdscr, "new value or <CR> for no change: ");
			input(stdscr, line);
			if (line[0] == '\0')
				break;
			lp->dkl_gap1 = atoi(line);
			break;
		case 3:
			wprintw(stdscr, " 3.  Size of gap2:   %u\n", lp->dkl_gap2);
			waddstr(stdscr, "new value or <CR> for no change: ");
			input(stdscr, line);
			if (line[0] == '\0')
				break;
			lp->dkl_gap2 = atoi(line);
			break;
		case 4:
			wprintw(stdscr, " 4.  Interleave factor:   %u\n", lp->dkl_intrlv);
			waddstr(stdscr, "new value or <CR> for no change: ");
			input(stdscr, line);
			if (line[0] == '\0')
				break;
			lp->dkl_intrlv = atoi(line);
			break;
		case 5:
			wprintw(stdscr, " 5.  Number of cylinders:  %u\n", lp->dkl_ncyl);
			waddstr(stdscr, "new value or <CR> for no change: ");
			input(stdscr, line);
			if (line[0] == '\0')
				break;
			lp->dkl_ncyl = atoi(line);
			break;
		case 6:
			wprintw(stdscr, " 6.  Number of alt cylinders:  %u\n", lp->dkl_acyl);
			waddstr(stdscr, "new value or <CR> for no change: ");
			input(stdscr, line);
			if (line[0] == '\0')
				break;
			lp->dkl_acyl = atoi(line);
			break;
		case 7:
			wprintw(stdscr, " 7.  Number of heads:  %u\n", lp->dkl_nhead);
			waddstr(stdscr, "new value or <CR> for no change: ");
			input(stdscr, line);
			if (line[0] == '\0')
				break;
			lp->dkl_nhead = atoi(line);
			break;
		case 8:
			wprintw(stdscr, " 8.  # of sectors/track:    %u\n", lp->dkl_nsect);
			waddstr(stdscr, "new value or <CR> for no change: ");
			input(stdscr, line);

			if (line[0] == '\0')
				break;
			lp->dkl_nsect = atoi(line);
			break;
		case 9:
			wprintw(stdscr, " 9.  Label location:  %u\n", lp->dkl_bhead);
			waddstr(stdscr, "new value or <CR> for no change: ");
			input(stdscr, line);
			if (line[0] == '\0')
				break;
			lp->dkl_bhead = atoi(line);
			break;
		case 10:
			wprintw(stdscr, " 10.  Physical partition #:  %u\n", lp->dkl_ppart);
			waddstr(stdscr, "new value or <CR> for no change: ");
			input(stdscr, line);
			if (line[0] == '\0')
				break;
			lp->dkl_ppart = atoi(line);
			break;
		case 11:
			if (ctype == RF3200) {
				printw(" 11.  Short Sector Present: %s\n",
				    (lp->dkl_asciilabel[127] & CONFSSP) ? "ON" :
					"OFF");
				addstr("new value or <CR> for no change (0=OFF/1=ON): ");
				input(stdscr, line);
				if (line[0] == '\0')
					break;
				lp->dkl_asciilabel[127] &= ~CONFSSP;
				if (line[0] == '1')
				    lp->dkl_asciilabel[127] |= CONFSSP;
				break;
			} /* RF3200 uses configuration bits */
		case 12:
			if (ctype == RF3200) {
				printw(" 12.  Extended Addressing: %s\n",
				    (lp->dkl_asciilabel[127] & CONFEAD) ? "ON" :
					"OFF");
				addstr("new value or <CR> for no change (0=OFF/1=ON): ");
				input(stdscr, line);
				if (line[0] == '\0')
					break;
				lp->dkl_asciilabel[127] &= ~CONFEAD;
				if (line[0] == '1')
				    lp->dkl_asciilabel[127] |= CONFEAD;
				break;
			} /* RF3200 uses configuration bits */
		default:
			wmove(stdscr, 6, 0);
			wprintw(stdscr, "Invalid item %d.\n", item);
			break;
		}
	}
}

/*******************************************************************************
*									       *
*	Subroutine:		showlabel                                      *
*									       *
*	Calling Sequence:	showlabel()                                    *
*									       *
*	Called by:		main                                           *
*									       *
*	Calling Parameters:	None                                           *
*				                                               *
*	Local Variables:	                                               *
*                         lp - pointer to label                                *
*                         wp - used to step through configuration info         *
*                         cksum - used to calculate the checksum of the label  *
*                         i - looping                                          *
*				                                               *
*	Calls Subroutines:  showpartitions - display partition information     *
*                                            from the label                    *
*				                                               *
*	Public/Global Variables:     					       *
*                               label - the current label                      *
*				                                               *
*	Description:							       *
*              This subroutine displays the current label.                     *
*		                                                               *
*******************************************************************************/

showlabel()
{
	struct dk_label *lp = &label;
	register int i;
	register unsigned short *wp;
	unsigned short cksum;

	clriwin();
	clrowin();

	if (!glabel) {
		move(5, 0);
		addstr("You must read the label using the 'r' command, or\n");
		addstr("set the label using the 'l' command.");
		pak();
		return;
		}

	wmove(stdscr, 10, 0);
	wprintw(stdscr, "<%s>\n", lp->dkl_asciilabel);
	cksum = 0;
	for (i = 0, wp = (unsigned short *)lp; i < sizeof(*lp)/2-1; i++)
		cksum ^= *wp++;
	if (cksum != lp->dkl_cksum)
		wprintw(stdscr, "Checksum error in label, was 0x%04x, should be 0x%04x\n",
			cksum, lp->dkl_cksum);
	else
		waddstr(stdscr, "Checksum in label is ok.\n");
	if (lp->dkl_magic != DKL_MAGIC)
		wprintw(stdscr, "Wrong magic number in label: 0x%04x\n", lp->dkl_magic);
	else
		waddstr(stdscr, "Magic number in label is ok.\n");
	wp = &lp->dkl_apc;
	for (i = 0; fieldnames[i] != NULL; i++)
		wprintw(stdscr, "%-20s %d\n", fieldnames[i], *wp++);
	showpartitions(lp->dkl_map);
	pak();
}

/*******************************************************************************
*									       *
*	Subroutine:		domap                                          *
*									       *
*	Calling Sequence:	domap()                                        *
*									       *
*	Called by:		main                                           *
*									       *
*	Calling Parameters:	None                                           *
*				                                               *
*	Local Variables:	                                               *
*                         geom - geometry information for the current device   *
*                         map - structure to pass map command to the driver    *
*                         sblk, eblk - Starting and End block to map           *
*                         tblk - Block to map To                               *
*                         lblk - Last block on disk                            *
*				                                               *
*	Calls Subroutines:	                                   	       *
*               ioctl - to get configuration information from the driver       *
*                       and to pass the map command to the driver              *
*               showgeometry - to display bad geometry information from kernel *
*				                                               *
*	Public/Global Variables:    					       *
*                                 line - buffer for user input                 *
*				                                               *
*	Description:							       *
*               This subroutine asks the user for a range of blocks            *
*               to map and passes the information to the driver.               *
*		                                                               *
*******************************************************************************/

domap()
{
	struct dk_geom geom;
	struct dk_mapr map;
	daddr_t sblock, eblock, tblock, lblock;
	int	scyl, shd, ecyl, ehd, tcyl, thd, inptype;
	extern char *index();
	int tors;	/* track or sector */

	clriwin();
	clrowin();
	wmove(stdscr, 5, 0);
	if (disk == -1) {
		waddstr(stdscr, "You must open the disk first.");
		pak();
		return;
	}

	if (!glabel) {
		addstr("You must read the label using the 'r' command, or\n");
		addstr("set the label using the 'l' command.");
		pak();
		return;
		}

	if (ioctl(disk, DKIOCGGEOM, &geom) < 0) {
		wprintw(stdscr, "DKIOCGGEOM: %s", sys_errlist[errno]);
		pak();
		return;
	}

	if (geom.dkg_ncyl == 0 || geom.dkg_acyl == 0 || geom.dkg_nhead == 0
	 || geom.dkg_nsect == 0) {
		waddstr(stdscr, "Bad disk geometry in kernel:");
		showgeometry(&geom);
		pak();
		return;
	}

	lblock = (geom.dkg_ncyl + geom.dkg_acyl) * geom.dkg_nhead * geom.dkg_nsect - 1;

	/* 1.16g */
	do {
		wmove(stdscr, 11, 0); clrtoeol();
		addstr("Map by Track or Sector (T/S)? ");
		tors = inp_char(stdscr);
		if (isupper(tors)) tors = tolower(tors);
		} while (tors != 't' && tors != 's' && tors != 'q' &&
			 tors != CHR_ESC);
	if (tors == CHR_ESC || tors == 'q') return;

	if (tors == 't') {
		addstr("Track\nYou may specify any block on the track\n");
		if (!getcyltrack("bad", &scyl, &shd, &sblock, &inptype)) {
			pak();
			return;
			}
		sblock = (sblock/geom.dkg_nsect)*geom.dkg_nsect;
		eblock = sblock + geom.dkg_nsect - 1;
		wmove(stdscr, 17, 0);
		printw("Starting block: %d\n", sblock);
		addstr("Enter a destination of -1 to have the Rimfire ");
		addstr("select the alternate track.\n");
		if (!getcyltrack("destination", &tcyl, &thd, &tblock, &inptype)) {
			pak();
			return;
			}
		}

	else { /* by sector */
		addstr("Sector\n\n");
		if (!getbnchs("starting", &sblock)) {
		    pak();
		    return;
		    }
	    wmove(stdscr, 17, 0);
	    wprintw(stdscr, "Starting block: %d\n", sblock);
	    if (!getbnchs("ending", &eblock)) {
		    pak();
		    return;
		    }
	    wmove(stdscr, 18, 0);
	    wprintw(stdscr, "Ending block:   %d\n", eblock);
	    addstr("Enter a destination of -1 to have the Rimfire select ");
	    addstr("the alternate sector(s).\n");
	    if (!getbnchs("destination", &tblock)) {
		    pak();
		    return;
		}

	} /* by sector */
	clriwin();
	wmove(stdscr, 5, 0);
	if (tblock != -1) {
		printw("Enter 'y' to map blocks %d through %d to:\n", sblock, eblock);
		printw("                 blocks %d through %d\n", tblock, tblock + eblock - sblock);
	} else {
		printw("Enter 'y' to map blocks %d through %d\n", sblock, eblock);
		addstr("to a controller selected alternate\n");
	}
	if ((inp_char(stdscr) | LCBIT) != 'y') {
		return;
	}
	waddstr(stdscr, "Yes");
	map.dkm_fblk = sblock;
	map.dkm_tblk = tblock;
	map.dkm_nblk = eblock - sblock + 1;
	map.dkm_recover = 2;
	ioctl(disk, DKIOCMAP, &map);
	if (map.dkm_error != 0) {
		wmove(stdscr, 7, 0);
		wprintw(stdscr, "DKIOCMAP: %02x - %s",  map.dkm_error,
			rfmsg[map.dkm_error]); /* 1.10a */
	}
	/* 1.5a */
	else {
		move(7,0);
		printw("%s mapped.", (tors == 't') ? "Track" : "Sector(s)");
		}
	pak();
} /* domap */

/*******************************************************************************
*									       *
*	Subroutine:		dogeometry                                     *
*									       *
*	Calling Sequence:	dogeometry()                                   *
*									       *
*	Called by:		main                                           *
*									       *
*	Calling Parameters:	None                                           *
*				                                               *
*	Local Variables:	                                               *
*                       geometry - geometry information to pass to the driver  *
*				                                               *
*	Calls Subroutines:	                                   	       *
*                           ioctl - to get geometry info from the driver       *
*                           	    and to pass new information to it          *
*                           showgeometry - to display the geometry information *
*				                                               *
*	Public/Global Variables:    					       *
*                                 label - current label                        *
*                                 line - buffer for user input                 *
*				                                               *
*	Description:							       *
*            This subroutine asks whether to use geometry information as it    *
*            exists in the kernel or from the current label.  It then displays *
*            the geometry and asks whether to send it to the kernel.           *
*		                                                               *
*******************************************************************************/

dogeometry()
{
	struct dk_geom geometry;

	clriwin();
	clrowin();
	wmove(stdscr, 5, 0);
	if (disk == -1) {
		waddstr(stdscr, "You must open the disk before modifying the geometry.");
		pak();
		return;
	}
	waddstr(stdscr, "Enter 'k' to get disk geometry from the kernel,\n");
	waddstr(stdscr, "anything else will get it from the label: ");
	if ((inp_char(stdscr) | LCBIT) == 'k') {
		waddstr(stdscr, "Kernel");
		if (ioctl(disk, DKIOCGGEOM, &geometry) < 0) {
			wmove(stdscr, 7, 0);
			wprintw(stdscr, "Get geometry: %s", sys_errlist[errno]);
			pak();
			return;
		}
	} else {
		waddstr(stdscr, "Label");
		geometry.dkg_ncyl = label.dkl_ncyl;
		geometry.dkg_acyl = label.dkl_acyl;
		geometry.dkg_bcyl = 0;
		geometry.dkg_nhead = label.dkl_nhead;
		geometry.dkg_bhead = label.dkl_bhead;
		geometry.dkg_nsect = label.dkl_nsect;
		geometry.dkg_intrlv = label.dkl_intrlv;
		geometry.dkg_gap1 = label.dkl_gap1;
		geometry.dkg_gap2 = label.dkl_gap2;
#ifdef NSPARES
		geometry.dkg_apc = label.dkl_apc;
#else
		geometry.dkg_extra[0] = label.dkl_pad[0];
#endif
		/* configuration flags */
		geometry.dkg_extra[1] = label.dkl_asciilabel[127];
	}
	showgeometry(&geometry);
	wmove(stdscr, 7, 0);
	waddstr(stdscr, "Enter 'y' to write the new geometry to the kernel: ");
	if ((inp_char(stdscr) | LCBIT) != 'y') {
		waddstr(stdscr, "No");
		return;
	}
	waddstr(stdscr, "Write");
	if (ioctl(disk, DKIOCSGEOM, &geometry) < 0) {
		wmove(stdscr, 8, 0);
		wprintw(stdscr, "Set geometry: %s", sys_errlist[errno]);
	}
	pak();
} /* dogeometry */

/*******************************************************************************
*  1.9b									       *
*	Subroutine:		abdefect                                       *
*									       *
*	Calling Sequence:	abdefect()                                     *
*									       *
*	Called by:		main                                           *
*									       *
*	Calling Parameters:	None                                           *
*				                                               *
*	Local Variables:	                                               *
*				                                               *
*	Calls Subroutines:	                                   	       *
*	   mkdefname							       *
*				                                               *
*	Public/Global Variables:    					       *
*				                                               *
*	Description:							       *
*	  Enter the defect list into a file -- useful for people who have      *
*         already formatted their drives or who have otherwise lost the        *
*         machine-readable version of the list and which to enter it from the  *
*         printed information provided by the manufacturer.		       *
*		                                                               *
*******************************************************************************/

abdefect()
{
	char dflname[128];	/* defect list file name		*/
	int cyl;		/* cylinder number			*/
	int head;		/* head number				*/
	int pos;		/* byte position			*/
	int len;		/* error length (in bits)		*/
	FILE *dflf;		/* point to file structure for defect list*/
	char c;			/* short input buffer			*/

	clrowin();
	clriwin();

	/* make sure disk device is opened */
	if (disk == -1) {
		wmove(stdscr,4,0);
		waddstr(stdscr, "Please use 'o' to open a disk device file.");
		pak();
		return;
	}
		
	/* build default name for defect list */
	mkdefname(diskname,dflname);

#ifndef STANDALONE
	move(5,0);
	/* attempt to open a file with the default name */
	while ((dflf = fopen(dflname,"r+")) != NULL) {
		
	    printw("File \"%s\" already exists.\n",dflname);
	    addstr("Enter a file name for the defect list, or press RETURN to quit: \n");

	    dflname[0] = '\0';
	    input(stdscr,dflname);
	    if (!dflname[0] || (dflname[0] == CHR_ESC)) {
		/* give up */
		return;
		}
	    } /* while file already exists */

	move (6,0);
	printw("About to create \"%s\" file for defect list.  OK? (y/n) ",
	    dflname);
	if ((inp_char(stdscr) | LCBIT) != 'y') {
	    return;
	}
#endif

	if ((dflf = fopen(dflname,"w+")) == NULL) {
	    printw("\nCannot create %s: %s",dflname,sys_errlist[errno]);
	    pak();
	    return;
	    }
	
	clriwin();

	move(4,0);
	addstr("Enter all numbers as decimal\n");
	addstr("If no error length is given on your list, specify 16\n");
	addstr("If a track is to be marked bad, give both byte offset and length as 0\n\n");

	for (;;)  {
	    move(8,0);
	    addstr("Cylinder number of error (-1 to quit): ");  clrtoeol();
	    line[0] = '\0'; input(stdscr,line);
	    sscanf(line," %d",&cyl);
	    if (cyl < 0) break;
	    move(9,0);
	    addstr("Head number: ");  clrtoeol();
	    line[0] = '\0'; input(stdscr,line);
	    sscanf(line," %d",&head);
	    move(10,0);
	    addstr("Byte offset: ");  clrtoeol();
	    line[0] = '\0'; input(stdscr,line);
	    sscanf(line," %d",&pos);
	    move(11,0);
	    addstr("Length of the defect: ");  clrtoeol();
	    line[0] = '\0'; input(stdscr,line);
	    sscanf(line," %d",&len);

	    fprintf(dflf,"%4d %2d: %d,%d\n", cyl, head, pos,len);


	    }

	fclose(dflf);


	}  /* abdefect */

/*******************************************************************************
*									       *
*	Subroutine:		readdef                                        *
*									       *
*	Calling Sequence:	readdef()                                     *
*									       *
*	Called by:		main                                           *
*									       *
*	Calling Parameters:	None                                           *
*				                                               *
*	Local Variables:	                                               *
*               	cyl, head - indexes for cylinder and head, respectively*
*               	defect - defect table for communicating w/driver       *
*               	dflname - name of defect list file                     *
*               	dflf - defect list file structure pointer              *
*               	geometry - disk geometry from kernel                   *
*				                                               *
*	Calls Subroutines:	                                   	       *
*                           ioctl - to get geometry info from the driver       *
*                           	    and to pass new information to it          *
*				                                               *
*	Public/Global Variables:    					       *
*                           NONE                                               *
*				                                               *
*	Description:							       *
*	     This routine reads an entire disk drive to obtain the mftrs's     *
*	     defect list, which is written to a file.  The file contains one   *
*	     ASCII record for each track containing a defect.  Tracks which    *
*	     are defective give 0 for all defect positions and lengths.        *
*		                                                               *
*******************************************************************************/

readdef()
{
    int cyl,head;		/* indexes for cylinders and heads	*/
    struct dk_rdb defect,*dp;	/* defect structure for system to return */
    FILE *dflf;			/* defect list file			*/
    struct dk_geom geometry;	/* geometry of drive (must be preset)	*/
    char dflname[256];		/* defect list file name		*/
    register int i;		/* counting index for ESDI errors	*/
    static int defmax[] = { 256, 512, 1024, 2048 };
				/* Maximum errors noted for ESDI list	*/
    int offset;			/* Position of an ESDI error -- also used
				   for a file descriptor to check existence */
				

    clrowin();
    clriwin();

    move(5,0);

    /* save the address of the block to pass through to ioctl
     * (SUN work-around)
     */
    dp = &defect;

    if (disk < 0) {
	addstr("Must open device file first!");
	pak();
	return;
	}

    if (!glabel) {
	addstr("Must specify label first!");
	pak();
	return;
	}

    move(7, 0);
    if (ioctl(disk, DKIOCGGEOM, &geometry) < 0) {
	wprintw(stdscr, "Get geometry: %s\n", sys_errlist[errno]);
	pak();
	return;
	}

    if (geometry.dkg_ncyl == 0 || geometry.dkg_acyl == 0 ||
	geometry.dkg_nhead == 0 || geometry.dkg_nsect == 0) {
	addstr("Bad disk geometry in kernel:  Set the geometry and try again.");
	pak();
	return;
	}

    mkdefname(diskname,dflname);

#ifndef STANDALONE
    /* 1.12a */
    /* until we get an unused file name */
    for (;;) {

	/* until the user likes the file name */
	for (;;) {
	    move(7,0); clrtobot();
	    printw("About to create \"%s\" file for defect list.  OK? (y/n) ",
		dflname);
	    if ((inp_char(stdscr) | LCBIT) == 'y')
		break;

	    move(7,0); clrtobot();
	    addstr("Enter a file name for the defect list, or press RETURN to quit:\n");

	    dflname[0] = '\0';
	    input(stdscr,dflname);
	    if (!dflname[0] || (dflname[0] == CHR_ESC)) {
		return;
		}
	    } /* until the user likes the file name */

	if ((offset = open(dflname,0)) < 0)
	    break;

	close(offset);
	move(7,0); clrtobot();
	printw("\"%s\" exists.  Press 'q' to quit or any other key to try again:",
	    dflname);
	if ((inp_char(stdscr) | LCBIT) == 'q')
	    return;

	move(7,0); clrtobot();
	addstr("Enter a file name for the defect list, or press RETURN to quit:\n");

	dflname[0] = '\0';
	input(stdscr,dflname);
	if (!dflname[0] || (dflname[0] == CHR_ESC)) {
	    return;
	    }
	} /* until we get an unused file name */
#endif STANDALONE
	
    move(7,0); clrtobot();
    if ((dflf = fopen(dflname,"w")) == NULL) {
	printw("Cannot create %s: %s",dflname,sys_errlist[errno]);
	pak();
	return;
	}

    /* Finally ready to get defect info from drive */

    if (ctype == RF3200) {

	/* Ask for the starting cylinder in case they formatted part
	   of the drive -- for completeness, should ask for starting
	   head, also, but for ease of use and implementation, we
	   will settle for the cylinder number for now */

	do {	/* 1.12b */
	    move(7,0); clrtobot();
	    addstr("Enter starting cylinder (0 for all): ");

	    line[0] = '\0';
	    cyl = 0;
	    input(stdscr,line);
	    sscanf(line," %d",&cyl);
	    } while (cyl < 0 || cyl >= geometry.dkg_ncyl + geometry.dkg_acyl);

	/* for each cylinder */
	for (; cyl < geometry.dkg_ncyl + geometry.dkg_acyl; cyl++ ) {

	    /* show cylinder number to assure that we are running */
	    move(9,0); printw("Cylinder %4d",cyl); refresh();

	    /* for each head */
	    for (head = 0; head < geometry.dkg_nhead; head ++ ) {
		defect.db.smd.cylinder = cyl;
		defect.db.smd.head = head;
		ioctl(disk, DKIOCRDDAT,&dp);
		/* 1.9d */
		if (defect.dkd_error){
		    fclose(dflf);
		    printw(", head %d\nDKIOCRDDAT: %02x - %s",head,
			defect.dkd_error,rfmsg[defect.dkd_error]); /* 1.10a */
		    pak();
		    return;
		    }
		else{
		    /* assume that if the first defect length is 0, the
		       track is clean */
		    if ( defect.db.smd.len1 | defect.db.smd.pos1 ) {

			/* a defective track will be marked as having zero-
			   length defects, since error-free tracks do not
			   appear in the disk file at all. */
			if ( (cyl == defect.db.smd.cylinder) && (defect.db.smd.xf0 == 0xf0)) {
			    /* defects in track 1.6 */
			    fprintf(dflf,"%4d %2d: %d,%d\n", cyl, head,
				    defect.db.smd.pos1,defect.db.smd.len1);
			    if (defect.db.smd.len2)
				fprintf(dflf,"%4d %2d: %d,%d\n", cyl, head,
				    defect.db.smd.pos2,defect.db.smd.len2);
			    if (defect.db.smd.len3)
				fprintf(dflf,"%4d %2d: %d,%d\n", cyl, head,
				    defect.db.smd.pos3,defect.db.smd.len3);
			    if (defect.db.smd.len4)
				fprintf(dflf,"%4d %2d: %d,%d\n", cyl, head,
				    defect.db.smd.pos4,defect.db.smd.len4);
			    }
			else {
			    /* defective track */
			    fprintf(dflf,"%4d %2d: 0,0\n", cyl, head);
			    }
			} /* not a clean track */

		    } /* no IOCTL error */

		} /* for each head */

	    } /* for each cylinder */

	} /* RF3200 */

    else { /* RF3400 1.6 */
	/* According the the ESDI spec., the defect information is stored
	   in two or three places on each surface of the drive.  The best
	   place to look is a special cylinder designated 0xfff, but this
	   is a new addition to the spec., and there are many drives in
	   the field that do not yet support this.  In most drives, it
	   will likely be a phantom cylinder outside of cylinder 0.  The
	   other two places are the last cylinder of the drive and the
	   cylinder 8 in from the last cylinder.  These are less desireable
	   to use because they are increasingly more likely to be formatted.
	   The most desirable method would be to try them in the order just
	   mentioned, checking the error return on cylinder 0xfff to accomodate
	   those drives which support that, and then silently trying the
	   last cylinder if the 0xfff access fails.  Since there are no drives
	   currently in the field which support this feature, we will wait
	   to include the first try at cylinder 0xfff until a slightly later
	   date. */

	/* for each head */
	for (head = 0; head < geometry.dkg_nhead; head ++ ) {

	    /* show head number to assure that we are running */
	    move(9,0); printw("Head %2d",head); refresh();

	    /* try last cylinder */
	    cyl = geometry.dkg_ncyl + geometry.dkg_acyl - 1;
	    defect.db.esdi.header.cylhigh = 0;
	    defect.db.esdi.header.cyllow = 0;
	    defect.db.esdi.header.head = head;
	    ioctl(disk, DKIOCRDDAT,&dp);
	    if (defect.dkd_error){
		fclose(dflf);
		printw("\nDKIOCRDDAT: %02x - %s",defect.dkd_error,
			rfmsg[defect.dkd_error]); /* 1.10a */
		pak();
		return;
		}

	    if ((defect.db.esdi.header.head != head) || 
		(defect.db.esdi.data.head != head) ||
		(defect.db.esdi.data.month > 12)) {

		/* try next location -- we didn't get good stuff */
		defect.db.esdi.header.cylhigh = 0;
		defect.db.esdi.header.cyllow = 1;
		defect.db.esdi.header.head = head;
		ioctl(disk, DKIOCRDDAT,&dp);
	        if (defect.dkd_error){
		    fclose(dflf);
		    printw("\nDKIOCRDDAT: %02x - %s",defect.dkd_error,
			rfmsg[defect.dkd_error]); /* 1.10a */
		    pak();
		    return;
		    }
		if ((defect.db.esdi.header.head != head) || 
		    (defect.db.esdi.data.head != head) ||
		    (defect.db.esdi.data.month > 12)) {
		    fclose(dflf);
		    printw("\nCannot find defect list on drive");
		    pak();
		    return;
		    }
		} /* try next location */

	    /* If we got this far, we must have gotten a plausible list */
	    for (i=0; i < 5*defmax[defect.db.esdi.header.flag & 3] &&
			defect.db.esdi.data.dl[i] != 0xff; i += 5 ) {
		cyl = (defect.db.esdi.data.dl[i] << 8) +
			defect.db.esdi.data.dl[i+1];
		offset = (defect.db.esdi.data.dl[i+2] << 8) +
			defect.db.esdi.data.dl[i+3];
		/* defects in track */
		fprintf(dflf,"%4d %2d: %d,%d\n",
		    cyl, head, offset, defect.db.esdi.data.dl[i+4]);
		} /* for each defect */

	    } /* for each head */

	} /* RF3400 */


    fclose(dflf);

    pak();

    } /* readdef */

/*******************************************************************************
*									       *
*	Subroutine:		doslip                                         *
*									       *
*	Calling Sequence:	doslip()                                       *
*									       *
*	Called by:		main                                           *
*									       *
*	Calling Parameters:	None                                           *
*				                                               *
*	Local Variables:	                                               *
*                       slip - sector slip information to pass to driver       *
*				                                               *
*	Calls Subroutines:	                                   	       *
*                           ioctl - to get geometry info from the driver       *
*                           	    and to pass new information to it          *
*			    Map sectors/track				       *
*				   - for mapping entire tracks.                *
*				                                               *
*	Public/Global Variables:    					       *
*                                 line - buffer for user input                 *
*				                                               *
*	Description:							       *
*            This subroutine will take the physical head, cylinder and byte    *
*	     count for a defect on the disk, then convert that number into     *
*	     a logical sector for that drive.  It will do this by reading      *
*	     a track of ID's then use the physical sector for the index        *
*	     into the logical sectors. Then issuing a slip sector for the      *
*	     correct sector.						       *
*		                                                               *
*******************************************************************************/

doslip()
{
	struct dk_slip slip;
	long bn;		/* block number to slip			*/

	clrowin();
	clriwin();
	wmove(stdscr, 5, 0);
	/* make sure disk device is opened */
	if (disk == -1) {
		waddstr(stdscr, "Please use 'o' to open a disk device file.");
		pak();
		return;
	}

	do {
		wmove(stdscr, 5, 0);
		wclrtoeol(stdscr);
		waddstr(stdscr,
	     "Enter the block number or cylinder/head/sector to be slipped: ");
		if (rdblkno(&bn,NULL)) /* 1.16i */
			return;
		} while (bn < 0);
	
	slip.dks_fblk = bn;

	wmove(stdscr, 6, 0);
  	waddstr(stdscr, "Do you want to recover previous data?");
	if ((inp_char(stdscr) | LCBIT) == 'y') {
		waddstr(stdscr, "Yes");
		slip.dks_recover = 1;
	} else {
		waddstr(stdscr, "No");
		slip.dks_recover = 0;
	}
	slip.dks_error = 0;
	wmove(stdscr, 7, 0);

	/* pass pointer to the driver for the dirty work */
	if (ioctl(disk, DKIOCSLP, &slip) < 0) {
		wprintw(stdscr, "DKIOCSLP: %s", sys_errlist[errno]);
	}

	/* check for error returned by the driver */
	else if (slip.dks_error != 0) {
		waddstr(stdscr, "Out of spare sectors, use map sector.");
        }
	/* 1.5a */
	else {
		addstr("Sector slipped.");
		}

	pak();
	return;
} /* doslip */

/*******************************************************************************
 *									       *
 *	Subroutine:		showgeomrtry                                   *
 *									       *
 *	Calling Sequence:	showgeometry(gp)                               *
 *									       *
 *	Called by:		dogeometry, domap                              *
 *									       *
 *	Calling Parameters:   gp - pointer to geometry information to display  *
 *				                                               *
 *	Local Variables:	None                                           *
 *				                                               *
 *	Calls Subroutines:	None                               	       *
 *				                                               *
 *	Public/Global Variables:None					       *
 *				                                               *
 *	Description:							       *
 *                   This subroutine displays the disk geometry information    *
 *                   passed to it.                                             *
 *		                                                               *
 ******************************************************************************/

showgeometry(gp)
register struct dk_geom *gp;
{
	wmove(stdscr, 11, 0);
	wprintw(stdscr, "# of data cylinders is %d\n", gp->dkg_ncyl);
	wmove(stdscr, 12, 0);
	wprintw(stdscr, "# of alternate cylinders is %d\n", gp->dkg_acyl);
	wmove(stdscr, 13, 0);
	wprintw(stdscr, "cyl offset (for fixed head area) is %d\n", gp->dkg_bcyl);
	wmove(stdscr, 14, 0);
	wprintw(stdscr, "# of heads is %d\n", gp->dkg_nhead);
	wmove(stdscr, 15, 0);
	wprintw(stdscr, "head offset (for Larks, etc.) is %d\n", gp->dkg_bhead);
	wmove(stdscr, 16, 0);
	wprintw(stdscr, "# of sectors per track is %d\n", gp->dkg_nsect);
	wmove(stdscr, 17, 0);
	wprintw(stdscr, "# of spare sectors per cylinder is %d\n", 
#ifdef NSPARES
		gp->dkg_apc);
#else
		gp->dkg_extra[0]);
#endif
	wmove(stdscr, 18, 0);
	wprintw(stdscr, "interleave factor is %d\n", gp->dkg_intrlv);
	wmove(stdscr, 19, 0);
	wprintw(stdscr, "gap 1 size is %d\n", gp->dkg_gap1);
	wmove(stdscr, 20, 0);
	wprintw(stdscr, "gap 2 size is %d\n", gp->dkg_gap2);
} /* showgeometry */

/*******************************************************************************
*									       *
*	Subroutine:		dopartitions                                   *
*									       *
*	Calling Sequence:	dopartitions()                                 *
*									       *
*	Called by:		main                                           *
*									       *
*	Calling Parameters:	None                                           *
*				                                               *
*	Local Variables:	None                                           *
*				                                               *
*	Calls Subroutines:	                                   	       *
*                   getpartitions - retrieve partition information from driver *
*                   showpartitions - display partition information             *
*				                                               *
*	Public/Global Variables:    					       *
*                                label - current label                         *
*                                line - buffer for user input                  *
*				                                               *
*	Description:							       *
*                This subroutine asks whether to retrieve partition            *
*                information from the driver or from the current label.        *
*                It then displays the partition information and allows         *
*                modification.  It allows allows the partition information     *
*                to be written back to the driver.                             *
*		                                                               *
*******************************************************************************/

dopartitions()
{
	clriwin();
	clrowin();
	move(5, 0);
 	addstr("Enter 'k' to get the partition table from the kernel,\n");
 	addstr("anything else will get it from the label: ");
 	if ((inp_char(stdscr) | LCBIT) == 'k') {
 		addstr("Kernel");
 		if (getpartitions(label.dkl_map) < 0) {
 			pak();
 			return;
 		}
 	} else
 		addstr("Label");
 	showpartitions(label.dkl_map);
 	move(7, 0);
 	addstr("Enter 'y' if you wish to modify these partitions: ");
 	if ((inp_char(stdscr) | LCBIT) == 'y') {
 		addstr("Yes");
 		askpartitions(label.dkl_map);
 	} else
 		addstr("No");
 	move(8, 0);
 	addstr("Enter 'y' to write the new partition table to the kernel: ");
 	if ((inp_char(stdscr) | LCBIT) != 'y') {
 		waddstr(stdscr, "No");
 		return;
 	}
 	waddstr(stdscr, "Yes");
 	setpartitions(label.dkl_map);
 	complbck(&label);
 	pak();
 	return;
 }

/*******************************************************************************
* 1.9c									       *
*	Subroutine:		enqdc		                               *
*									       *
*	Calling Sequence:	enqdc()		                               *
*									       *
*	Called by:		main                                           *
*									       *
*	Calling Parameters:	None                                           *
*				                                               *
*	Local Variables:	None                                           *
*				                                               *
*	Calls Subroutines:	                                   	       *
*				                                               *
*	Public/Global Variables:    					       *
*                                label - current label                         *
*                                line - buffer for user input                  *
*				                                               *
*	Description:							       *
*		This routine does a command 0x13 (Return ESDI configuration    *
*		for the 3400 or Interrogate Disk for the 3200) and displays    *
*		such information as was gotten therefrom.		       *
*		                                                               *
*******************************************************************************/

enqdc()
{
	clrowin();
	clriwin();

	move (5,0);

	if (ctype == RF3200) {
		interrog();
		}
	else {
		esdiconf();
		}
	
	pak();
	} /* enqdc */

interrog()
{
	SMDCHRS charp;

	if (ioctl(disk, DKIOCSMDCHR, &charp) < 0) {
		move(4, 0);
		printw("DKIOCSMDCHR: %s", sys_errlist[errno]);
		return;
 		}

	if (charp.dsctrk == 0) {
		printw("Disk is unformatted\nPhysical sectors/track\t%4d\n",
			charp.psctrk);
		if (charp.cyld) {
			printw("Cylinders\t\t%4d\n",charp.cyld);
			}
		else
			addstr("Cylinders\t\t1024 (possibly)\n");
		if (charp.nhead) {
			printw("Heads\t\t\t%4d",charp.nhead);
			}
		} /* unformatted disk */
	else {
		/* disk is formatted -- much better information returned */

		/* display sector info */
		printw("Disk is formatted\nPhysical sectors/track\t%4d",
			charp.psctrk);
		move(6,COLS/2);
		printw("Data sectors/track\t%4d\n",charp.dsctrk);
		if (charp.flags & CONFSSP)
			addstr("Short Sector Present (SSP)\n");
		else
			addstr("No Short Sector (/SSP)\n");

		/* display cylinder information or best guess */
		if (charp.cyld) {
			printw("Cylinders\t\t%4d\n",charp.cyld);
			}
		else
			addstr("Cylinders\t\t1024 (possibly)\n");
		
		/* Finally, number of heads */
		printw("Heads\t\t\t%4d",charp.nhead);

		} /* formatted disk */
	} /* interrog */

esdiconf()
{
	ESDIPARAMS parmp;
    
	if (ioctl(disk, DKIOCRESDI, &parmp) < 0) {
		move(4, 0);
		printw("DKIOCRESDI: %s", sys_errlist[errno]);
		return;
 		}

	printw("Physical sectors/track\t\t%4d", parmp.sectrk);
	move(5,COLS/2);
	printw("Physical bytes/sector\t\t%4d\n",parmp.bytesec);

	printw("Total cylinders (Fixed)\t\t%4d", parmp.fixcyl);
	move(6,COLS/2);
	printw("Total cylinders (Removable)\t%4d\n", parmp.remcyl);

	printw("Heads (Fixed)\t\t\t%4d", parmp.fixhead);
	move(7,COLS/2);
	printw("Heads (Removable)\t\t%4d\n", parmp.remhead);

	printw("Physical bytes/track\t%12d", parmp.minbytrk);

	} /* esdiconf */


/*******************************************************************************
 *									       *
 *	Subroutine:		showpartitions                                 *
 *									       *
 *	Calling Sequence:	showpartitions(map)                            *
 *									       *
 *	Called by:		main                                           *
 *									       *
 *	Calling Parameters:                                                    *
 *                       map - pointer to the partition mapping information    *
 *				                                               *
 *	Local Variables:	                                               *
 *                                i - looping                                  *
 *				                                               *
 *	Calls Subroutines:	None                               	       *
 *				                                               *
 *	Public/Global Variables:None					       *
 *				                                               *
 *	Description:							       *
 *                This subroutine displays the partition mapping information   *
 *                passed to it.                                                *
 *		                                                               *
 ******************************************************************************/
 
showpartitions(map)
register struct dk_map *map;
{
 	register int i;
 
 	for (i = LINES < 24 ? LINES : 23; i > 14; i--) {
 		wmove(stdscr, i, COLS/2);
 		wclrtoeol(stdscr);
 	}
 	wmove(stdscr, 15, COLS/2);
 	wprintw(stdscr, "%2s %4s %8s", "", "cyl", "nblks");
 	for (i = 0; i < 8; i++) {
 		wmove(stdscr, 12 + i, COLS/2);
 		wprintw(stdscr, "%c: %4ld %8ld", 'a' + i, map->dkl_cylno, map->dkl_nblk);
 		map++;
 	}
 }
 
/*******************************************************************************
 *									       *
 *	Subroutine:		askpartitions                                  *
 *									       *
 *	Calling Sequence:	askpartitions(map)                             *
 *									       *
 *	Called by:		dopartitions, dolabel                          *
 *									       *
 *	Calling Parameters:	                                               *
 *                      map - pointer to partition mapping info to modify      *
 *				                                               *
 *	Local Variables:	                                               *
 *                         i - looping                                         *
 *                         cp - point to input buffer                          *
 *                         cyl, nblk - starting cylinder and number of         *
 *                                     blocks read from user                   *
 *				                                               *
 *	Calls Subroutines:	None                               	       *
 *				                                               *
 *	Public/Global Variables:    					       *
 *                          line - user input buffer                           *
 *				                                               *
 *	Description:							       *
 *              This subroutine allows the user to edit the partition          *
 *              table passed to it.  It steps through each entry,              *
 *              displays it, and accepts replacement entries if specified.     *
 *		                                                               *
 ******************************************************************************/
 
askpartitions(map)
register struct dk_map *map;
{
 	register int i;
 	register char *cp;
 	int cyl, nblk;
 	extern char *index();
 	struct dk_map *savemap;
 	char partition[5];
 	int count;
 	int first;
 
 	savemap = map;
 	clriwin();
 again:
 	first = TRUE;
 	do {
 		if (!first) {
 			wmove(stdscr, 10, 0);
 			waddstr(stdscr, "Illegal partition, try again - ");
 		}
 		first = FALSE;
 		wmove(stdscr, 10, 32);
 		waddstr(stdscr, "Partition to change, <CR> when done: ");
 		line[0] = inp_char(stdscr);
 		if (line[0] == '\n')
 			return;
 		if (isalpha(line[0])) {
 			if (isupper(line[0]))
 				line[0] = tolower(line[0]);
 		}
 	} while (line[0] < 'a' || line[0] > 'h');
 
 	partition[0] = line[0];
 	partition[1] = '\0';
 	map = savemap;
 	count = (int)line[0] - (int)'a';
 	for (i=0; count > 0; count--)
 		map++;
 retry:
 	wmove(stdscr, 10, 0);
 	wclrtoeol(stdscr);
 	wprintw(stdscr, "%1s: %4ld %8ld >", partition, map->dkl_cylno, map->dkl_nblk);
 
 	input(stdscr, line);
 	if ((line[0] | LCBIT)  == 'q')
 		goto again;
 	if ((cp = index(line, ',')) == NULL
 	&& (cp = index(line, ' ')) == NULL
 	&& (cp = index(line, '\t')) == NULL) {
 		wmove(stdscr, 9, 0);
 		waddstr(stdscr, "Need 2 numbers separated by a blank, please try again.\n");
 		goto retry;
 	}
 	if ((cyl = atoi(line)) < 0 || !isdigit(line[0])) {
 		wmove(stdscr, 9, 0);
 		*cp = '\0';
 		wprintw(stdscr, "%s is not a valid cylinder number, use 0-?\n", line);
 		goto retry;
 	}
 	while (*cp == ',' || isspace(*cp))
 		cp++;
 	if ((nblk = atoi(cp)) < 0 || !isdigit(*cp)) {
 		wmove(stdscr, 9, 0);
 		wprintw(stdscr, "%s is not a valid number of blocks, use 0-?\n", cp);
 		goto retry;
 	}
 	map->dkl_cylno = cyl;
 	map->dkl_nblk = nblk;
 	showpartitions(savemap);
 	goto again;
 	/*NOTREACHED*/
} /* askpartitions */

/*******************************************************************************
 *									       *
 *	Subroutine:		setpartitions                                  *
 *									       *
 *	Calling Sequence:	setpartitions()                                *
 *									       *
 *	Called by:		dopartitions                                   *
 *									       *
 *	Calling Parameters:	None                                           *
 *				                                               *
 *	Local Variables:	                                               *
 *                     i - looping                                             *
 *                     file - to open the device file for each partition       *
 *                     cp - pointer to partition character of device file name *
 *				                                               *
 *	Calls Subroutines:	                                   	       *
 *                      ioctl - to set partition information for each partition*
 *				                                               *
 *	Public/Global Variables:    					       *
 *                      diskname - device file name of device currently in use *
 *				                                               *
 *	Description:							       *
 *               This subroutine passes the partition information from         *
 *               "map" to the device driver.                                   *
 *                                                                             *
 ******************************************************************************/
 
setpartitions(map)
register struct dk_map *map;
{
 	register int i;
 	register int file;
 	register char *cp;
 	char savechar;
 	int savegflag;		/* save the contents of the "good label" flag */
 
 	cp = &diskname[strlen(diskname)-1];
 	savechar = *cp;
 	for (i = 0; i < NDKMAP; i++) {
 		*cp = 'a' + i;
 		if (map[i].dkl_nblk == 0)
 			continue;
 		if ((file = open(diskname, 2)) < 0) {
 			wprintw(stdscr, "\nopen: %s %s\n", diskname,
				sys_errlist[errno]); /* 1.10b */
 			continue;
 		}
 		if (ioctl(file, DKIOCSPART, &map[i]) < 0) {
 			wmove(stdscr, 4, 0);
 			wprintw(stdscr, "\nDKIOCSPART: %s", sys_errlist[errno]);
 		}
 		close(file);
 	}
 	*cp = savechar;	/* back to the first partition */
 
 	/* preserve copy of disk name and label status flag (nulled at close) */
 	strcpy(line,diskname);
 	savegflag = glabel;
 
 	/* So as to refresh the possible size change. */
 	if (opendevice(line) > 0)
 		glabel = savegflag;
 
 	return;
 }
 
/*******************************************************************************
 *									       *
 *	Subroutine:		getpartitions                                  *
 *									       *
 *	Calling Sequence:	getpartitions()                                *
 *									       *
 *	Called by:		dopartitions                                   *
 *									       *
 *	Calling Parameters:	None                                           *
 *				                                               *
 *	Local Variables:	                                               *
 *                      i - looping                                            *
 *                      file - to open the device file for each partition      *
 *                      cp - pointer to partition character of device file name*
 *				                                               *
 *	Calls Subroutines:	                                   	       *
 *                     ioctl - to set partition information for each partition *
 *				                                               *
 *	Public/Global Variables:    					       *
 *                      diskname - device file name of device currently in use *
 *				                                               *
 *	Description:							       *
 *                This subroutine gets the partition information from          *
 *                the device driver and places it in the array of              *
 *                structures pointed to by "map".                              *
 *		                                                               *
 ******************************************************************************/
 
getpartitions(map)
register struct dk_map *map;
 {
 	register int i;
 	register int file;
 	register char *cp;
 	char savepart;
 	int good = -1;
 
 	cp = &diskname[strlen(diskname)-1];
 	savepart = *cp;
 	for (i = 0; i < NDKMAP; i++) {
 		*cp = 'a' + i;
 		if ((file = open(diskname, 2)) < 0) {
 			map[i].dkl_cylno =  map[i].dkl_nblk = 0;
 			continue;
 		}
 		if (ioctl(file, DKIOCGPART, &map[i]) < 0) {
 			wmove(stdscr, 4, 0);
 			wprintw(stdscr, "\nDKIOCGPART: %s\n", sys_errlist[errno]);
 			*cp = savepart;
 			return(-1);
 		}
 		good = 0;
 		close(file);
 	}
 	*cp = savepart;
 	return(good);
 }
 
/*******************************************************************************
 *									       *
 *	Subroutine:		showstatistics                                 *
 *									       *
 *	Calling Sequence:	showstatistics()                               *
 *									       *
 *	Called by:		main                                           *
 *									       *
 *	Calling Parameters:	None                                           *
 *				                                               *
 *	Local Variables:	statb - the statistics data                    *
 *				                                               *
 *	Calls Subroutines:	                                   	       *
 *                      ioctl - to pass the statistics command to the driver   *
 *				                                               *
 *	Public/Global Variables:None					       *
 *                  disk - file descriptor of disk to get the statistics from  *
 *				                                               *
 *	Description:							       *
 *                This routine prints out the data returned by the             *
 *                statistics command                                           *
 *		                                                               *
 ******************************************************************************/
 
showstatistics() {
 	BOARDST statb;
 
 	clriwin();
 	clrowin();
 	wmove(stdscr, 10, 0);
 	if (disk == -1) {
 		waddstr(stdscr, "You must open the disk first.");
 		pak();
 		return;
 	}
 
 	statb.commands = FALSE;	/* Leave the statistics alone for now. */
 	if (ioctl(disk, DKIOCGSTAT, &statb) < 0) {
 		wprintw(stdscr, "DKIOCGSTAT: %s", sys_errlist[errno]);
 		pak();
 		return;
 	}
 
 	move(8,0);
 	addstr("Totals are:\n");
 	printw("Commands executed:\t%u\n", statb.commands);
 	printw("Read commands:\t\t%u\n", statb.reads);
 	printw("Sectors read:\t\t%u\n", statb.diskrd);
 	move(9,COLS/2);
 	printw("Total sectors read from cache:\t%u", statb.cacherd);
 	move(10,COLS/2);
 	printw("Write commands:\t\t\t%u", statb.writes);
 	move(11,COLS/2);
 	printw("Sectors written:\t\t%u\n\n", statb.diskwr);
 	if (statb.diskrd != 0)
 		printw("Cache hit rate:\t%4.2f%%\n", (float)
 			(100.0 * ((float)statb.cacherd / (float)statb.diskrd)));
 	else
 		printw("Cache hit rate:\t0.0%%\n");
 
 	/* drive 0 statistics */
 	addstr("\n\t    Errors   Alternate  ECC corrections   Non-ECC\n");
 	addstr("Drive #\t Seek   SMD    seeks    Good    Failed  data errors\n");
 	printw("  %2d\t%5d%6d%9d%8d%10d%13d\n",0,statb.d1_seek,statb.d1_smdint,
 		statb.d1_altseek, statb.d1_ecccor, statb.d1_corfail,
 		statb.d1_nonecc);
 	printw("  %2d\t%5d%6d%9d%8d%10d%13d\n",1,statb.d2_seek,statb.d2_smdint,
 		statb.d2_altseek, statb.d2_ecccor, statb.d2_corfail,
 		statb.d2_nonecc);
 	printw("  %2d\t%5d%6d%9d%8d%10d%13d\n",2,statb.d3_seek,statb.d3_smdint,
 		statb.d3_altseek, statb.d3_ecccor, statb.d3_corfail,
 		statb.d3_nonecc);
 	printw("  %2d\t%5d%6d%9d%8d%10d%13d\n",3,statb.d4_seek,statb.d4_smdint,
 		statb.d4_altseek, statb.d4_ecccor, statb.d4_corfail,
 		statb.d4_nonecc);
 
 	move(5, 0);
 	addstr("Do you want to clear the board statistics? ");
 	if ((inp_char(stdscr) | LCBIT) != 'y') {
 		addstr("No");
 		return;
 	}
 
 	move(6, 0);
 	statb.commands = TRUE;	/* Clear the statistics */
 	if (ioctl(disk, DKIOCGSTAT, &statb) < 0)
 		printw("DKIOCGSTAT: %s", sys_errlist[errno]);
 }
 
showtune(tunesb)
register TUNEST *tunesb;
{
 	wmove(stdscr, 11, 0);
 	waddstr(stdscr, "Tuning Statistics:\n");
 	wprintw(stdscr, "Combining:\t%s\t", tunesb->combining ? "ON" : "OFF");
 	wprintw(stdscr, "Ready Limit:\t%u\t", tunesb->ready_limit);
 	wprintw(stdscr, "Sort Limit:\t%u\n", tunesb->sort_limit);
 	wprintw(stdscr, "Max Commands On Board:\t%u\t", tunesb->max_cboard);
 	wprintw(stdscr, "Max Commands In A List:\t%u\n", tunesb->max_clist);
 	wprintw(stdscr, "Max SB Transfer Length:\t%u\t", tunesb->max_sbxferlen);
 	wprintw(stdscr, "Max PB Transfer Length:\t%u\n", tunesb->max_pbxferlen);
 	wprintw(stdscr, "# of Times Board Went Empty:\t%u\n", tunesb->board_empty);
 	wprintw(stdscr, "Total Requests Sorted:\t%u\t", tunesb->tot_req_sorted);
 	wprintw(stdscr, "Total Requests Combined:\t%u\n", tunesb->tot_req_combined);
 	wprintw(stdscr, "# of Times Sort Limit Reached:\t%u\n", tunesb->hit_sort_limit);
 	wprintw(stdscr, "Count of PB Transfers:\t%u\t", tunesb->cnt_pbxfer);
 	wprintw(stdscr, "Total PB Transfer Size:\t%u\n", tunesb->tot_pbxfer_size);
 	wprintw(stdscr, "Count of SB Transfers:\t%u\t", tunesb->cnt_sbxfer);
 	wprintw(stdscr, "Total SB Transfer Size:\t%u\n", tunesb->tot_sbxfer_size);
 	wprintw(stdscr, "Normal Memory Xfer Count:\t%u\n", tunesb->norm_mem_xfer_cnt);
 	wprintw(stdscr, "Corequest Memory Xfer Count:\t%u\n", tunesb->corequest_mem_xfer_cnt);
 	wprintw(stdscr, "# of Time No PB Was Found For CA:\t%u\n", tunesb->nopb_ca);
 	wprintw(stdscr, "Block Allocate Failures:\t%u\t", tunesb->blk_alloc_fail);
 	wprintw(stdscr, "Cache Expand Count:\t%u\n", tunesb->cache_exp_count);
 	wprintw(stdscr, "Cache Allocate Failures:\t%u\t", tunesb->cache_alloc_fail);
 	wprintw(stdscr, "Xfer Restart Count:\t%u\n", tunesb->xfer_restart);
 	wprintw(stdscr, "Xfer Idle Count:\t\t%u\n", tunesb->xfer_idle);
 	wprintw(stdscr, "Disk Idle Count:\t\t%u\n", tunesb->disk_idle);
 	wprintw(stdscr, "Disk Collision Count:\t\t%u\n", tunesb->disk_collision);
 	wprintw(stdscr, "Max PB Transfer Size:\t%u\t", tunesb->max_pbxfersize);
 	wprintw(stdscr, "Max SB Transfer Size:\t%u\n", tunesb->max_sbxfersize);
 	wprintw(stdscr, "MTR Idle Count:\t\t\t%u\n", tunesb->mtr_idle);
 	wprintw(stdscr, "MTR Collision Count:\t\t%u\n", tunesb->mtr_collision);
 	wprintw(stdscr, "Drive 0 FDL Errors:\t\t%u\n", tunesb->drv0_fdl);
 	wprintw(stdscr, "Drive 0 LI Errors:\t\t%u\n", tunesb->drv0_li);
 	wprintw(stdscr, "Drive 0 Sync Byte Errors:\t%u\n", tunesb->drv0_sb);
 	wprintw(stdscr, "Drive 0 Clock Glitch Errors:\t%u\n", tunesb->drv0_cg);
 	wprintw(stdscr, "Drive 1 FDL Errors:\t\t%u\n", tunesb->drv1_fdl);
 	wprintw(stdscr, "Drive 1 LI Errors:\t\t%u\n", tunesb->drv1_li);
 	wprintw(stdscr, "Drive 1 Sync Byte Errors:\t%u\n", tunesb->drv1_sb);
 	wprintw(stdscr, "Drive 1 Clock Glitch Errors:\t%u\n", tunesb->drv1_cg);
 }
 
doiocg() {
 	struct iocgrp iocg;
 	int group;
 
 	clriwin();
 	clrowin();
 	if (disk == -1) {
 		wmove(stdscr, 10, 0);
 		waddstr(stdscr, "You must open the disk first.");
 		pak();
 		return;
 	}
 
 	wmove(stdscr, 9, 0);
 	waddstr(stdscr, "\t     1     \t    2     \t   3    \t   4   \t   5\n");
 	waddstr(stdscr, "Group\tCache Cntrl\tRead-ahead\tRecovery\tD-retry\tND-retry\n");
 	for (group = 1; group <= MAXIOCG; group++) {
 		iocg.group = group;
 		if (ioctl(disk, DKIOCGIOCG, &iocg) < 0) {
 			wmove(stdscr, 9, 0);
 			wprintw(stdscr, "DKIOCGIOCG: %s", sys_errlist[errno]);
 			pak();
 			return;
 		}
 		showiocg(&iocg);
 	}
 
 	wmove(stdscr, 5, 0);
 	wclrtoeol(stdscr);
 	wmove(stdscr, 5, 0);
 	waddstr(stdscr, "Enter 'y' if you wish to modify these settings: ");
 	if ((inp_char(stdscr) | LCBIT) == 'y') {
 		waddstr(stdscr, "Yes");
 		askiocg(&iocg);
 	} else {
 		waddstr(stdscr, "No");
 	}
 }
 
showiocg(iocg)
struct iocgrp *iocg;
{
 	wmove(stdscr, 11 + iocg->group - 1, 0);
 	wprintw(stdscr, "  %2d \t", iocg->group);
 	wprintw(stdscr, "  0x%-2x\t", iocg->cache);
 	wprintw(stdscr, "\t  %2d\t", iocg->readahead);
 	wprintw(stdscr, "\t  0x%-2x\t", iocg->recovery);
 	wprintw(stdscr, "\t  %2d\t", iocg->dretry);
 	wprintw(stdscr, "  %2d\n", iocg->ndretry);
 }
 
askiocg(iocg)
struct iocgrp *iocg;
{
 	int item;
 
 	while (TRUE) {
 		wmove(stdscr, 5, 0);
 		wclrtoeol(stdscr);
 		wmove(stdscr, 5, 0);
 		waddstr(stdscr, "Which group do you want to change? ");
 		input(stdscr, line);
 		iocg->group = atoi(line);
 		if (iocg->group <= 0)
 			return;
 
 		wmove(stdscr, 6, 0);
 		wclrtoeol(stdscr);
 		wmove(stdscr, 6, 0);
 		if (iocg->group > MAXIOCG) {
 			waddstr(stdscr, "Invalid IO Control group");
 			continue;
 		}
 		wmove(stdscr, 7, 0);
 		waddstr(stdscr, "\n\n");
 		wmove(stdscr, 7, 0);
 		if (ioctl(disk, DKIOCGIOCG, iocg) < 0) {
 			wprintw(stdscr, "DKIOCGIOCG: %s", sys_errlist[errno]);
 			pak();
 			return;
 		}
 		wmove(stdscr, 6, 0);
 		waddstr(stdscr, "Enter item number to change, <CR> when done: ");
 		input(stdscr, line);
 		if (line[0] == '\0')
 			break;
 		item = atoi(line);
 
 		wmove(stdscr, 7, 0);
 		waddstr(stdscr, "\n\n");	/* Clears two lines */
 		wmove(stdscr, 7, 0);
 		switch (item) {
 		case 1:
 			wprintw(stdscr, " 1.  Cache control:   %x\n", iocg->cache);
 			waddstr(stdscr, "new value (in hex) or <CR> for no change: ");
 			input(stdscr, line);
 			if (line[0] == '\0')
 				break;
 			sscanf(line, "%hx", &iocg->cache);
 			break;
 		case 2:
 			wprintw(stdscr, " 2.  Readahead:   %d\n", iocg->readahead);
 			waddstr(stdscr, "new value or <CR> for no change: ");
 			input(stdscr, line);
 			if (line[0] == '\0')
 				break;
 			iocg->readahead = (byte)atoi(line);
 			break;
 		case 3:
 			wprintw(stdscr, " 3.  Recovery:   %d\n", iocg->recovery);
 			waddstr(stdscr, "new value or <CR> for no change: ");
 			input(stdscr, line);
 			if (line[0] == '\0')
 				break;
 			iocg->recovery = (byte)atoi(line);
 			break;
 		case 4:
 			wprintw(stdscr, " 4.  Data retry count:   %d\n", iocg->dretry);
 			waddstr(stdscr, "new value or <CR> for no change: ");
 			input(stdscr, line);
 			if (line[0] == '\0')
 				break;
 			iocg->dretry = (byte)atoi(line);
 			break;
 		case 5:
 			wprintw(stdscr, " 5.  Non-data retry count:   %d\n", iocg->ndretry);
 			waddstr(stdscr, "new value or <CR> for no change: ");
 			input(stdscr, line);
 			if (line[0] == '\0')
 				break;
 			iocg->ndretry = (byte)atoi(line);
 			break;
 		default:
 			waddstr(stdscr, "Invalid selection...Please try again.\n");
 			break;
 		}
 		wmove(stdscr, 7, 0);
 		waddstr(stdscr, "\n\n");
 		wmove(stdscr, 7, 0);
 		if (ioctl(disk, DKIOCSIOCG, iocg) < 0) {
 			wprintw(stdscr, "DKIOCSIOCG: %s", sys_errlist[errno]);
 			pak();
 			return;
 		}
 		showiocg(iocg);
 	}
}

doconf()
{
 	struct dk_label *lp = &label;	/* make a pointer to the static label */
 	struct smd_param smdpb;		/* structure used to issue changes */
 	RETCONF retconf;	/* info returned from driver from getconf */
 	int done;		/* flat indicating completion of input	*/
 	WINDOW *confwin;   /* window in which to display configuration data */
 	int clabel = 0;	   /* TRUE indicates label info was changed in memory */
 
 	clriwin();
 	clrowin();
 	wmove(stdscr, 11, 0);
 	if (disk == -1) {
 		waddstr(stdscr, "You must open a disk device first.");
 		pak();
 		return;
 	}
 	if (ioctl(disk, DKIOCGSMD, &retconf) < 0) {
 		wprintw(stdscr, "DKIOCGSMD: %s", sys_errlist[errno]);
 		pak();
 		return;
 	}
 
 	/* create a window to display the modifyable parameters */
 	confwin = newwin (7,COLS/2,10,COLS/2);
 
 	/* copy the modifyable parameters into the structure used to change
 	   them */
 	smdpb.smd_interleave = retconf.interleave;
 	smdpb.smd_headskew = retconf.headskew;
 	smdpb.smd_cylskew = retconf.cylskew;
 	smdpb.smd_hgskew = retconf.hgskew;
 	smdpb.smd_recovery = retconf.recovery;
 	smdpb.smd_idpre = retconf.idpre;
 	smdpb.smd_datapre = retconf.datapre;
 
 	/* display the info which, though interesting, cannot be changed here */
 	wmove(stdscr, 10, 0);
 	waddstr(stdscr, "Non-changeable parameters\n");
 	wprintw(stdscr, " bytesec : %d\n", retconf.bytsec);
 	wprintw(stdscr, "*cyldsk : %d\n", retconf.cyldsk);
 	wprintw(stdscr, "*nspares : %d\n", retconf.nspares);
 	wprintw(stdscr, "*sectrk : %d\n", retconf.sectrk);
 	wprintw(stdscr, "*headcyl : %d\n", retconf.headcyl);
 	wprintw(stdscr, " basehead : %d\n", retconf.basehead);
 	wprintw(stdscr, " flags : %d\n", retconf.flags );
 	wprintw(stdscr, "*secdsk : %d\n", retconf.secdsk);
 	waddstr(stdscr, "   * - Changeable in label\n"); /* 1.9f */
 
 	/* since we have changed parts of stdscr which overlay the
 	   confwin, we better refresh stdscr before confwin, or
 	   we'll erase it as fast as we write it */
 	refresh();
 	showconf(confwin, &smdpb);
 
 	wmove(stdscr, 5, 0);
 	waddstr(stdscr, "Do you wish to change these parameters? ");
 	if ((inp_char(stdscr) | LCBIT) != 'y') {
 		waddstr(stdscr, "No");
 		delwin(confwin);
 		return;
 	}
 	waddstr(stdscr, "Yes");
 
 	done = FALSE;
 	do {
 		wmove(stdscr, 6, 0);
 		wclrtoeol(stdscr);
 		wmove(stdscr, 6, 0);
 		waddstr(stdscr, "Enter the number to change > ");
 		input(stdscr, line);
 		wmove(stdscr, 7, 0);
 		wclrtoeol(stdscr);
 		wmove(stdscr, 7, 0);
 		switch (atoi(line)) {
 			case 1 :
 				wprintw(stdscr, "interleave = %d > ", retconf.interleave);
 				input(stdscr, line);
 				smdpb.smd_interleave = atoi(line);
 				break;
 			case 2 :
 				wprintw(stdscr, "headskew = %d > ", retconf.headskew);
 				input(stdscr, line);
 				smdpb.smd_headskew = atoi(line);
 				break;
 			case 3 :
 				wprintw(stdscr, "cylskew = %d > ", retconf.cylskew);
 				input(stdscr, line);
 				smdpb.smd_cylskew = atoi(line);
 				break;

 			case 4 :
 				wprintw(stdscr, "hgskew = %d > ", retconf.hgskew);
 				input(stdscr, line);
 				smdpb.smd_hgskew = atoi(line);
 				break;
 			case 5 :
 				wprintw(stdscr, "recovery = %d > ", retconf.recovery);
 				input(stdscr, line);
 				smdpb.smd_recovery = atoi(line);
 				break;
 			case 6 :
 				wprintw(stdscr, "idpre = %d > ", retconf.idpre);
 				input(stdscr, line);
 				smdpb.smd_idpre = atoi(line);
 				clabel = 1;
 				break;
 			case 7 :
 				wprintw(stdscr, "datapre = %d > ", retconf.datapre);
 				input(stdscr, line);
 				smdpb.smd_datapre = atoi(line);
 				clabel = 1;
 				break;
 			default:
 				done = TRUE;
 				break;
 		}
 		showconf(confwin, &smdpb);
 	} while (!done);
 
 	wmove(stdscr, 6, 0); clrtoeol();
 	addstr("Do you wish to change the configuration to match this data? ");
 	if ((inp_char(stdscr)|LCBIT) == 'y') {
 	    waddstr(stdscr, "Yes");
 
 	    /* if the information also included in the label is different from
 	       the corresponding info currently in the label, change it,
 	       mark the label as modified, and recompute the checksum. */
 	    if ((lp->dkl_gap1 != smdpb.smd_idpre) || 
 			(lp->dkl_gap2 != smdpb.smd_datapre)) {
 		lp->dkl_gap1 = smdpb.smd_idpre;
 		lp->dkl_gap2 = smdpb.smd_datapre;
 		lbcur = 0;  /* mark label as "not current" */
 		complbck(lp);
 		} /* if information in label changed */
 
 	    wmove(stdscr, 9, 0);
 	    if (ioctl(disk, DKIOCSSMD, &smdpb) < 0) {
 		wprintw(stdscr, "DKIOCSSMD: %s", sys_errlist[errno]);
 		}
 	    }  /* go ahead and issue change to drive */
 	else { /* do not issue change */
 	    waddstr(stdscr, "No");
 	    }
 
 	pak();
 
 	delwin(confwin);
 	return;
 
     } /* doconf */
 
showconf(win,smdinfo)
WINDOW *win;
struct smd_param *smdinfo;
{
 
 	wclear(win);
 	wprintw(win, "1. interleave : %d\n", smdinfo->smd_interleave);
 	wprintw(win, "2. headskew : %d\n", smdinfo->smd_headskew);
 	wprintw(win, "3. cylskew : %d\n", smdinfo->smd_cylskew);
 	wprintw(win, "4. hgskew : %d\n", smdinfo->smd_hgskew);
 	wprintw(win, "5. recovery : %d\n", smdinfo->smd_recovery);
 	wprintw(win, "6. idpre : %d\n", smdinfo->smd_idpre);
 	wprintw(win, "7. datapre : %d\n", smdinfo->smd_datapre);
 	wrefresh(win);
 
 	} /* showconf */
 
addbadblock(which)
daddr_t which;
{
 	if (badblki >= maxbadblks) {
 		printf("Too many bad blocks - Increase size of badblocks array.\n");
 		pak();
 		return;
 	}
 
 	badblocks[badblki++] = which;
 }
 
mapbadblocks() {
	printf("Map Bad Blocks!\n");
}

/* rdblkno will figure a block number from input, either using a logical block
 * as entered, or cylinder/head/sector and the current geometry.
 */
rdblkno(bn,gp)		/* 1.16i */
long *bn;
struct dk_geom *gp;
{
	struct dk_geom geometry;
	int cyl, head, sect;

	*bn = -1L;
	input(stdscr, line);

	if (isupper(line[0])) line[0] = toupper(line[0]);
	if (line[0] == '\0' || line[0] == 'q') return 1;

	if (sscanf(line," %d / %d / %d",&cyl, &head, &sect) == 3) {
		if (gp == NULL) {
			gp = &geometry;
			if ((ioctl(disk, DKIOCGGEOM, &geometry) < 0) ||
					(geometry.dkg_ncyl == 0)) {
				if (glabel) {
					geometry.dkg_ncyl = label.dkl_ncyl;
					geometry.dkg_nhead = label.dkl_nhead;
					geometry.dkg_nsect = label.dkl_nsect;
					
					}
				else {
					addstr("Must have geometry or label to compute block number\n");
					return 0;
					}
				}
			}
		if (cyl < 0 || cyl >= gp->dkg_ncyl || head < 0 
			|| head >= gp->dkg_nhead || sect < 0 ||
			sect >= gp->dkg_nsect) return 0;
		*bn = ((cyl * gp->dkg_nhead + head) * gp->dkg_nsect) + sect;
		return 0;
		}
	else {
		if (sscanf(line, " %ld", bn) != 1) return 0;
		if (*bn < 0L) return 0;
		}
	} /* rdblkno */
