/*******************************************************************************
********************************************************************************
****                                                                        ****
****                   Copyright 1986, 1989 Ciprico, Inc.                   ****
****                                                                        ****
****                         Ciprico Incorporated                           ****
****                           2955 Xenium Lane                             ****
****                          Plymouth, MN  55441                           ****
****                            (612) 559-2034                              ****
****                                                                        ****
****	Module Name:	rf.c                                                ****
****	Package:	Rimfire 3200/3400 Sun OS driver for Sun 3's and 4's ****
****	Date:		$Date:   13 Feb 1989 11:47:14  $                    ****
****	Module Rev:	$Revision:   1.16  $                                ****
****	                                                                    ****
****	Subroutines:    see below                                           ****
****	                                                                    ****
****	Description:                                                        ****
****		Rimfire 3200 / Rimfire 3400 Sun OS driver                   ****
****	                                                                    ****
********************************************************************************
*******************************************************************************/

/*******************************************************************************
*									       *
*				Revision History			       *
*									       *
********************************************************************************

    Revision    Date		Author
	Description of Change
--------------------------------------------------------------------------------

    1.1		07/01/86	Michael J. Tobias
	Initial Release
    1.2		07/18/86	Michael J. Tobias
	Update includes sector slipping and board statistics

    1.3 	3/17/87		Robert Simning
	Add the read track of id's command

    1.4         5/7/87          Mark Erickson
	Added partition protection to the strategy routine

    1.5		5/29/87		John Lind
	Added ability to read defect data from drive

    1.6		6/3/87		John Lind
	Added conf flags to label, extra bytes of geometry

    1.7		6/30/87		John Lind
	Added post-mortem core dump and error numbers to error messages.

    1.8         9/21/87		John Lind	Annotated
	a) Made block numbers in error messages relative to cyl 0, hd 0 instead
	   of current parition.
	b) Decode device number into unit number, rather than the full
	   16-bit quantity (major and minor in hex).
	c) Added address operator to dumpsb commands where missing.
	d) Defined out "register" attributes for all pointer types because
	   of compiler bugs dealing with register pointers.  This can be
	   disabled by undefining CREGBUG.
	e) Replaced hard-coded timeout checks with Sun-provided CDELAY
	   macro for SOURCE portability across all Sun models.  Since
	   CDELAY relies on an external value "cpudelay", this may
	   be object code portable, too.
	f) Commented out code which set gap1 and gap2 to default (0)
	   during set geometry.

    1.9		87/11/11 - 88/1/3	John Lind	Annotated
	a) Added the support for reading the defect list from the 3400.
	   [rf.c & rfioctl.h, see also rfutil]
	b) Increased the driver support to 15 logical units from 4.
	   [rfreg.h]
	c) Eliminated printing of error messages for ioctl calls, changed
	   error returns to return the error through the ioctl parameters
	   in all cases.
	   [rf.c & rfioctl.h, see also rfutil]
	d) Allowed truncated read of long transfer near end of partition.
	   [rf.c]
	e) Change loop wait for room in the PB queue to a sleep for sb return.
	   [rf.c]
	f) Add "Interrogate SMD" and "Return ESDI configuration" ioctl calls.
	   [rf.c, rfioctl.h, see also rfutil]
	g) If the transfer size is not a multiple of DEV_BSIZE, bump up the
	   size and allow the transfer.  This is because the kernel swap
	   routine calls us with a transfer size which is not a multiple of
	   the block size when it is swapping page tabel entries.  Since 
	   network clients have their swap partitions on a host that doesn't
	   know (at the driver level) that it is anyone's swap partition, we
	   can't even check to just do this for swaps.  *WHY* does Berkely
	   *DO* this to us?!?!
	   [rf.c]

    1.10	88/1/28 - 88/3/14	John Lind	Annotated
	a) Fix the new, absolute block error messages to use the address from
	   the status block (oops!).  See 1.8a.
	   [rf.c]
	b) Print out the version number of the driver on boot-up.
	   [rf.c]
	c) Decrement nserv appropriately for the extra status block
	   returned from C_INTERROGATE on some 3200-type controllers
	   REGARDLESS of whether or not DSKBUG is set (oops!).
	   [rf.c]
	d) Use S/G commands to handle transfers which are not exact multiples
	   of the sector size (see 1.9g above -- SIGH!).  av_back is used
	   to both flag the existence of the SG list structure and to remember
	   its address for release.  Since the av_back pointer was already in
	   use to keep md_hd information for mbrelse, that code was changed to
	   look up this information in the rfintr routine, freeing av_back
	   for use in this new application.  Also needed to change the
	   SGDESCRIPTOR field order, since it was backwards!
	   [rf.c,rfreg.h]
	e) Use the MBI_ADDR macro instead of RF_ADDR for addresses returned
	   by mbsetup.  Continue using RF_ADDR for rmalloc-acquired addresses.
	   This won't actually matter until the driver is set up as a dual-
	   addresses space device, since the current configuration makes
	   the system think that we can only handle 16 bit addresses, anyway.
	   [rf.c]
	f) Issue motor start to rf3400 controllers, but not Pick and Hold
	   to 3200s because they blow up.
	   [rf.c,rfreg.h]
	g) Eliminated redundant printing of error codes as found in rfmsg.
	   [rferr.h]
	h) Generalized the include file specifications.
	   [rf.c]
	i) Use address modifier 3D for sun machines other than 3/1XX instead
	   of 0D.
	   [rfreg.h]
	j) Added parameter to rferr to force printing of status blocks
	   when desired.
	   [rf.c]
	k) Added defines for TRUE and FALSE if not already defined.
	   [rfreg.h]
	l) Reduced the number of parameter blocks and status blocks
	   to be more in keeping with the limitations of the DVMA resource.
	   [rfreg.h]

   1.11		4/27/88		J. D. Flatt
        1. No Change
    
   1.12		88/5/31 -- 88/7/5	J. E. Lind
	a) Fixed an ifndef with no tag in introduced in 1.11.
	   [rferr.h]
	b) Added support for OS4.
	   [rfreg.h,rf.c]

   1.13		88/7/20 - 88/7/27		J. E. Lind
	a) Changed rfslave to actually work, using the Issue Tag/Return
	   Status command for SMDs, RETURN STANDARD STATUS for ESDIs.
	   [rf.c,rfreg.h]
	b) Dropped support for the errintr field -- no longer used.
	   [rf.c]
	c) Made error message for no ST_CC during sglcmd clearer.
	   [rf.c]

    1.14	88/8/8		J. E. Lind
	a) No change.

   $Log:   I:\software\drivers\rf3200\sun\vcs\rf.c_v  $
   
      Rev 1.16   13 Feb 1989 11:47:14   JLind
   1.16:
     a) Increase the I/O control group representation to one per controller
        rather than one per driver.  This is really a bug fix for a bug
        left over from when the driver was first expanded to support
        multiple controllers.
        [rf.c,rfparam.h]
   
     b) Removed the GOPTWDL (Write Data Loaded) flags since the handling
        of Write Data Loaded had never been completely implemented.  This
        problem first showed up when the controllers began looking at
        this flag.
        [rf.c]
   
     c) Cleared the label magic number before reading the label to
        ensure that we don't get a false label display.
        [rf.c]
   
     d) Any sector containing an error according to the manufacturer's
        defect list is now considered bad.  Before, errors contained
        within the head scatter and gap3 were considered "harmless".
        [rfutil.c]
   
     e) Add the ability to examine the IDs of a track.  The track is reported
        as being map or bad, or the skew is shown, along with any "inter-
        esting" sectors.  The ioctl DKIOCTRAKID command (as used by compsec
        within rfutil) was renamed to DKIOCPHYSID to better reflect its
        function (returning the ID of a specified physical sector on a track),
        while DKIOCTRAKID returns the IDs of all sectors on a track. 
        DKIOCPHYSID is the same value as the old DKIOTRAKID for binary
        compatibility.  In the process, the interesting id FLAG, SECTOR,
        and ALT values were given meaningful names and moved to rfreg.h.
        XTCOFF and POSCOFF were moved from rfparam.h to rfreg.h.
        [rfutil.c,rfreg.h,rfparam.h,rfioctl.h,rf.c]
   
     f) The input routine (getcyltrack) which requests block numers or
        cylinder/head pairs now does so on one line, hopefully reducing
        user confusion.
        [rfutil.c]
   
     g) The map sector routine now asks whether you wish to map by track
        or sector.  Mapping by track helps out by figuring the track
        boundaries.
        [rfutil.c]
   
     h) A bug in ver_cyl incorrectly computed the block number from c/h/s,
        failing to verify the full disk.  The block numbers reported as
        failing really did fail and really were processed properly, but did
        not correspond to the c/h/s displayed.  Also, the variable "badhead"
        and the associated computations were discarded from the "ver_to"
        routines, since this information was already available.
        [rfutil.c]
   
     i) Slipping now accepts the block to slip either in c/h/s or as
        a block number through the use of a new routine, rdblkno.
        Rdblkno will accept input in either form if the label or
        geometry is available, but only as a block number if they are
        not.
        [rfutil.c]
   
     j) Corrected bzero(...,sizeof (EXTB)) to be bzero(...,sizeof(EXTPB)).
        [rf.c]
   
     k) Corrected comparison of unit number to be >= NRF instead of > NRF
        in several locations and moved assignment of the automatic variable
        device out of the declaration in rfdump() so that a bounds check is
        performed on unit.
        [rf.c]
   
      Rev 1.15   23 Nov 1988 13:07:50   JLind
           a) Corrected the indexing of rfcontroller_info to be by
   	   CONTROLLER number rather than DRIVE number (duh!),
   	   thereby fixing the dreaded dkinfo and newfs crashes.
   	   [rf.c]
   	b) Inserted debug flag tests on "not alive" messages.
   	   [rf.c]
   	c) Dual-port support.
   	   [rf.c,rfreg.h]
   	d) STANDALONE support for standalone rfutil.
   	   [rf.c, rfutil.c]
   	e) Changed default control group cache control to use full cacheing
   	   and sorting. [rfparam.h]
   	f) added safealloc and safefree (replacing rmalloc and rmfree,
   	   respectively) to make the driver insensitive to misalignment of
   	   the iopbmap resource, while also not causing it for anyone else.
   	   [rf.c]
    
*******************************************************************************/

/*

	The major points of interest in this file are ordered as follows:
		Include the header files
                break;
 		Subroutine definitions
		Sun driver and controller structures
		Rimfire 3200 driver internal structures
		The subroutines themselves

*******************************************************************************

	The subroutines interact as follows.  The interaction shown is only
	for those subroutines within this driver, UNIX subroutines are
	not shown.

	ROUTINE or SYSTEM		CALLS
Level 0:
	configuration startup	==>	rfattach, rfprobe, rfslave
	I/O system		==>	rfdump, rfioctl, rfopen,
					rfread, rfwrite, rimfsize,
					rfstrategy
	Interrupt mechanism	==>	rfintr

Level 1:
	rfattach		==>	sglcmd
	rfprobe			==>	sglcmd
	rfslave			==>	sglcmd

	rfdump			==>	(nothing)
	rfioctl			==>	config, format, maptrack, slip, retconf,
					defsmd, setiocg, verify, getstat,
					readtrkid, readmap, readddat
	rfread			==>	rfstrategy
	rfwrite			==>	rfstrategy
	rimfsize			==>	(nothing)
	rfstrategy		==>	putpb

	more follow...
*/


/* 1.8d */
/*#define CREGBUG		/* comment out this line if you don't have the register bug */

#ifdef CREGBUG
# define REGISTER
#else
# define REGISTER register
#endif

#ifndef lint
static char	Sccsid[] = "@(#)rf.c  75182013 $Revision:   1.16  $ $Date:   13 Feb 1989 11:47:14  $ Copyright 1986, 1989 Ciprico, Inc.";
#endif

/* The include file rf.h is created during the configuration process on Sun
 * machines.  Its name itself is dependent on the configuration process.
 * Basically, what is needed in the include file are two defines:
 *	NRF	- Defines how many disk drives may be used
 *	NRFC	- Defines how many controllers may be used
 */
#include "rf.h"

/* don't bother compiling if there aren't any drives */
#if NRF > 0

/* Include the required system files	Begin 1.10h */

#include <sys/param.h>
#include <sys/dir.h>
#include <sys/user.h>
#include <sys/buf.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/map.h>
#include <sys/ioctl.h>
#include <sys/uio.h>
#include <sys/file.h>
#include <sys/dk.h>
#include <sys/vnode.h>

#include "../machine/psl.h"
#include "../sun/dkio.h"
#include "../sun/dklabel.h"

/* end 1.10h */

/* and our required files */
#include "../sundev/rfreg.h"
#include "../sundev/rfioctl.h"
#include "../sundev/rfparam.h"
#include "../sundev/rferr.h"

#include "../sundev/mbvar.h"

#define GNB		/* Add gnb's code to make iostat work */

/* Get a Direct Virtual Memory Address (DVMA) from a virtual kernel address */
#define	RF_ADDR(x) ((int)((int)(x) & 0xfffff))	/* DVMA address (20 bits) */

/* For those routines in the driver which are visible to the outside world,
 * define what they return.
 */
int rfprobe(), rfattach(), rfslave(), rfopen(), rfintr();
int rfioctl(), rfread(),   rfwrite(), rfdump();
daddr_t rimfsize();
void rfstrategy();
static caddr_t safealloc();
static void safefree();

/* Define the structures and storage needed for configuration and runtime */
static struct mb_ctlr *rfcontroller_info[NRFC];	/* per controller */
static struct mb_device *rfdrive_info[NRF];	/* per drive */
static int nservint[NRFC];		/* pending requests per controller */
#ifdef NOHANG
static int waitsb = FALSE;		/* wait for sb */
#endif NOHANG

/* the driver structure */
struct mb_driver rfcdriver = {
	rfprobe,		/* the probe routine */
	rfslave,		/* the slave routine */
	rfattach,		/* the attach routine */
	0,			/* there is no mdr_go routine */
	0,			/* there is no mdr_done routine */
	rfintr,			/* the intr routine */
	0,			/* no memory used */
	"rf",			/* the device name */
	rfdrive_info,		/* info per drive */
	"rfc",			/* the controller name */
	rfcontroller_info,	/* info per controller */
	MDR_BIODMA,		/* we do DMA transfers */
	0			/* interrupt routine linked list */
};

/********************************************
 *
 * Structures used internally by this driver
 *
 ********************************************/


/* Debugging aids */
/* #define	DSKBUG          /* comment this line out to run faster */
static int	debuglevel = 0x00;	/* all debug levels OFF */
#ifdef	RFSTATS
/* Statistic gathering information: */
static int	iorqsi = 0;
static IORQ	iorq;
static int	niorqs = sizeof (iorq.iorqs) / (sizeof iorq.iorqs[0]);
extern struct timeval	time;
#endif RFSTATS

/* How long can we hold the bus? */
#define THROTTLE	8		/* 8 transfers (32 bytes)is the size
					   of the bus fifo */

/* how long until controller reset or single command should complete */
/* 1.8e -- changed SECONDS to 1000000 from 60000 for compatibility w/DELAY */
#define	SECONDS		1000000		/* one second of DELAY */
#define TOOLONG 	(20*SECONDS)	/* wait 20 seconds */

/* UNIX buffer requests */
static struct buf ioctlbuf;		/* for ioctl calls */
static struct buf rrfbuf[NRF];		/* for raw reads and writes */

/* The Rimfire 3200 and 3400 use circular buffer structures for passing
 * parameter blocks to the controller and returning status blocks
 * from the controller.  This structure must be physically contiguous.
 */
static CMDLIST	*cmdq[NRFC];		/* one per controller */

/* This is the BSY bit of the Rimfire's Status Port as the controller
 * pictures it.  It should be set to zero on a controller reset and toggle
 * for each type zero channel attention.
 */
static int	bsybit[NRFC];
/* This is the controller type of each controller.  The value is read from
 * the status register of the board after a reset.  See rfprobe().
 */
static int	rfctype[NRFC];

/* To keep track of all the parameter blocks in the RF3200, the ID field
 * of the standard parameter block is actually a pointer to the associated
 * UNIX buffer.
 * Thus, when we receive a status block from the controller, we can
 * use its identifier to determine which UNIX buffer to free.
 * We also need to keep the return value of mbsetup (mbinfo) so that
 * we can release the entries in the kernel memory map via mbrelse.
 * Since we don't use the available list pointers in the buf structure
 * we will abuse them by saving mbinfo as a pointer to struct buf.
 * 1.10d
 * The av_back pointer will be abused to point to a buf
 * structure allocated through geteblk for the purpose of putting the
 * trivial case of a S/G list in it for the brain-damaged swapping code
 * which does non-block transfers through the strategy routine.
 */
#define b_mbinfo av_forw	/* from struct buf */
#define	b_sglist av_back	/* from struct buf */

/* the geometry for each of the drives */
static struct dk_geom geometry[NRF];

/* the partition information for each of the drives */
static struct dk_map partitions[NRF][NDKMAP];

/*******************************************************************************
*									       *
*	Subroutine:		rfprobe					       *
*									       *
*	Calling Sequence:	rfprobe()				       *
*									       *
*	Called by:		UNIX initialization			       *
*									       *
*	Calling Parameters:	None					       *
*									       *
*	Local Variables:	                                               *
*				setup - setup command list parameter block     *
*				iocgpb - set iocg parameter block              *
*				gopt - set general board options               *
*									       *
*	Calls Subroutines:	pokec, sglcmd                          	       *
*									       *
*	Public/Global Variables:    					       *
*									       *
*	Description:							       *
*		This routine is called at startup time to probe for a          *
*		controller, to reset it, and to configure it for any options   *
*		                                                               *
*******************************************************************************/

rfprobe(rf, unit)
REGISTER RF32REG *rf;
{
	int finished;	/* save reset complete status		*/
	static char tellem = 1;	/* Identify driver for them just once 1.10b */

#ifdef DSKBUG
	if (debuglevel&1)
		printf("rfprobe: controller %d at %x\n", unit, rf);
#endif DSKBUG

	/* Reset controller and check if it exists */
	if (pokec((char *)&rf->reset, 1) != 0) {
		/* it's not there */
#ifdef DSKBUG
        	if (debuglevel&1)
			printf("rfprobe: no controller present at %x\n", rf);
#endif DSKBUG
		return(0);
	}
	bsybit[unit] = 0;

	/* Wait for reset to complete 1.8e*/
#ifdef TOOLONG
	CDELAY((finished = (rf->status & SWAB(STATRST)) == SWAB(RESETDONE)),
		TOOLONG);
	if (!finished) {
		printf("rfprobe: Cannot reset controller: status %x\n",
			SWAB(rf->status));
		return(0);
		}
#else TOOLONG
	while ((rf->status & SWAB(STATRST)) != SWAB(RESETDONE))
		;
#endif TOOLONG
	rfctype[unit] = rf->status & STATCTYPE;

#ifdef DSKBUG
	if (debuglevel&1)
		printf("Controller %d is a ", unit);
	if (debuglevel&1)
		printf("%s\n", rfctype[unit] == STATRF3200 ? "3200" : "3400");
#endif

	/* If all has gone well for at least one controller, tell them that
	 * we are here, but only once.
	 */
	if (tellem) {
		printf("rf: RF3200/RF3400 driver version 1.16\n");
		tellem = 0;
		}
	return(sizeof(RF32REG));
}

/*******************************************************************************
*									       *
*	Subroutine:		rfslave					       *
*									       *
*	Calling Sequence:	rfslave(dev, rf)			       *
*									       *
*	Called by:		UNIX configuration startup		       *
*									       *
*	Calling Parameters:	dev - pointer to device structure  	       *
*	                        rf - controllers I/O address     	       *
*									       *
*	Local Variables:	xpb - an extended parameter block              *
*	                        pb - standard parameter block    	       *
*	                        pbaddr - address of extended block    	       *
*	                        success - TRUE for drive present    	       *
*									       *
*	Calls Subroutines:	NONE	                           	       *
*									       *
*	Public/Global Variables:None					       *
*									       *
*	Description:							       *
*		This routine checks for the existence of the specified drive.  *
*		We do this by issuing a an ISSUE TAG/RETURN STATUS command     *
*		for the 3200, or ISSUE ESDI SERIAL COMMAND with REQUEST        *
*		STANDARD STATUS for the 2400.  If the controller fails this    *
*		command, then the drive is not usable.			       *
*		                                                               *
*******************************************************************************/

static
/*ARGSUSED*/
rfslave(dev, rf)
REGISTER struct mb_device *dev;
REGISTER RF32REG *rf;
{
	REGISTER EXTPB *xpb;
	REGISTER PARMBLK *pb;
	dword	pbaddr;
	int success;

	/* Check the device "flags" parameter...see if they set up the */
	/* configuration file correctly. */
	if ((VOL(dev->md_flags) != 1 && VOL(dev->md_flags) != 2)
#ifndef ENABLE_DUAL_PORT
		|| (rfctype[dev->md_ctlr] != STATRF3200 && 
			VOL(dev->md_flags) != dev->md_flags)
#endif
			) {
		printf("**The configuration is not correct for the Rimfire %s #%d unit %d.\n",
			rfctype[dev->md_ctlr] == STATRF3200 ? "3200" : "3400",
			dev->md_ctlr,dev->md_slave);
		printf("**The \"flags\" value should contain a 1 for the main ");
		printf("volume, or 2 for the\n  second volume");
#ifdef ENABLE_DUAL_PORT
		printf(" with other flags as needed");
#endif ENABLE_DUAL_PORT
		printf(".\n**Please check your manual.\n");
		return(0);
	}

	/* get an extended parameter block */
	xpb = (EXTPB *)safealloc(sizeof(EXTPB));

	/* initialize extended parameter block - disable interrupts */
	xpb->intr = 0;
	xpb->errintr = 0;
	xpb->resv = 0;

	/* Set up the drive command 1.13a */
	pb = (PARMBLK *)&xpb->pb;
	if (rfctype[dev->md_ctlr] == STATRF3200) {
		pb->command = C_TAGSTATUS;
		pb->diskaddr = 0;
		}
	else {
		pb->command = C_ISSUEESDI;
		/* This decodes into the "Status returned" bit plus
		   the return status word flag */
		pb->diskaddr = 0x12000;
		}
	pb->id = 0;
	pb->control = 0;
	pb->unit = (dev->md_slave << 1) + VOL(dev->md_flags);
	pb->addrmod = 0;
	pb->vmeaddr = 0;
	pb->blkcount = 0;
	pb->reserved = 0;

	/* Clear the status block */
	xpb->sb.id = 0;
	xpb->sb.status = 0;
	xpb->sb.errcode = 0;
	xpb->sb.errextra = 0;
	xpb->sb.errdata = 0;

	/* convert extpb pointer to a long for shifting */
	pbaddr = RF_ADDR(xpb);

	/* Write the control and address modifier to the address buffer port */
	rf->addrbuf = SWAB(CNTRL << 8 | VME_ADD_MOD);
	rf->addrbuf = SWAB(pbaddr >> 16 & 0xffff);
	rf->addrbuf = SWAB(pbaddr & 0xffff);

	/* Get the controller's attention */
	rf->attention = 0;

	/* Toggle our busy bit, as the controller should do when it finishes
	 * the command. */
	bsybit[dev->md_ctlr] ^= 1;

#ifdef TOOLONG
	CDELAY((xpb->sb.status & ST_CC),TOOLONG);
	if ((xpb->sb.status & ST_CC) == 0) {
		printf("rfslave: timeout -- command did not complete\n");
		printf("rfslave: command = %x, status port = %x\n",
			xpb->pb.command, rf->status);
		return(0);
	}
#else TOOLONG
	while ((xpb->sb.status & ST_CC) == 0)
		;
#endif TOOLONG

	/* check for an error */
	success = !(xpb->sb.status & ST_ERR);

	safefree((caddr_t)xpb);
	return(success);
}

/*******************************************************************************
*									       *
*	Subroutine:		rfattach				       *
*									       *
*	Calling Sequence:	rfattach(device)			       *
*									       *
*	Called by:		UNIX configuration startup		       *
*									       *
*	Calling Parameters:	device - the device to do the housework for    *
*									       *
*	Local Variables:	rf - the controller address                    *
*	                        xpb - extended parameter block   	       *
*	                        label - disk lable structure     	       *
*	                        cpb - configuration parameter block            *
*	                        pb - standard parameter block                  *
*	                        usable - usable number of disk blocks          *
*	                        physical - physical number of disk blocks      *
*	                        gp - pointer to disk geometry                  *
*									       *
*	Calls Subroutines:	sglcmd                             	       *
*									       *
*	Public/Global Variables:None					       *
*									       *
*	Description:							       *
*		This function does preliminary setup work for a                *
*		specified slave number.  If the drive is found, the label is   *
*		read and the controller configured as the label states.  If    *
*		the label doesn't exist, we try to determine one using the     *
*		interrogate command.                                           *
*		                                                               *
*******************************************************************************/

rfattach(device)
REGISTER struct mb_device *device;
{
	REGISTER EXTPB *xpb;
	REGISTER CMDLIST *cmdptr;
	struct dk_label *label;
	CONFPB *cpb;
	DEFSMDPB *smdpb;
	PARMBLK *pb;
	SETUPPB *setup;
	IOCGPB *iocgpb;
	GOPTPB *gopt;
	int usable, physical;
	struct dk_geom *gp;
	register int i;
	struct vec *vp;
	ESDIPARAMS *esdip;
	int esdi_spt;
	unsigned short *wp;
	unsigned short cksum;

/* Something is bashing ioctlbuf.b_flags just before the last call on this
   routine, so we will clear it here until we can find the real problem.
   This will be fun, because our routines are not being executed at the
   time of the problem. */

ioctlbuf.b_flags = 0;

	/* allocate space for extended parameter block */
	xpb = (EXTPB *)safealloc( sizeof(EXTPB));
	bzero(xpb,sizeof (EXTPB));	/* 1.16j */
	/* heaven only knows why the above bzero fails to nuke these few
	 * words, but this seems to make everyone happy -- is there a monster
	 * out there, waiting to get us?  I don't know...
	 */
	xpb->intr = 0 ; xpb->errintr = 0; xpb->resv = 0;
#ifdef DSKBUG
	if (debuglevel & 2)
		printf("rfattach: allocated EXTPB at %x\n", xpb);
#endif DSKBUG

	/* Allocate a command queue for each controller. */
	if (cmdq[device->md_ctlr] == (CMDLIST *) NULL) {
		cmdq[device->md_ctlr] = (CMDLIST *)safealloc(sizeof(CMDLIST));
#ifdef DSKBUG
		if (debuglevel & 2) {
			printf("rfattach: allocated cmdlist at %x", cmdq[device->md_ctlr]);
			printf(" for controller %d\n", device->md_ctlr);
		}
#endif DSKBUG
		cmdptr = cmdq[device->md_ctlr];


		/* Initialize the command list buffer:
		 *	set IN and OUT pointers to zero
		 *	set the list size fields
		 */
		cmdptr->pbin = 0;
		cmdptr->pbout = 0;
		cmdptr->sbin = 0;
		cmdptr->sbout = 0;
		cmdptr->pbsize = RF32NPB;
		cmdptr->sbsize = RF32NSB;
		cmdptr->resv1 = 0;
		cmdptr->resv2 = 0;

		/* Use an extended parameter block to issue type 0 commands
		 * disable interrupts and wait for these commands to complete.
		 */
		xpb->intr = 0;
		xpb->errintr = 0;
		xpb->resv = 0;

		/* Let the controller know where the command list buffer is */
		setup = (SETUPPB *)&xpb->pb;
		setup->id = 0;
		setup->command = C_SETUPCL;
		setup->iocg = 0;
		setup->unit = 0;
		vp = device->md_mc->mc_intr;
		setup->errintr = 0;	/* 1.13b */
		setup->doneintr = (device->md_mc->mc_intpri << 8) | vp->v_vec;
#ifdef DSKBUG
		if (debuglevel & 1) {
			printf("Doneintr=%x\n", setup->doneintr);
		}
#endif
		setup->start = RF_ADDR(cmdptr);
		setup->addrmod = VME_ADD_MOD;
		setup->listno = HOST_ID;
		setup->resv[0] = 0;
		setup->resv[1] = 0;
		setup->resv[2] = 0;
		if (sglcmd(xpb, device->md_mc)) {
			/* error, release all resources */
			printf("Cannot setup command list!\n");
			cmdq[device->md_ctlr] = (CMDLIST *) NULL;
			safefree((caddr_t)cmdptr);
			safefree((caddr_t)xpb);
			return(0);
		}

		/* Set some I/O Control Groups for later use
		 * IOCG(1) - disable cache
		 * IOCG(2) - read IOCG
		 * IOCG(3) - write IOCG
		 */
#ifdef DSKBUG
		if (debuglevel&1)
			printf("Setting I/O Control Groups\n");
#endif DSKBUG
		for (i = 1; i <= MAXIOCG; i++) {
			iocgpb = (IOCGPB *)&xpb->pb;
			iocgpb->id = 0;
			iocgpb->command = C_SETIOCG;
			iocgpb->control = i;
			iocgpb->unit = 0;
			iocgpb->resv1 = 0;
			/* 1.16a next 5 lines */
			iocgpb->recovery = iocg[device->md_ctlr][i].recovery;
			iocgpb->readahead = iocg[device->md_ctlr][i].readahead;
			iocgpb->cache = iocg[device->md_ctlr][i].cache;
			iocgpb->ndretry = iocg[device->md_ctlr][i].ndretry;
			iocgpb->dretry = iocg[device->md_ctlr][i].dretry;
			iocgpb->resv2 = 0;
			iocgpb->resv3 = 0;

			/* issue command and check for error */
			if (sglcmd(xpb, device->md_mc)) {
				/* error */
				printf("Cannot set I/O Control Group %d!\n", i);
				safefree((caddr_t)xpb);
				return(0);
			}
		}

		/* Set the General Board Options */
#ifdef DSKBUG
		if (debuglevel&1)
			printf("Setting General Board Options\n");
#endif DSKBUG
		gopt = (GOPTPB *)&xpb->pb;
		gopt->id = 0;
		gopt->command = C_OPTION;
		gopt->control = 0;
		gopt->unit = 0;
		gopt->throttle = THROTTLE;
		gopt->sbenable = GOPTDTE | GOPTIDE;	/* 1.16b */
		gopt->sbintr = 0;			/* 1.16b */
		gopt->resv[0] = 0;
		gopt->resv[1] = 0;

		/* issue general board options command and check for error */
		if (sglcmd(xpb, device->md_mc)) {
			/* error */
			printf("Cannot set General Board Options!\n");
			safefree((caddr_t)xpb);
			return(0);
		}

		/* initialize it for no interrupts */
		xpb->intr = 0;
		xpb->errintr = 0;
		xpb->resv = 0;
	}
	cmdptr = cmdq[device->md_ctlr];

	if (rfctype[device->md_ctlr] == STATRF3400) {
		esdip = (ESDIPARAMS *) safealloc(sizeof(ESDIPARAMS));
		if (esdip == (ESDIPARAMS *) NULL) {
#ifdef DSKBUG
			printf("rf: Can't alloc mem for ESDIPARAMS\n");
#endif
			goto out;
		}
		/* Set up command to get the parameters. */
		pb = &xpb->pb;
		pb->id = 0;
		pb->command = C_GETESDICONF;
		pb->control = 0;
		pb->unit = (device->md_slave << 1) + VOL(device->md_flags);
		pb->addrmod = VME_ADD_MOD;
		pb->diskaddr = 0;
		pb->vmeaddr = RF_ADDR(esdip);
		pb->blkcount = 0;
		pb->reserved = 0;

		if (sglcmd(xpb, device->md_mc)) {
			/* failed, this should never happen unless the drive is
			 * missing or dead.
			 */
			printf("Can't get ESDI Configuration/status!\n");
			safefree((caddr_t) esdip);
			safefree((caddr_t)xpb);
			return(0);
		}


		if (esdip->genconf & GC_DRVHARD) {
			/* Hard-sectored Drive */
			esdi_spt = 0;
		} else {
			/* Soft-sectored Drive */
			esdi_spt = esdip->minbytrk / 700;
		}

		safefree((caddr_t) esdip);
	}

out:
	/* Set up a configuration for a minimal disk */
	cpb = (CONFPB *)&xpb->pb;
	cpb->id = 0;
	cpb->command = C_SETCONFIG;
	cpb->control = 0;
	cpb->unit = (device->md_slave << 1) + VOL(device->md_flags);
	cpb->resv1 = 0;
	cpb->bytsec = 512;
	cpb->basehead = 0;
	if (rfctype[device->md_ctlr] == STATRF3200) {
		cpb->sectrk = 1;
		cpb->cyldsk = 1;
		cpb->headcyl = 1;
	} else {
		cpb->sectrk = esdi_spt;
		cpb->cyldsk = 0;
		cpb->headcyl = 0;
	}
	cpb->nspares = 0;
	if (rfctype[device->md_ctlr] == STATRF3200)
		cpb->flags = CONFSSP;
	else
		cpb->flags = 0;	/* STATRF3400 */

#ifdef ENABLE_DUAL_PORT
	if (device->md_flags & MDF_DUALPORTED) {
		cpb->flags |= (CONFEDP|CONFBSY);
	}
#endif

	cpb->resv2[0] = 0;
	cpb->resv2[1] = 0;
	cpb->resv2[2] = 0;
#ifdef DSKBUG
	if (debuglevel & 1)
		printf("Configuring minimal disk unit %d, volume %d\n", device->md_slave, VOL(device->md_flags));
#endif DSKBUG
	/* try the command */
	if (sglcmd(xpb, device->md_mc)) {
		/* failed, this should never happen */
		printf("Preliminary configuration failed!\n");
		safefree((caddr_t)xpb);
		return(0);
	}

	/* Issue the motor start command to ESDI drives  1.10f */
	if (rfctype[device->md_ctlr] != STATRF3200) {
		pb->id = 0;
		pb->command = C_MSTART;
		pb->control = 0;
		pb->unit = (device->md_slave << 1) + VOL(device->md_flags);
		pb->addrmod = VME_ADD_MOD;
		pb->diskaddr = 0;
		pb->vmeaddr = 0;
		pb->blkcount = 0;
		pb->reserved = 0;

		/* issue motor start */
		if (sglcmd(xpb, device->md_mc)) {
			/* error occured, assume no drive there */
			printf("slave: Cannot issue start to drive %d, volume %d!\n",
				device->md_slave, VOL(device->md_flags));
			safefree((caddr_t)xpb);
			return(0);
			}
		}

	/* get some memory for holding a disk label */
	label = (struct dk_label *)safealloc(sizeof(struct dk_label));

	/* try to read the label */
	if (label == (struct dk_label *)NULL) {
		/* couldn't get memory for the disk label */
		printf("No memory for the disk label.\n");
		safefree((caddr_t)xpb);
		return(0);
	}

	/* set up a read command for the first sector */
	pb = &xpb->pb;
	pb->id = 0;
	pb->command = C_READ;
	pb->control = 0;
#ifdef ENABLE_DUAL_PORT
	if(device->md_flags & MDF_RELEASE) pb->control |= RF_RLS;
#endif
	pb->unit = (device->md_slave << 1) + VOL(device->md_flags);
	pb->addrmod = VME_ADD_MOD;
	pb->diskaddr = 0;
	pb->vmeaddr = RF_ADDR(label);
	pb->blkcount = 1;
	pb->reserved = 0;

	/* make sure that we won't find a "leftover" label. 1.16c */
	label->dkl_magic = 0;

	/* issue the read command and check for errors */
	if (sglcmd(xpb, device->md_mc)) {

		/* the read failed, the drive is probably not formatted
		printf("The disk may not exist or be formatted.\n");


		/* release memory used for disk label */
		safefree((caddr_t)label);

		/* zero the disk geometry */
		bzero(&geometry[device->md_unit], sizeof(struct dk_geom));

		/* release resources and return */
		safefree((caddr_t)xpb);
		return(1);	/* Leave it alive */
	}

	/* read worked, see if there is a valid label on this sector */

	/* Compute what the checksum should be... */
	cksum = 0;
	for (i = 0, wp = (unsigned short *)label; i < sizeof(struct dk_label)/2-1; i++)
		cksum ^= *wp++;
	if (cksum != label->dkl_cksum || label->dkl_magic != DKL_MAGIC) {
		if (cksum != label->dkl_cksum)
			printf("Bad checksum in label on unit %d\n", device->md_unit);
		else
			printf("No label on unit %d\n", device->md_unit);

		/* release memory used for disk label */
		safefree((caddr_t)label);

		/* interrogate the drive to guess what the disk looks like */
		/* for now keep minimal config */
		gp = &geometry[device->md_unit];
		gp->dkg_ncyl = 1;
		gp->dkg_acyl = 0;
		gp->dkg_nhead = 1;
		gp->dkg_bhead = 0;
		if (rfctype[device->md_ctlr] == STATRF3200)
			gp->dkg_nsect = 1;
		else
			gp->dkg_nsect = esdi_spt;
		gp->dkg_intrlv = 1;
		gp->dkg_gap1 = 0;
		gp->dkg_gap2 = 0;

		/* release resources and return */
		safefree((caddr_t)xpb);
		return(1);
	}

	/* this disk has a valid label, lets see what it says */
	/* set the disk partition map according to the label */
	for (i = 0; i < NDKMAP; i++)
		partitions[device->md_unit][i] = label->dkl_map[i];

	/* set the geometry according to the label */
	gp = &geometry[device->md_unit];
	gp->dkg_ncyl = label->dkl_ncyl;
	gp->dkg_acyl = label->dkl_acyl;
	gp->dkg_nhead = label->dkl_nhead;
	gp->dkg_bhead = label->dkl_bhead;
	gp->dkg_nsect = label->dkl_nsect;
	gp->dkg_intrlv = label->dkl_intrlv;
	gp->dkg_gap1 = label->dkl_gap1;
	gp->dkg_gap2 = label->dkl_gap2;
#ifdef NSPARES
	gp->dkg_apc = label->dkl_apc;
#else
	gp->dkg_extra[0] = label->dkl_pad[0];
#endif

	/* remember the drive configuration flags for future
	 * C_SETCONF commands and test for unknown bits.  */
	if ((gp->dkg_extra[XTCFF]=(label->dkl_asciilabel[POSCFF] & MASKCFF))
		!= label->dkl_asciilabel[POSCFF]) 
		printf("rfattach: warning -- unknown config flag bits in label.\n");

	/* determine usable and physical sizes */
	usable = gp->dkg_ncyl * gp->dkg_nsect * gp->dkg_nhead;
	physical = usable + gp->dkg_acyl * gp->dkg_nsect * gp->dkg_nhead;
	/* add in the spare sectors for all tracks */
#ifdef NSPARES
	physical += (gp->dkg_ncyl + gp->dkg_acyl) * gp->dkg_apc;
#else
	physical += (gp->dkg_ncyl + gp->dkg_acyl) * gp->dkg_extra[0];
#endif
	printf("%s%d: <%s(%d/%d)>\n", rfcdriver.mdr_dname, device->md_unit,
		label->dkl_asciilabel, usable, physical);

	/* release the label structure */
	safefree((caddr_t)label);

	/* re-configure the disk according to the new geometry */
	cpb = (CONFPB *)&xpb->pb;
	cpb->id = 0;
	cpb->command = C_SETCONFIG;
	cpb->control = 0;
	cpb->unit = (device->md_slave * 2) + VOL(device->md_flags);
	cpb->resv1 = 0;
	cpb->cyldsk = gp->dkg_ncyl + gp->dkg_acyl;
	cpb->bytsec = 512;
	if (rfctype[device->md_ctlr] == STATRF3200)
		cpb->basehead = gp->dkg_bhead;
	else
		cpb->basehead = 0; /* STATRF3400 */
	cpb->headcyl = gp->dkg_nhead;
	cpb->sectrk = gp->dkg_nsect;
#ifdef NSPARES
	cpb->nspares = gp->dkg_apc / gp->dkg_nhead;
#else
	cpb->nspares = gp->dkg_extra[0] / gp->dkg_nhead;
#endif
	if (rfctype[device->md_ctlr] == STATRF3200)
		cpb->flags = gp->dkg_extra[XTCFF];
	else
		cpb->flags = 0;	/* STATRF3400 */
#ifdef	ENABLE_DUAL_PORT
	if (device->md_flags & MDF_DUALPORTED) {
		cpb->flags |= (CONFEDP|CONFBSY);
	}
#endif
	cpb->resv2[0] = 0;
	cpb->resv2[1] = 0;
	cpb->resv2[2] = 0;

#ifdef DSKBUG
	if (debuglevel & 1)
		printf("Final configuration.\n");
#endif DSKBUG

	/* issue the final configuration and check for errors */
	if (sglcmd(xpb, device->md_mc)) {
		/* ugh... */
		printf("Final configuration failed, reconfiguring.\n");
		/* reset the drive as a minimal disk so the label
		 * can be written.
		 */
		cpb->id = 0;
		cpb->command = C_SETCONFIG;
		cpb->control = 0;
		cpb->unit = (device->md_slave << 1) + VOL(device->md_flags);
		cpb->resv1 = 0;
		cpb->cyldsk = 1;
		cpb->bytsec = 512;
		cpb->basehead = 0;
		cpb->headcyl = 1;
		if (rfctype[device->md_ctlr] == STATRF3200)
			cpb->sectrk = 1;
		else
			cpb->sectrk = esdi_spt;
		cpb->nspares = 0;
		if (rfctype[device->md_ctlr] == STATRF3200)
			cpb->flags = CONFSSP;
		else
			cpb->flags = 0;	/* STATRF3400 */
		cpb->resv2[0] = 0;
		cpb->resv2[1] = 0;
		cpb->resv2[2] = 0;
		if (sglcmd(xpb, device->md_mc)) {
			printf("Minimal reconfiguration failed.\n");
			safefree((caddr_t)xpb);
			return(1);
		}

		/* put some sense in the geometry information */
		if (rfctype[device->md_ctlr] == STATRF3200)
			gp->dkg_nsect = 1;
		else
			gp->dkg_nsect = esdi_spt;
		gp->dkg_nhead = 1;
		gp->dkg_ncyl = 1;
		gp->dkg_acyl = 0;
		printf("%s%d: <Assuming: cyl %d alt 0 hd %d sec %d(1/1)>\n",
			rfcdriver.mdr_dname, device->md_unit,
			gp->dkg_ncyl, gp->dkg_nhead, gp->dkg_nsect);
	} else {
		/* set the SMD parameters optimally */
		smdpb = (DEFSMDPB *)&xpb->pb;
		smdpb->id = 0;
		smdpb->resv1 = 0;
		smdpb->unit = (device->md_slave << 1) + VOL(device->md_flags);
		smdpb->control = 0;
		smdpb->command = C_SETSMDPARAM;
		smdpb->interleave = 1;
		smdpb->headskew = gp->dkg_nsect > 5 ? 5 : 1;
		smdpb->cylskew = gp->dkg_nsect > 18 ? 21 : 1;
		smdpb->hgskew = 0;
		smdpb->recovery = 0;
		if (rfctype[device->md_ctlr] == STATRF3200) {
			smdpb->idpre = gp->dkg_gap1;
			smdpb->datapre = gp->dkg_gap2;
		} else {
			smdpb->idpre = 0;
			smdpb->datapre = 0;
		}
		smdpb->resv3 = 0;

#ifdef DSKBUG
		if (debuglevel & 1)
			printf("Defining Disk Parameters.\n");
#endif DSKBUG

		/* issue the command and check for errors */
		if (sglcmd(xpb, device->md_mc)) {
			printf("Define Disk Parameter Command Failed.\n");
		}
	}

#ifdef GNB
/*	Assume it's OK -> add to dk info for iostat */
	if (device->md_dk >= 0 && device->md_dk < DK_NDRIVE) {	/* Is OK */
		/* bytes/sec = bytes/sector * rev/sec * nsectors */
		/*	interleave = 1 */
		dk_bps[device->md_dk] = DEV_BSIZE * 60 * gp->dkg_nsect;
	}
#endif GNB

	/* release resources */
	safefree((caddr_t)xpb);
	return(1);
}

/*******************************************************************************
*									       *
*	Subroutine:		rfopen					       *
*									       *
*	Calling Sequence:	rfopen(dev, rw)				       *
*									       *
*	Called by:		UNIX I/O system 			       *
*									       *
*	Calling Parameters:	dev - major and minor device number	       *
*	                        rw - open for reading and/or writing	       *
*									       *
*	Local Variables:	unit - the drive number                        *
*									       *
*	Calls Subroutines:	None                               	       *
*									       *
*	Public/Global Variables:None					       *
*									       *
*	Description:							       *
*		Configure each drive on the first open.                        *
*		                                                               *
*******************************************************************************/

rfopen(dev, rw)
dev_t dev;
int rw;
{
	register int unit = UNIT(dev);
	REGISTER struct mb_device *device;

#ifdef DSKBUG
	if (debuglevel&1)
		printf("rfopen: dev=%x, unit=%x\n", dev, unit);
#endif DSKBUG

	/* does the unit make sense and is the drive alive */
	if (unit >= NRF || (device = rfdrive_info[unit]) == NULL ||
	    !device->md_alive) {
#ifdef DSKBUG
		if (debuglevel&1) /* 1.15b */
			printf("rfopen: bad device or drive not alive!\n");
#endif DSKBUG
		return(ENXIO);
	}

	/* may want to check for write protected drive someday */
	/* Sure do. */
	return(0);
}

/*******************************************************************************
*									       *
*	Subroutine:		rfstrategy				       *
*									       *
*	Calling Sequence:	rfstrategy(bp)				       *
*									       *
*	Called by:		UNIX I/O system 			       *
*									       *
*	Calling Parameters:	bp - UNIX buffer pointer		       *
*									       *
*	Local Variables:	unit - drive number   			       *
*	                        part - UNIX logical drive partition            *
*	                        sizeofxfer - number of blocks to xfer          *
*	                        blksleft - number of blocks left in part.      *
*	                        mbinfo - returned by mbsetup                   *
*	                        failbp - set if bp request is invalid          *
*	                        s - keeps old priority level                   *
*	                        secnum - logical sector number                 *
*	                        stdpb - standard parameter block for r/w       *
*									       *
*	Calls Subroutines:	None                               	       *
*									       *
*	Public/Global Variables:None					       *
*									       *
*	Description:							       *
*		Check validity of request                                      *
*		                                                               *
*******************************************************************************/

static int nqbuf[NRF];

void
rfstrategy(bp)
REGISTER struct buf *bp;
{

	REGISTER struct mb_device *device;
	REGISTER struct mb_ctlr *ctlr;
	register int unit = UNIT(bp->b_dev);
	register int part = PARTITION(bp->b_dev);
	register long sizeofxfer;
	register long blksleft;	/* block in partition from b_blkno to end */
	int mbinfo;
	int failbp;
	register int s;			/* "save priority" */
	register long secnum;
	PARMBLK stdpb;
#ifdef RFSTATS
	IORQ_STAT *iorqptr;
#endif

#ifdef DSKBUG
	if (debuglevel&1)
		printf("rfstrategy: dev=%o flags=%x bn=%d ma=%x count=%x\n",
			minor(bp->b_dev), bp->b_flags, bp->b_blkno,
			RF_ADDR(bp->b_un.b_addr), bp->b_bcount);
#endif DSKBUG

	/* Check validity of transfer request */
	if (unit >= NRF)
		failbp = 1;
	else
		failbp = 0;

	/* is the unit alive? */
	if ((device = rfdrive_info[unit]) == (struct mb_device *)NULL ||
	     device->md_alive == 0)
		failbp = 1;

	/* check the drive geometry */
	if (!geometry[unit].dkg_nsect || !geometry[unit].dkg_ncyl)
		failbp = 1;

	/* 1.9g */
	/* number of blocks requested */
	sizeofxfer = btodb(bp->b_bcount + DEV_BSIZE - 1);

	/* 1.9d changes */

	/* compute blocks remaining in parition */
	blksleft = partitions[unit][part].dkl_nblk - bp->b_blkno;

	if (blksleft < 0) {
		/* no good -- transfer starts outside of parition */
		failbp = TRUE;
		}
	else if (blksleft == 0) {
		/* EOF -- request starts at end of parition */
		bp->b_resid = bp->b_bcount;
		iodone(bp);
		return;
		}
	else if (sizeofxfer > blksleft) {
		/* hit end of partition -- trim transfer to available size */
		bp->b_resid = (sizeofxfer - blksleft) * DEV_BSIZE;
		sizeofxfer = blksleft;
		}
	else
		bp->b_resid = 0;


	/* if for any reason we failed so far do not execute this bp */
	if (failbp) {
		/* notify the system */
		bp->b_flags |= B_ERROR;
		bp->b_error = EIO;
		bp->b_resid = bp->b_bcount;
		iodone(bp);
		return;
	}

#ifndef SunOS4
	/* map memory to main bus */
	mbinfo = mbsetup(device->md_hd, bp, intsvc() ? MB_CANTWAIT : 0);

	/* if we didn't get it, try again later */
	if (!mbinfo) {
		timeout(rfstrategy, bp, hz);
		return;
	}
#else
	/* map memory to main bus */
	mbinfo = mbsetup(device->md_hd, bp, 0);
#endif

	++nqbuf[device->md_ctlr];

	/* save mbinfo for release in rfintr */
	bp->b_mbinfo = (struct buf *)mbinfo;

	/* Compute logical starting sector number */
	secnum = bp->b_blkno + (partitions[unit][part].dkl_cylno *
		geometry[unit].dkg_nsect * geometry[unit].dkg_nhead);

	stdpb.id = (dword)bp;
	if (bp->b_flags & B_PHYS)
		stdpb.control = 1;
	else if ((bp->b_bcount % 4096) == 0)	/* Multiple of 4K? */
		stdpb.control = 2;
	else
		stdpb.control = 3;

	/* If this is a dual-ported drive and the release flag is
	 * set, add the release bit to the parameter block 1.15c */
#ifdef ENABLE_DUAL_PORT
	if (device->md_flags & MDF_RELEASE) stdpb.control |= RF_RLS;
#endif

	stdpb.unit = (device->md_slave * 2) + VOL(device->md_flags);
	stdpb.addrmod = VME_ADD_MOD;
	stdpb.diskaddr = secnum;
	stdpb.reserved = 0;

#ifdef RFSTATS
	if (iorqsi >= niorqs) {
		printf("rfstrategy: WRAP!");
		iorqsi = 0;
	}
	iorqptr = &iorq.iorqs[iorqsi];
	iorqptr->iorq_time = time;
	iorqptr->iorq_bn = secnum;
	iorqptr->iorq_rq = bp->b_flags & B_READ;
	iorqptr->iorq_ns = sizeofxfer > 255 ? 255 : sizeofxfer;
	iorqptr->iorq_type = IORQ_START;
	iorqsi++;
#endif
	/* Physical memory mapped for us, issue normal I/O unless we
	 * need to do a transfer of size other than the sector size
	 * (thank you, swap routines!)
	 */
	stdpb.blkcount = sizeofxfer;
	if (bp->b_bcount % DEV_BSIZE) { /* 1.10d */
		REGISTER SGHEADER *sglist;
		/* if we can't get the memory resources we need for this
		 * request, we better try again later
		 */
		if((sglist = (SGHEADER *)safealloc(sizeof(SGHEADER)))
				== NULL) {
			mbrelse(device->md_hd, &bp->b_mbinfo);
			timeout(rfstrategy, bp, hz);
			return;
		}
		bp->b_sglist = (struct buf *)sglist;
		stdpb.command = (bp->b_flags & B_READ) ? C_SREAD : C_GWRITE;
		stdpb.vmeaddr = RF_ADDR(sglist);
		sglist->next = -1;	/* this is the only one */
		sglist->desc[0].bytes = bp->b_bcount;
		sglist->desc[0].sdptr = MBI_ADDR(mbinfo);
		sglist->desc[1].bytes = 0;  /* ignore rest of header */
	} else {
		stdpb.command = (bp->b_flags & B_READ) ? C_READ : C_WRITE;
		stdpb.vmeaddr = MBI_ADDR(mbinfo); /* 1.10e */
		bp->b_sglist = NULL;
	}

	ctlr = device->md_mc;
	/* Add this buffer to the RF 3200 command list */
	s = splx(pritospl(ctlr->mc_intpri));

	/* The following linking information isn't really used in the driver
	 * but it aids UNIX with finding incore() information. */
	if (debuglevel & 0x20) {
	bp->b_forw = (struct buf *) NULL;
	if (ctlr->mc_tab.b_actf == (struct buf *) NULL) {
		ctlr->mc_tab.b_actf = bp;
		bp->b_back = (struct buf *) NULL;
	} else {
		ctlr->mc_tab.b_actl->b_forw = bp;
		bp->b_back = ctlr->mc_tab.b_actl;
	}
	ctlr->mc_tab.b_actl = bp;
	}

	/* Start transfer */
	putpb(&stdpb, device);
	splx(s);

} /* rfstrategy */

/*******************************************************************************
*									       *
*	Subroutine:		rfintr					       *
*									       *
*	Calling Sequence:	rfintr()				       *
*									       *
*	Called by:		UNIX interrupt mechanism.		       *
*									       *
*	Calling Parameters:	None					       *
*									       *
*	Local Variables:	None					       *
*									       *
*	Calls Subroutines:	None                               	       *
*									       *
*	Public/Global Variables:None					       *
*									       *
*	Description:							       *
*		Complete the disk transfer.                                    *
*		                                                               *
*******************************************************************************/

rfintr(controller)
int controller;
{
	REGISTER struct mb_ctlr *mb_ctlr;
	REGISTER STATBLK *sb;
	REGISTER struct buf *bp;
	register int unit;
	STATBLK *getsb();
	register int nserv = 0;
	int	part;
#ifdef RFSTATS
	IORQ_STAT	*iorqptr;
#endif

#ifdef DSKBUG
	if (debuglevel&1)
		printf("rfintr: %x\n", controller);
#endif DSKBUG

	/* Is this controller alive */
	if (controller < 0 || controller >= NRFC) {
		printf("DEAD(%x)", controller);
		return(0);
	}
	if ((mb_ctlr = rfcontroller_info[controller]) == (struct mb_ctlr *)NULL) {
		printf("INFO(%x)", controller);
		return(0);
	}

/* Read a controller register to force the CPU to release IACK
 * on the VMEbus */
unit = ((RF32REG *)mb_ctlr->mc_addr)->status;

	/* Get the status block(s) which may have caused this interrupt */
	while ((sb = getsb(mb_ctlr)) != (STATBLK *)NULL) {

		/* this interrupt was for us */
		nserv++;
#ifdef DSKBUG
		if (debuglevel&2)
			printf("status %x err %x id %x data %x\n", sb->status,
				sb->errcode, sb->id, sb->errdata);
#endif DSKBUG

		/* check for general board error */
		if (sb->id == (unsigned)0xffffffff) {
			rferr(sb,FALSE);
			continue;
		}

		/* Get the buffer pointer for this status block */
		bp = (struct buf *)sb->id;
		unit = UNIT(bp->b_dev);
		part = PARTITION(bp->b_dev);

		/* What does this status block say? */
		if (sb->status & ST_RTY) {

			/* Retry required */
#ifdef DSKBUG
		if (debuglevel&1)
			printf("rfintr: retry required, dev=0x%x, block=%d\n",
					bp->b_dev, bp->b_blkno + 
					(partitions[unit][part].dkl_cylno *
					geometry[unit].dkg_nsect *
					geometry[unit].dkg_nhead));
#endif DSKBUG
		} /* retry required */

		if (sb->status & ST_COR) {

			/* Correction required */
			/* 1.8a&b 1.9d */
			if (bp != &ioctlbuf)
				printf("rfintr: correction required, unit=%d, block=%d\n",
					unit, bp->b_blkno +
					(partitions[unit][part].dkl_cylno *
					geometry[unit].dkg_nsect *
					geometry[unit].dkg_nhead));
		}  /* correction required */

		if (sb->status & ST_CC) {

			/* command completed, get UNIX buf for this command */
			if (bp == NULL)
				panic("rfintr: Disk buffer gone!");
		
			/* check for errors */
			if (sb->status & ST_ERR) {

				/* Error condition */

				/* 1.8a&b */

				/* for ioctl commands, put the errdata
				 * in bp->b_resid and do not print out the
				 * error information.  It is the responsibility
				 * of the applications program to report ioctl
				 * errors.  For r/w determine
				 * how many blocks did not transfer.
				 */
				/* 1.9c */
				if (bp == &ioctlbuf) {
					bp->b_resid = (long) sb->errdata;
					bp->b_error = (0xff & sb->errcode);
				} else {
					printf("rfintr: Hard error! unit=%d, block=%d\n",
						unit, sb->errdata);
					rferr(sb,FALSE);
					bp->b_resid = bp->b_bcount;
					bp->b_error = EIO;
				}
				bp->b_flags |= B_ERROR;
			}
			else {

				/* Transfer complete */

				/* 1.9c -- moved setting b_resid to strategy
				 * for successful completion
				 */

				bp->b_error = 0;
			}

			/* Keep some statistics (no difference between
			 * buffer response and controller active times).
			 */
			/* later */

			/* release the Main bus resource, if it was allocated */
			if (bp != &ioctlbuf) {
				/* ioctl's don't use mbinfo */
#ifdef RFSTATS
				if (iorqsi >= niorqs) {
					printf("rfintr: WRAP!");
					iorqsi = 0;
				}
				iorqptr = &iorq.iorqs[iorqsi++];
				iorqptr->iorq_time = time;
				iorqptr->iorq_bn = bp->b_blkno +
					(partitions[unit][part].dkl_cylno *
					 geometry[unit].dkg_nsect *
					 geometry[unit].dkg_nhead);
				iorqptr->iorq_type = IORQ_FINISH;
#endif
				/* 1.10d */
				mbrelse(rfdrive_info[UNIT(bp->b_dev)]->md_hd,
					&bp->b_mbinfo);

				/* Free the scatter/gather list if there
				 * is one
				 */
				if (bp->b_sglist != NULL)
					safefree((caddr_t)bp->b_sglist);

				--nqbuf[controller];
			}

			/* 1.10c */
			if (bp->b_flags & B_DONE) {
			    /* This can, unfortunately, happen for
			     * C_INTERROGATE with some firmware... */
#ifndef DSKBUG
			    if (bp != &ioctlbuf)
#else
			    if (debuglevel & 0x40 || bp != &ioctlbuf)
#endif 
				printf("rfintr: Two sb's for one command!!!\n");
			    nserv--;
			    }
			else

			    /* tell upper half we are done */
			    iodone(bp);

#ifdef GNB
			if(!nqbuf[controller]) {  /* Mark drive inactive */
				register int dk = rfdrive_info[unit]->md_dk;

				if (dk >= 0 && dk < DK_NDRIVE) {
					dk_busy &= ~(1 << dk);
				}
			}
#endif
		}  /* command complete */

	}  /* while there are status blocks to process */

	/* if this interrupt had nothing to do, and the last one for
	   this controller handled one or fewer requests, then something
	   is probably out of kilter. */
     /*	if (nserv == 0 && nservint[controller] <= 1)
	        printf("rfintr: no sb for rfc%d, last served%d\n", controller,
		nservint[controller]); */
	nservint[controller] = nserv;

	return(nserv);

} /* rfintr */

/*******************************************************************************
*									       *
*	Subroutine:		rimfsize					       *
*									       *
*	Calling Sequence:	rimfsize(dev)				       *
*									       *
*	Called by:		UNIX I/O system 			       *
*									       *
*	Calling Parameters:	dev - major and minor device number	       *
*									       *
*	Local Variables:	device - pointer to device information         *
*									       *
*	Calls Subroutines:	None                               	       *
*									       *
*	Public/Global Variables:None					       *
*									       *
*	Description:							       *
*		This routine returns the size of the requested partition in    *
*		blocks.                                                        *
*		                                                               *
*******************************************************************************/

daddr_t
rimfsize(dev)
dev_t dev;
{
	REGISTER struct mb_device *device = rfdrive_info[UNIT(dev)];

#ifdef DSKBUG
	if (debuglevel & 1)
		printf("rimfsize:\n");
#endif DSKBUG

	if (device == NULL || device->md_alive == 0) {
#ifdef DSKBUG
		if (debuglevel & 1) /* 1.15b */
			printf("rimfsize: device not alive\n");
#endif DSKBUG
		return(-1);
	}
	return(partitions[device->md_unit][PARTITION(dev)].dkl_nblk);
}

/*******************************************************************************
*									       *
*	Subroutine:		rfioctl					       *
*									       *
*	Calling Sequence:	rfioctl(dev, ioctlcmd, addr, flag)             *
*									       *
*	Called by:		UNIX I/O system 			       *
*									       *
*	Calling Parameters:	dev - major & minor device number              *
*				ioctlcmd - the i/o command to perform          *
*				addr - command data address in user memory     *
*				flag - ??????????                              *
*									       *
*	Local Variables:	None					       *
*									       *
*	Calls Subroutines:	None                               	       *
*									       *
*	Public/Global Variables:None					       *
*									       *
*	Description:							       *
*		Some thing happens                                             *
*		                                                               *
*******************************************************************************/

/*ARGSUSED*/
rfioctl(dev, ioctlcmd, addr, flag)
dev_t	dev;
int	ioctlcmd, flag;
caddr_t addr;
{
	register int unit = UNIT(dev);
	REGISTER struct mb_device *device = rfdrive_info[unit];
	struct mb_ctlr *mbctlr; /* 1.15a */
	int	s, group;
	int	error;
	int	retval;
	struct dk_info	information;
#ifdef RFSTATS
	IORQS	tiorq;
	int	i;
#endif

#ifdef DSKBUG
	if (debuglevel&1)
		printf("rfioctl: dev %x cmd %x addr %x flags %x\n",
			dev, ioctlcmd, addr, flag);
#endif DSKBUG

	/* Don't let normal users mess with this stuff */
	if (!suser()) {
		return(EPERM);
	}

	/* only allow one ioctl at a time */
	s = splx(pritospl(device->md_mc->mc_intpri));

	while (ioctlbuf.b_flags & B_BUSY) {
		ioctlbuf.b_flags |= B_WANTED;
		sleep((caddr_t)&ioctlbuf.b_flags, PRIBIO);
	}

	/* initialize this UNIX buffer */
	ioctlbuf.b_flags |= B_BUSY;
	ioctlbuf.b_flags &= ~(B_ERROR | B_DONE);
	ioctlbuf.b_error = 0;
	ioctlbuf.b_dev = dev;
	
	/* assume bad command number */
	retval = EINVAL;

	/* what needs to be done */
	switch(ioctlcmd) {

	case DKIOCFMT:		/* Format track(s) */
		/* 1.9c */
		retval = format(unit, (struct dk_fmt *)addr, &ioctlbuf);
		break;

	case DKIOCGGEOM:	/* Get drive configuration */
		*(struct dk_geom *)addr = geometry[unit];
		retval = 0;
		break;

	case DKIOCSGEOM:	/* Set a drives configuration */

		/* Reconfigure the drive */
		retval = config(unit, (struct dk_geom *)addr, &ioctlbuf);

		if (retval == 0) {
			/* success, save this geometry */
			geometry[unit] = *(struct dk_geom *)addr;
		}
		break;

	case DKIOCGPART:	/* Get partition info */

		/* just copy what we have */
		*(struct dk_map *)addr = partitions[unit][PARTITION(dev)];
		retval = 0;
		break;

	case DKIOCSPART:	/* Set partition info */

		/* change our copy */
		partitions[unit][PARTITION(dev)] = *(struct dk_map *)addr;
		retval = 0;
		break;

	case DKIOCINFO:		/* Get information */

		/* 1.15a */
		mbctlr = rfcontroller_info[rfdrive_info[unit]->md_ctlr];
		information.dki_ctlr = getdevaddr(mbctlr->mc_addr);
		information.dki_unit = rfdrive_info[unit]->md_slave << 1 +
					VOL(rfdrive_info[unit]->md_flags) - 1;
		information.dki_ctype = DKC_RF3200;
		information.dki_flags = DKI_MAPTRK;
		*(struct dk_info *)addr = information;
		retval = 0;
		break;

	case DKIOCVFY:		/* Verify disk blocks */

		retval = verify(unit, (struct dk_vfy *)addr, &ioctlbuf);
		break;

	case DKIOCMAP:		/* Map track(s) or sector(s) */

		retval = maptrack(unit, (struct dk_mapr *)addr, &ioctlbuf);
		break;

	case DKIOCSLP:		/* Slip sector */

		retval = slip(unit, (struct dk_slip *)addr, &ioctlbuf);
		break;

	case DKIOCSDEBUG:	/* set driver debug level */
		debuglevel = *(int *) addr;
		retval = 0;
		break;

	case DKIOCGDEBUG:	/* get driver debug level */
		*(int *)addr = debuglevel;
		retval = 0;
		break;

	case DKIOCGSTAT:	/* get board statistics */
		retval = getstat(unit, (BOARDST *)addr, &ioctlbuf);
		break;

	case DKIOCSSMD:		/* define SMD parameters */
 
                retval = defsmd(unit, (struct smd_param *)addr, &ioctlbuf);
                break;
 
        case DKIOCGSMD:         /* return configuration */
                retval = retconf(unit, (RETCONF *)addr, &ioctlbuf);
                break;
 
	case DKIOCGTUNE:	/* get board tuning statistics */
		retval = tune(unit, (TUNEST *)addr, 0);
		break;

	case DKIOCSTUNE:	/* set board tuning statistics */
		retval = tune(unit, (TUNEST *)addr, 1);
		break;

	case DKIOCGIOCG:	/* Get I/O control group info */
		group = ((struct iocgrp *)addr)->group;
		if (group < 1 || group > MAXIOCG)
			retval = EINVAL;
		else {
			/* 1.16a */
			*(struct iocgrp *)addr = iocg[device->md_ctlr][group];
			retval = 0;
		}
		break;

	case DKIOCSIOCG:	/* Set I/O control group info */

		/* Modify current I/O control settings */
		retval = setiocg(unit, (struct iocgrp *)addr, &ioctlbuf);
		break;

	case DKIOCGIORQN:	/* Get # of IO Request Statistics */

#ifdef	RFSTATS
		/* just copy what we have */
		*(int *)addr = iorqsi;
		retval = 0;
#else
		retval = EINVAL;
#endif
		break;

	case DKIOCIORQC:	/* Clear IO Request Statistics */

#ifdef	RFSTATS
		/* Just set our index back to zero. */
		iorqsi = 0;
		retval = 0;
#else
		retval = EINVAL;
#endif
		break;

        case DKIOCIDTRAK:         /* read track of id's */
		retval = readtrkid(unit, (TRAKID **)addr, &ioctlbuf);
                break;
 
        case DKIOCPHYSID:         /* read id of one sector */
		retval = readphysectid(unit, (TRAKID *)addr, &ioctlbuf);
                break;
 
	case DKIOCRDDAT:	/* read defect data from track	*/
		retval = readddat(unit,(struct dk_rdb *)addr, &ioctlbuf);
		break;

	case DKIOCRDMAP:	/* read defect map */
		retval = readmap(unit, (DEFECTMAP *)addr, &ioctlbuf);
		break;

	case DKIOCGIORQS:	/* Get an IORQ_STAT Request */

#ifdef	RFSTATS
		tiorq = *(IORQS *) addr;
		i = tiorq.which;
		((IOR(p *)addr)->iorq = iorq.iorqs[i];
		retval = 0;
#else
		retval = EINVAL;
#endif
		break;

	case DKIOCGBTYP:	/* Get Board TYPe (3200 or 3400) */
		*(int *)addr = rfctype[device->md_ctlr] >> 8;
		retval = 0;
		break;

	case DKIOCGBID:		/* Get Board IDentification */
                retval = retident(device, (RETID *)addr);
                break;

	case DKIOCSMDCHR:	/* return SMD characteristics (interrogate) */
		retval = smdchr(device, (SMDCHRS *)addr, &ioctlbuf);
		break;

	case DKIOCRESDI:	/* return SMD characteristics (interrogate) */
		retval = retesdi(device, (ESDIPARAMS *)addr, &ioctlbuf);
		break;
	}

	/* release the UNIX buffer used for this request */
	ioctlbuf.b_flags &= ~B_BUSY;

	/* wakeup anyone waiting for it */
	if (ioctlbuf.b_flags & B_WANTED) {
		wakeup(&ioctlbuf.b_flags);
		}

	splx(s);
	return(retval);
}

/*******************************************************************************
*									       *
*	Subroutine:		rfread					       *
*									       *
*	Calling Sequence:	rfread(dev, uio)			       *
*				rfwrite(dev, uio)      		     	       *
*									       *
*	Called by:		UNIX I/O system (Raw device)                   *
*									       *
*	Calling Parameters:	dev - major & minor device number	       *
*				uio - user's IO request		     	       *
*									       *
*	Local Variables:	None					       *
*									       *
*	Calls Subroutines:	rfstrategy (via physio)          	       *
*									       *
*	Public/Global Variables:None					       *
*									       *
*	Description:							       *
*		Handle unbuffered reads and writes to the disk.                *
*		                                                               *
*******************************************************************************/

rfread(dev, uio)
dev_t dev;
struct uio *uio;
{
#ifdef DSKBUG
	if (debuglevel&1)
		printf("rfread:\n");
#endif DSKBUG

	if (UNIT(dev) >= NRF)
		return(ENXIO);
	return(physio(rfstrategy, &rrfbuf[UNIT(dev)], dev, B_READ, minphys, uio));
}

rfwrite(dev, uio)
dev_t dev;
struct uio *uio;
{
#ifdef DSKBUG
	if (debuglevel&1)
		printf("rfwrite:\n");
#endif DSKBUG

	if (UNIT(dev) >= NRF)
		return(ENXIO);
	return(physio(rfstrategy, &rrfbuf[UNIT(dev)], dev, B_WRITE, minphys, uio));
}

/* Using the system's default minphys

rfminphys(bp)
struct buf *bp;
{
	if (bp->b_bcount > MAX_RF_BSIZE)
		bp->b_bcount = MAX_RF_BSIZE;
}
*/

/*******************************************************************************
*									       *
*	Subroutine:		sglcmd					       *
*									       *
*	Calling Sequence:	sglcmd(xpb, ctlr)			       *
*									       *
*	Called by:		rfprobe, rfslave, rfattach                     *
*									       *
*	Calling Parameters:	xpb - extended RF 3200 parameter block         *
*				ctlr - Pointer to mb_ctlr structure	       *
*									       *
*	Local Variables:	pb - address of the extended parameter block   *
*									       *
*	Calls Subroutines:	None                               	       *
*									       *
*	Public/Global Variables:None					       *
*									       *
*	Description:							       *
*		A type 0 command is sent to the RF 3200.  The address of the   *
*		extended parameter block is written to the address buffer port *
*		and a 0 is written to the channel attention port.  The status  *
*		block is then polled until the command completes.              *
*		                                                               *
*******************************************************************************/

sglcmd(xpb, ctlr)
REGISTER EXTPB *xpb;
struct mb_ctlr *ctlr;
{
	REGISTER RF32REG *rf = (RF32REG *) ctlr->mc_addr;
	register int waitfor;
	dword	pb;
	word	lsb, msb;
	int finished;		/* save completion status	*/

	/* Clear the status block */
	xpb->sb.id = 0;
	xpb->sb.status = 0;
	xpb->sb.errcode = 0;
	xpb->sb.errextra = 0;
	xpb->sb.errdata = 0;

	/* Wait until we can write to address buffer port */
	/* this should test toggling of the busy bit */
	waitfor = SWAB(bsybit[ctlr->mc_ctlr]);
	/* 1.8e */
#ifdef TOOLONG
	CDELAY((finished = (rf->status & waitfor) == waitfor), TOOLONG);
	if (!finished) {
		printf("sglcmd: timeout ... BSY\n");
		printf("sglcmd: status port = %x\n", rf->status);
		rferr(&xpb->sb,TRUE);
		return(1);
	}
#else TOOLONG
	while ((rf->status & waitfor) != waitfor)
		;
#endif TOOLONG

#ifdef DSKBUG
	if (debuglevel & 0x10)
		printf("sglcmd: status port = %x\n", rf->status);
#endif DSKBUG

	/* Tell the RF3200 our system characteristics */
#ifdef DSKBUG
	if (debuglevel & 2) {
		printf("sglcmd: control and address modifiers = %x\n",
			SWAB(CNTRL << 8 | VME_ADD_MOD));
	}
#endif DSKBUG

	/* Write the control and address modifier to the address buffer port */
	rf->addrbuf = SWAB(CNTRL << 8 | VME_ADD_MOD);

#ifdef DSKBUG
	if (debuglevel & 2)
		printf("sglcmd: status port = %x\n", rf->status);
#endif DSKBUG

	/* convert extpb pointer to a long for shifting */
	pb = RF_ADDR(xpb);
	msb = SWAB(pb >> 16 & 0xffff);
	lsb = SWAB(pb & 0xffff);
#ifdef DSKBUG
	if (debuglevel & 0x10)
		printf("sglcmd: pb=%x msb=%x lsb=%x\n", pb, msb, lsb);
#endif DSKBUG

	/* write the msb and lsb of the pb to the address buffer port */
	rf->addrbuf = msb;
	rf->addrbuf = lsb;

	/* Get the controller's attention */
	rf->attention = 0;

	/* Toggle our busy bit, as the controller should do when it finishes
	 * the command. */
	bsybit[ctlr->mc_ctlr] ^= 1;

	/* Wait for the command to complete */
#ifdef DSKBUG
	if (debuglevel & 0x10) {
		printf("sglcmd: Waiting for ST_CC\n");
		printf("sglcmd: status port = %x\n", rf->status);
		dumpepb((caddr_t)xpb);
	}
#endif DSKBUG

	/* 1.8e */
#ifdef TOOLONG
	CDELAY((xpb->sb.status & ST_CC),TOOLONG)
	if ((xpb->sb.status & ST_CC) == 0) {
		/* 1.13c */
		printf("rf: sglcmd: timeout waiting for command complete (ST_CC)\n");
		printf("rf: sglcmd: command = %x, status port = %x\n",
			xpb->pb.command, rf->status);
		rferr(&xpb->sb,FALSE);
		return(1);
	}
#else TOOLONG
	while ((xpb->sb.status & ST_CC) == 0)
		;
#endif TOOLONG

	/* check for an error */
	if (xpb->sb.status & ST_ERR) {
		REGISTER struct buf *bp;

		bp = (struct buf *)xpb->sb.id;

		/* Error condition */

		/* 1.8a&b */

		/* for ioctl commands, put the errdata
		 * in bp->b_resid and do not print out the
		 * error information.  It is the responsibility
		 * of the applications program to report ioctl
		 * errors.  For r/w determine
		 * how many blocks did not transfer.
		 */
		/* 1.9c */
		if (bp == &ioctlbuf) {
			bp->b_resid = (long) xpb->sb.errdata;
			bp->b_error = (0xff & xpb->sb.errcode);
			bp->b_flags |= B_ERROR;
		} else {
			/* error found, dump status block */
			rferr(&xpb->sb,TRUE);
		}
		return(1);
	}
#ifdef DSKBUG
	if (debuglevel & 0x10)
		printf("sglcmd: return\n");
#endif DSKBUG
	/* no error */
	return(0);
}

/*******************************************************************************
*									       *
*	Subroutine:		format					       *
*									       *
*	Calling Sequence:	format(unit, fmt, bp)			       *
*									       *
*	Called by:		rfioctl                                        *
*									       *
*	Calling Parameters:	                                               *
*		unit - device unit number 				       *
*		fmt - the format request    			               *
*		bp - UNIX buffer pointer    			               *
*									       *
*	Local Variables:	    					       *
*		device - drive information       			       *
*		geom - disk geometry used for format                           *
*		fmtpb - parameter block for format command                     *
*									       *
*	Calls Subroutines:	putpb                              	       *
*									       *
*	Public/Global Variables:None					       *
*									       *
*	Description:							       *
*		Format the drive for the given number of tracks.               *
*		                                                               *
*******************************************************************************/

format(unit, fmt, bp)
int unit;
struct dk_fmt *fmt;
struct buf *bp;
{
	struct mb_device *device = rfdrive_info[unit];
	struct dk_geom *geom = &geometry[device->md_unit];
	FMTVERPB fmtpb;
	struct dk_fmt ufmt_rq;

#ifdef DSKBUG
	if (debuglevel & 1)
		printf("format: unit=%d\n", unit);
#endif DSKBUG

	/* If the number of sectors for this unit is 0, no configuration */
	if (geom->dkg_nsect == 0) {
		printf("%s%d: drive geometry not set!\n", rfcdriver.mdr_dname,
			device->md_unit);
		return(ENXIO);
	}

	/* Get the user format request */
	ufmt_rq = *(struct dk_fmt *)fmt;

	/* be sure this is for full tracks only */
	if ((ufmt_rq.dkf_blkno % geom->dkg_nsect) != 0 ||
	    (ufmt_rq.dkf_nblk  % geom->dkg_nsect) != 0) {
		printf("%s%d: must format full tracks!\n",
			rfcdriver.mdr_dname, device->md_unit);
		return(ENXIO);
	}

	/* create a format command */
	fmtpb.id = (dword)bp;
	fmtpb.command = C_FORMAT;
	fmtpb.control = 0;
	fmtpb.unit = (device->md_slave << 1) + VOL(device->md_flags);
	fmtpb.diskaddr = (dword)ufmt_rq.dkf_blkno;
	fmtpb.blkcount = (dword)ufmt_rq.dkf_nblk;
	fmtpb.vmeaddr = 0;
	fmtpb.addrmod = 0;

	/* issue command to the controller */
	putpb(&fmtpb, device);

	/* wait for the command to finish */
	iowait(bp);
	 
	/* if an error occured the bad block is in b_resid */
	ufmt_rq.dkf_error = bp->b_error;  /* 1.9c */
	ufmt_rq.dkf_badblk = (daddr_t) bp->b_resid;
	/* And now copy the verify information back. */
	*(struct dk_fmt *) fmt = ufmt_rq;

	return(0);
}

/*******************************************************************************
*									       *
*	Subroutine:		config					       *
*									       *
*	Calling Sequence:	config(unit, geom, bp)			       *
*									       *
*	Called by:		rfioctl                                        *
*									       *
*	Calling Parameters:	                                               *
*		unit - device unit number				       *
*		geom - disk geometry pointer			               *
*		bp - UNIX buffer pointer    			               *
*									       *
*	Local Variables:	    					       *
*		device - drive information     			               *
*		conf - configuration parameters			               *
*									       *
*	Calls Subroutines:	putpb                              	       *
*									       *
*	Public/Global Variables:None					       *
*									       *
*	Description:							       *
*		Configure the given drive according to the specified drive     *
*		configuration information.                                     *
*		                                                               *
*******************************************************************************/

config(unit, ugeom_rq, bp)
int unit;
struct dk_geom *ugeom_rq;
struct buf *bp;
{
	REGISTER struct mb_device *device = rfdrive_info[unit];
	CONFPB	conf;
	DEFSMDPB defsmdpb;
	RETCONF confparams;
	struct dk_geom geom;

	/* Get the user's new geometry. */
	geom = *(struct dk_geom *) ugeom_rq;

	/* Set up a configuration parameter block */
	conf.id = (dword)bp;
	conf.command = C_SETCONFIG;
	conf.control = 0;
	conf.unit = (device->md_slave << 1) + VOL(device->md_flags);
	conf.resv1 = 0;
	conf.cyldsk = geom.dkg_ncyl + geom.dkg_acyl;
	conf.bytsec = 512;			/* always 512 byte sectors */
	if (rfctype[device->md_ctlr] == STATRF3200)
		conf.basehead = geom.dkg_bhead;
	else
		conf.basehead = 0;	/* STATRF3400 */
	conf.headcyl = geom.dkg_nhead;
	conf.sectrk = geom.dkg_nsect;
#ifdef NSPARES
	conf.nspares = geom.dkg_apc / geom.dkg_nhead;
#else
	conf.nspares = geom.dkg_extra[0] / geom.dkg_nhead;
#endif
	if (rfctype[device->md_ctlr] == STATRF3200)
		conf.flags = geom.dkg_extra[XTCFF];
	else
		conf.flags = 0;	/* STATRF3400 */

#ifdef ENABLE_DUAL_PORT
	if (device->md_flags & MDF_DUALPORTED) {
		conf.flags |= (CONFEDP|CONFBSY);
	}
#endif

	conf.resv2[0] = 0;
	conf.resv2[1] = 0;
	conf.resv2[2] = 0;

	/* queue the command */
	putpb(&conf, device);

	/* wait for completion */
	iowait(bp);

	bp->b_flags &= ~(B_ERROR | B_DONE);
	/* Get whatever params the drive already has. */
	if (retconf(unit, &confparams, bp)) {
		/* ERROR */
#ifdef DSKBUG
		printf("rf: Couldn't get drive configuration\n");
#endif
		confparams.interleave = 1;
		confparams.headskew = 3;
		confparams.cylskew = 5;
		confparams.hgskew = 0;
	}

	bp->b_flags &= ~(B_ERROR | B_DONE);
	/* Set up a define SMD parameter block */
	defsmdpb.id = (dword)bp;
	defsmdpb.resv1 = 0;
	defsmdpb.unit = (device->md_slave << 1) + VOL(device->md_flags);
	defsmdpb.control = 0;
	defsmdpb.command = C_SETSMDPARAM;
	defsmdpb.interleave = confparams.interleave;
	defsmdpb.headskew = confparams.headskew;
	defsmdpb.cylskew = confparams.cylskew;
	defsmdpb.hgskew = confparams.hgskew;
	defsmdpb.recovery = 0;
	/* This seems wrong to me.  Why reset to the default values at
	   this point?    87/9/29 -J.E.Lind   1.8f
	if (rfctype[device->md_ctlr] == STATRF3400) {
		defsmdpb.idpre = geom.dkg_gap1;
		defsmdpb.datapre = geom.dkg_gap2;
	} else {
		defsmdpb.idpre = 0;
		defsmdpb.datapre = 0;
	}  */
	/* We'll just use the geometry values we just got.  Why not? 1.8f */
	defsmdpb.idpre = geom.dkg_gap1;
	defsmdpb.datapre = geom.dkg_gap2;

	defsmdpb.resv3 = 0;

	/* queue the command */
	putpb(&defsmdpb, device);

	/* wait for completion */
	iowait(bp);

	/* return error status */
	return(bp->b_error ? EIO : 0);
}

/*******************************************************************************
*									       *
*	Subroutine:		smdchr					       *
*									       *
*	Calling Sequence:	smdchr(unit, chrs, bp)			       *
*									       *
*	Called by:		rfioctl                                        *
*									       *
*	Calling Parameters:	                                               *
*		unit - device unit number				       *
*		u_smdchr - buffer to return characteristics		       *
*		bp - UNIX buffer pointer    			               *
*									       *
*	Local Variables:	    					       *
*		device - drive information     			               *
*		charp - pointer to DVMA buffer for characteristics from cont.  *
*									       *
*	Calls Subroutines:	putpb                              	       *
*									       *
*	Public/Global Variables:None					       *
*									       *
*	Description:							       *
*		Issue Interrogate Drive command and return results to user     *
*		                                                               *
*******************************************************************************/

smdchr(device, u_smdchr, bp)
REGISTER struct mb_device *device;
SMDCHRS *u_smdchr;
struct buf *bp;
{
	SMDCHRS *charp;
	PARMBLK pb;

	/* get some memory to DMA the characteristics into */
	charp = (SMDCHRS *)safealloc(sizeof(SMDCHRS));
	if (charp == NULL) {
		return(ENOMEM);
	}

	/* set up interrogate command for controller */
	pb.id = (dword)bp;
	pb.command = C_INTERROGATE;
	pb.unit = (device->md_slave << 1) + VOL(device->md_flags);
	pb.control = 0;
	pb.addrmod = VME_ADD_MOD;
	pb.vmeaddr = RF_ADDR(charp);
	pb.reserved = 0;
	pb.diskaddr = 0;
	pb.blkcount = 0;

	/* queue the command */
	putpb(&pb, device);

	/* wait for completion */
	iowait(bp);

	/* copy characteristics into where the user wants it */
	*u_smdchr = *charp;

	/* free DVMA memory */
	safefree((caddr_t)charp);

	/* return error status */
	return(bp->b_error ? EIO : 0);
} /* smdchr */

/*******************************************************************************
*									       *
*	Subroutine:		retesdi					       *
*									       *
*	Calling Sequence:	retesdi(unit, u_parm, bp)		       *
*									       *
*	Called by:		rfioctl                                        *
*									       *
*	Calling Parameters:	                                               *
*		unit - device unit number				       *
*		u_parm - buffer to return characteristics		       *
*		bp - UNIX buffer pointer    			               *
*									       *
*	Local Variables:	    					       *
*		device - drive information     			               *
*		charp - pointer to DVMA buffer for characteristics from cont.  *
*									       *
*	Calls Subroutines:	putpb                              	       *
*									       *
*	Public/Global Variables:None					       *
*									       *
*	Description:							       *
*		Issue Interrogate Drive command and return results to user     *
*		                                                               *
*******************************************************************************/

retesdi(device, u_parm, bp)
REGISTER struct mb_device *device;
ESDIPARAMS *u_parm;
struct buf *bp;
{
	ESDIPARAMS *parms;
	PARMBLK pb;

	/* get some memory to DMA the characteristics into */
	parms = (ESDIPARAMS *)safealloc(sizeof(ESDIPARAMS));
	if (parms == NULL) {
		return(ENOMEM);
	}

	/* set up interrogate command for controller */
	pb.id = (dword)bp;
	pb.command = C_GETESDICONF;
	pb.unit = (device->md_slave << 1) + VOL(device->md_flags);
	pb.control = 0;
	pb.addrmod = VME_ADD_MOD;
	pb.vmeaddr = RF_ADDR(parms);
	pb.reserved = 0;
	pb.diskaddr = 0;
	pb.blkcount = 0;

	/* queue the command */
	putpb(&pb, device);

	/* wait for completion */
	iowait(bp);

	/* copy characteristics into where the user wants it */
	*u_parm = *parms;

	/* free DVMA memory */
	safefree((caddr_t)parms);

	/* return error status */
	return(bp->b_error ? EIO : 0);
}

/*******************************************************************************
*									       *
*	Subroutine:		setiocg	         			       *
*									       *
*	Calling Sequence:	setiocg(unit, iocgptr, bp)		       *
*									       *
*	Called by:		rfioctl                                        *
*									       *
*	Calling Parameters:	                                               *
*		unit - device unit number				       *
*		iocgptr - i/o control group pointer		               *
*		bp - UNIX buffer pointer    			               *
*									       *
*	Local Variables:	    					       *
*		device - drive information     			               *
*									       *
*	Calls Subroutines:	putpb                              	       *
*									       *
*	Public/Global Variables:None					       *
*									       *
*	Description:							       *
*		Changes the I/O control group settings.                        *
*		                                                               *
*******************************************************************************/

setiocg(unit, iocgptr, bp)
int unit;
struct iocgrp *iocgptr;
struct buf *bp;
{
	REGISTER struct mb_device *device = rfdrive_info[unit];
	IOCGPB	iocgpb;
	struct iocgrp niocg;

	/* Get the users new group. */
	niocg = *(struct iocgrp *)iocgptr;
	/* Is the group requested within bounds? */
	if (niocg.group < 1 || niocg.group > MAXIOCG)
		return(EINVAL);

	/* Set up a configuration parameter block */
	iocgpb.id = (dword)bp;
	iocgpb.command = C_SETIOCG;
	iocgpb.control = (word)niocg.group;
	iocgpb.unit = 0; 
	iocgpb.resv1 = 0;
	iocgpb.cache = (word)niocg.cache;
	iocgpb.readahead = (byte)niocg.readahead;
	iocgpb.recovery = (byte)niocg.recovery;
	iocgpb.resv2 = 0;
	iocgpb.dretry = (byte)niocg.dretry;
	iocgpb.ndretry = (byte)niocg.ndretry;
	iocgpb.resv3 = 0;

	/* queue the command */
	putpb(&iocgpb, device);

	/* wait for completion */
	iowait(bp);

	/* return error status */
	if ((bp->b_flags & B_ERROR) == 0) {
		/* Save the information. */
		iocg[device->md_ctlr][niocg.group] = niocg; /* 1.16a */
	}
	return(bp->b_error ? EIO : 0);
}

/*******************************************************************************
*									       *
*	Subroutine:		verify					       *
*									       *
*	Calling Sequence:	verify(unit, ver, bp)			       *
*									       *
*	Called by:		rfioctl                                        *
*									       *
*	Calling Parameters:	                                               *
*		device - main bus device pointer			       *
*		ver - the verify request    			               *
*		bp - UNIX buffer pointer    			               *
*									       *
*	Local Variables:	    					       *
*		device - drive information       			       *
*		verpb - parameter block for verify command                     *
*									       *
*	Calls Subroutines:	putpb                              	       *
*									       *
*	Public/Global Variables:None					       *
*									       *
*	Description:							       *
*		Verify the specified drives format.                            *
*		                                                               *
*******************************************************************************/

verify(unit, uver_rq, bp)
int unit;
struct dk_vfy *uver_rq;
struct buf *bp;
{
	REGISTER struct mb_device *device = rfdrive_info[unit];
	FMTVERPB verpb;
	struct dk_vfy ver;

#ifdef DSKBUG
	if (debuglevel & 1)
		printf("verify: unit=%x\n", unit);
#endif DSKBUG

	/* Get the users verify request. */
	ver = *(struct dk_vfy *) uver_rq;

	/* create a verify command */
	verpb.id = (dword)bp;
	verpb.command = C_VERIFY;
	verpb.control = 0;
#ifdef ENABLE_DUAL_PORT
	if (device->md_flags & MDF_RELEASE) verpb.control |= RF_RLS;
#endif
	verpb.unit = (device->md_slave << 1) + VOL(device->md_flags);
	verpb.diskaddr = (dword)ver.dkv_blkno;
	verpb.blkcount = (dword)ver.dkv_nblk;
	verpb.vmeaddr = 0;
	verpb.addrmod = 0;

	/* issue command to the controller */
	putpb(&verpb, device);

	/* wait for the command to finish */
	iowait(bp);
	 
	/* if an error occured the bad block is in b_resid */
	ver.dkv_error = bp->b_error;
	ver.dkv_badblk = (daddr_t) bp->b_resid;
	/* And now copy the verify information back. */
	*(struct dk_vfy *) uver_rq = ver;

	return(0);
}

/*******************************************************************************
*									       *
*	Subroutine:		maptrack				       *
*									       *
*	Calling Sequence:	maptrack(unit, maprequest, bp)		       *
*									       *
*	Called by:		rfioctl                                        *
*									       *
*	Calling Parameters:	                                               *
*		unit - the device unit number                                  *
*		maprequest - the mapping request		               *
*		bp - UNIX buffer pointer    			               *
*									       *
*	Local Variables:	    					       *
*		device - drive information       			       *

*		gp - pointer to disk geometry  			               *
*		mappb - parameter block for mapping commands                   *
*		count - number of sectors or tracks to map                     *
*		next - 1: map sector, sec/track: map track                     *
*		error - did an error occur on any map request                  *
*									       *
*	Calls Subroutines:	putpb                              	       *
*									       *
*	Public/Global Variables:None					       *
*									       *
*	Description:							       *
*		Process a user's track mapping request.                        *
*		                                                               *
*******************************************************************************/

maptrack(unit, u_maprequest, bp)
int unit;
REGISTER struct dk_mapr *u_maprequest;
struct buf *bp;
{
	REGISTER struct mb_device *device = rfdrive_info[unit];
	REGISTER struct dk_geom *gp = &geometry[device->md_unit];
	MAPPB mappb;
	long count;
	int next;
	word error;
	struct dk_mapr mapreq;

#ifdef DSKBUG
	if (debuglevel & 1)
		printf("maprequest: unit %x\n", unit);
#endif DSKBUG

	/* Get the user map request. */
	mapreq = *(struct dk_mapr *) u_maprequest;

	/* no errors yet */
	error = 0;

	/* Check validity of request */
	if (gp->dkg_nsect == 0) {
		printf("%s%d: drive geometry not set!\n", rfcdriver.mdr_dname,
			device->md_unit);
		return(EINVAL);
	}

	/* Set up mapping command */
	mappb.id = (dword)bp;
	mappb.control = 0;
	mappb.unit = (device->md_slave << 1) + VOL(device->md_flags);

	/* get the sector to map (first sector if mapping track) */
	mappb.baditem = (dword)mapreq.dkm_fblk;

	/* and the place to map it to */
	mappb.alternate = (dword)mapreq.dkm_tblk;

	/* clear the reserved fields of this command */
	mappb.resv1 = 0;
	mappb.resv2 = 0;

	/* set recover data field according to user specification */
	mappb.recover = mapreq.dkm_recover;

	/* Set command for either mapping sector or track:
	 * if the bad sector is the first on the track and
	 * the number of sectors to map is a multiple of track size
	 * then map tracks, otherwise map sectors
	 */
	if ((mapreq.dkm_fblk % gp->dkg_nsect) == 0 &&
	    (mapreq.dkm_nblk % gp->dkg_nsect) == 0) {

		/* map tracks - keep count in tracks */
#ifdef DSKBUG
		if (debuglevel & 4)
			printf("Mapping tracks\n");
#endif DSKBUG
		mappb.command = C_MAPTRACK;
		count = mapreq.dkm_nblk / gp->dkg_nsect;
		next = gp->dkg_nsect;
	}
	else {	/* map sectors - keep count in sectors */
#ifdef DSKBUG
		if (debuglevel & 4)
			printf("Mapping sectors\n");
#endif DSKBUG
		mappb.command = C_MAPSECTOR;
		count = mapreq.dkm_nblk;
		next = 1;
	}

	/* execute successive mapping commands until all sectors or tracks
	 * are mapped.  If an error occurs on any request stop.
	 */
	while (count-- > 0) {

		putpb(&mappb, device);

		/* wait for the command to finish */
		iowait(bp);

		/* check for an error */
		if (error = bp->b_error)
			break;

		/* clear done bit so we can execute again */
		bp->b_flags &= ~B_DONE;

		/* get next sectors to map */
		mappb.baditem += next;
		if (mapreq.dkm_tblk == -1)
			mappb.alternate = -1;
		else
			mappb.alternate += next;
	}

	/* pass back error information */
	mapreq.dkm_error = error;
	/* and the place it was mapped to */
	mapreq.dkm_tblk = mappb.alternate;
	*(struct dk_mapr *) u_maprequest = mapreq;
	return(0);
}

/*******************************************************************************
*									       *
*	Subroutine:		slip                                           *
*									       *
*	Calling Sequence:	slip(unit, slipr, bp)		               *
*									       *
*	Called by:		rfioctl                                        *
*									       *
*	Calling Parameters:	                                               *
*		unit - the device unit number                                  *
*		slipr - the sector slipping request		               *
*		bp - UNIX buffer pointer    			               *
*									       *
*	Local Variables:	    					       *
*		device - drive information       			       *
*		gp - pointer to disk geometry  			               *
*		mappb - parameter block for slip commands                      *
*									       *
*	Calls Subroutines:	putpb                              	       *
*									       *
*	Public/Global Variables:None					       *
*									       *
*	Description:							       *
*		Process a user's sector slipping request.                      *
*		                                                               *
*******************************************************************************/

slip(unit, u_slipr, bp)
int unit;
struct dk_slip *u_slipr;
struct buf *bp;
{
	REGISTER struct mb_device *device = rfdrive_info[unit];
	REGISTER struct dk_geom *gp = &geometry[device->md_unit];
	MAPPB mappb;
	struct dk_slip slipr;

#ifdef DSKBUG
	if (debuglevel & 1)
		printf("slip: unit %x\n", unit);
#endif DSKBUG

	/* Check validity of request */
	if (gp->dkg_nsect == 0) {
		printf("%s%d: drive geometry not set!\n", rfcdriver.mdr_dname,
			device->md_unit);
		return(EINVAL);
	}

	/* Get the user's request. */
	slipr = *(struct dk_slip *) u_slipr;

	/* Set up slip command */
	mappb.id = (dword)bp;
	mappb.command = C_SLIPSECTOR;
	mappb.control = 0;
	mappb.unit = (device->md_slave << 1) + VOL(device->md_flags);

	/* get the sector to slip */
	mappb.baditem = (dword)slipr.dks_fblk;
	mappb.alternate = 0;

	/* set recover data field according to user specification */
	mappb.recover = slipr.dks_recover;

	/* clear the reserved fields of this command */
	mappb.resv1 = 0;
	mappb.resv2 = 0;

	/* send command to the controller */
	putpb(&mappb, device);

	/* wait for the command to finish */
	iowait(bp);

	/* return any errors */
	slipr.dks_error = bp->b_error;
	*(struct dk_slip *) u_slipr = slipr;
	return(0);
}

/*******************************************************************************
*									       *
*	Subroutine:		getstat                                        *
*									       *
*	Calling Sequence:	getstat(unit, stats, bp)	               *
*									       *
*	Called by:		rfioctl                                        *
*									       *
*	Calling Parameters:	                                               *
*		unit - the device unit number                                  *
*		stats - the statistics request		                       *
*		bp - UNIX buffer pointer    			               *
*									       *
*	Local Variables:	    					       *
*		device - drive information       			       *
*		dmastats - parameter block for statistics commands             *
*									       *
*	Calls Subroutines:	putpb                              	       *
*									       *
*	Public/Global Variables:None					       *
*									       *
*	Description:							       *
*		Process a user's board statistics request.                     *
*		                                                               *
*******************************************************************************/

getstat(unit, u_stats, bp)
int unit;
BOARDST *u_stats;
struct buf *bp;
{
	STATPB pb;
	REGISTER struct mb_device *device = rfdrive_info[unit];
	BOARDST *dmastats;
	BOARDST stats;

	/* get some memory to DMA the statistics into */
	dmastats = (BOARDST *)safealloc(sizeof(BOARDST));
	if (dmastats == NULL) {
		printf("getstat: Cannot rmalloc statistics buffer!\n");
		return(ENOMEM);
	}

	/* Get the users statistic information so we know if we should clear */
	/* the statistics or not. */
	stats = *(BOARDST *) u_stats;

	/* set up statistics command for controller */
	/* clear all board statistics */
	pb.id = (dword)bp;
	pb.command = C_STATS;
	pb.unit = 0;
	pb.control = 0;
	pb.addrmod = VME_ADD_MOD;
	pb.vmeaddr = RF_ADDR(dmastats);
	pb.spare = 0;
	pb.resv = 0;
	pb.resv1 = 0;
	pb.clear = (byte)stats.commands;

        putpb(&pb, device);

        /* wait for the command to finish */
	iowait(bp);

	/* Was there an error? */
	if (bp->b_error)
		return(bp->b_error ? EIO : 0);
#ifdef DSKBUG
        if (debuglevel & 1) {
	printf("\nDMA Statistics are:");
	printf("Total number of commands executed:\t%u\n", dmastats->commands);
	printf("Total number of read commands:\t%u\n", dmastats->reads);
	printf("Total number of sectors read from disk:\t%u\n", dmastats->diskrd);
	printf("Total sectors read from cache:\t%u\n", dmastats->cacherd);
	printf("Total number of write commands:\t%u\n", dmastats->writes);
	printf("Total number of sectors written:\t%u\n", dmastats->diskwr);

	/* drive 0 statistics */
	printf("\nDrive 0:");
	printf("\tSeek errors: %d\tSMD errors: %d\tAlternate seeks: %d\n",
		dmastats->d1_seek, dmastats->d1_smdint, dmastats->d1_altseek);
	printf("\tGood ECC corrections: %d\tFailed corrections: %d\n",
		dmastats->d1_ecccor, dmastats->d1_corfail);
	printf("\tNon ECC data errors: %d\n", dmastats->d1_nonecc);

	/* drive 1 statistics */
	printf("\nDrive 1:");
	printf("\tSeek errors: %d\tSMD errors: %d\tAlternate seeks: %d\n",
		dmastats->d2_seek, dmastats->d2_smdint, dmastats->d2_altseek);
	printf("\tGood ECC corrections: %d\tFailed corrections: %d\n",
		dmastats->d2_ecccor, dmastats->d2_corfail);
	printf("\tNon ECC data errors: %d\n", dmastats->d2_nonecc);

	/* drive 2 statistics */
	printf("\nDrive 2:");
	printf("\tSeek errors: %d\tSMD errors: %d\tAlternate seeks: %d\n",
		dmastats->d3_seek, dmastats->d3_smdint, dmastats->d3_altseek);
	printf("\tGood ECC corrections: %d\tFailed corrections: %d\n",
		dmastats->d3_ecccor, dmastats->d3_corfail);
	printf("\tNon ECC data errors: %d\n", dmastats->d3_nonecc);

	/* drive 3 statistics */
	printf("\nDrive 3:");
	printf("\tSeek errors: %d\tSMD errors: %d\tAlternate seeks: %d\n",
		dmastats->d4_seek, dmastats->d4_smdint, dmastats->d4_altseek);
	printf("\tGood ECC corrections: %d\tFailed corrections: %d\n",
		dmastats->d4_ecccor, dmastats->d4_corfail);
	printf("\tNon ECC data errors: %d\n", dmastats->d4_nonecc);
	}
#endif DSKBUG

	/* copy statistics into where the user wants it */
	*u_stats = *dmastats;

	/* free DVMA memory */
	safefree((caddr_t)dmastats);

	/* get error info */
	return(0);
}

tune(unit, u_stats, flag)
int unit, flag;
TUNEST *u_stats;
{
	PARMBLK pb;
	REGISTER struct mb_device *device = rfdrive_info[unit];
	TUNEST *dmastats;

	/* get some memory to DMA the statistics into */
	dmastats = (TUNEST *)safealloc(sizeof(TUNEST));
	if (dmastats == NULL) {
		printf("tune: Cannot rmalloc statistics buffer!\n");
		return(1);
	}

	/* Set the tuning parameters */
	if (flag) {
		*dmastats = *u_stats;
	}

	/* set up statistics command for controller */
	/* clear all board statistics */
	pb.id = (dword)&ioctlbuf;
	pb.command = flag ? 0xB : 0xA;
	pb.unit = 0;
	pb.control = 0;
	pb.addrmod = VME_ADD_MOD;
	pb.vmeaddr = RF_ADDR(dmastats);
	pb.diskaddr = 0;
	pb.reserved = 0;
	pb.blkcount = 0;

        putpb(&pb, device);

        /* wait for the command to finish */
	iowait(&ioctlbuf);

	/* copy statistics into where the user wants it */
	if (!flag) {
		*u_stats = *dmastats;
	}

	/* free DVMA memory */
	safefree((caddr_t)dmastats);

	/* get error info */
	return(ioctlbuf.b_error ? EIO : 0);
}

/*******************************************************************************
*									       *
*	Subroutine:		defsmd                                         *
*									       *
*	Calling Sequence:	defsmd(unit, smd, bp)	                       *
*									       *
*	Called by:		rfioctl                                        *
*									       *
*	Calling Parameters:	                                               *
*		unit - the device unit number                                  *
*		smd - the statistics request		                       *
*		bp - UNIX buffer pointer    			               *
*									       *
*	Local Variables:	    					       *
*		device - drive information       			       *
*		defsmdpb - parameter block for define SMD parameter commands   *
*									       *
*	Calls Subroutines:	putpb                              	       *
*									       *
*	Public/Global Variables:None					       *
*									       *
*	Description:							       *
*		Process a user's define SMD parameter request.                 *
*		                                                               *
*******************************************************************************/

defsmd(unit, u_smd, bp)
int unit;
struct smd_param *u_smd;
struct buf *bp;
{
	REGISTER struct mb_device *device = rfdrive_info[unit];
	DEFSMDPB defsmdpb;
	struct smd_param smd;

	/* Get the SMD parameters from user memory. */
	smd = *(struct smd_param *) u_smd;

	/* Set up a define SMD parameter block */
	defsmdpb.id = (dword)bp;
	defsmdpb.resv1 = 0;
	defsmdpb.unit = (device->md_slave << 1) + VOL(device->md_flags);
	defsmdpb.control = 0;
	defsmdpb.command = C_SETSMDPARAM;
	defsmdpb.interleave = smd.smd_interleave;
	defsmdpb.headskew = smd.smd_headskew;
	defsmdpb.cylskew = smd.smd_cylskew;
	defsmdpb.hgskew = smd.smd_hgskew;
	if (rfctype[device->md_ctlr] == STATRF3200)
		defsmdpb.hgskew = 0;
	defsmdpb.recovery = smd.smd_recovery;
	defsmdpb.idpre = geometry[unit].dkg_gap1 = smd.smd_idpre;
	defsmdpb.datapre = geometry[unit].dkg_gap2 = smd.smd_datapre;
	if (rfctype[device->md_ctlr] != STATRF3200) {
		defsmdpb.idpre = geometry[unit].dkg_gap1 = smd.smd_idpre = 0;
		defsmdpb.datapre = geometry[unit].dkg_gap2 = smd.smd_datapre = 0;
	}
	defsmdpb.resv3 = 0;


	/* queue the command */
	putpb(&defsmdpb, device);

	/* wait for completion */
	iowait(bp);

	/* return error status */
	return(bp->b_error ? EIO : 0);
}

/*******************************************************************************
*									       *
*	Subroutine:		retconf                                        *
*									       *
*	Calling Sequence:	retconf(unit, retconf, bp)	               *
*									       *
*	Called by:		rfioctl                                        *
*									       *
*	Calling Parameters:	                                               *
*		unit - the device unit number                                  *
*		retconf - the return configuration request		       *
*		bp - UNIX buffer pointer    			               *
*									       *
*	Local Variables:	    					       *
*		device - drive information       			       *
*		conf - parameter block for return configuration commands       *
*									       *
*	Calls Subroutines:	putpb                              	       *
*									       *
*	Public/Global Variables:None					       *
*									       *
*	Description:							       *
*		Process a user's return configuration request.                 *
*		                                                               *
*******************************************************************************/

retconf(unit, u_retconf, bp)
int unit;
RETCONF *u_retconf;
struct buf *bp;
{
	PARMBLK pb;
	REGISTER struct mb_device *device = rfdrive_info[unit];
	RETCONF *conf;

	/* get some memory to DMA the configuration into */
	conf = (RETCONF *)safealloc(sizeof(RETCONF));
	if (conf == NULL) {
		printf("retconf: Cannot rmalloc configuration buffer!\n");
		return(ENOMEM);
	}

	/* set up return configuration  command for controller */
	pb.id = (dword)bp;
	pb.command = C_GETCONFIG;
	pb.unit = (device->md_slave << 1) + VOL(device->md_flags);
	pb.control = 0;
	pb.addrmod = VME_ADD_MOD;
	pb.vmeaddr = RF_ADDR(conf);
	pb.diskaddr = 0;
	pb.blkcount = 0;
	pb.reserved = 0;

        putpb(&pb, device);

        /* wait for the command to finish */
	iowait(bp);

	/*stopy configuration into where the user wants it */
	*(RETCONF *)u_retconf = *conf;

	/* free DVMA memory */
	safefree((caddr_t)conf);

	/* get error info */
	return(bp->b_error ? EIO : 0);
}

 
/*******************************************************************************
*									       *
*	Subroutine:		readphysectid				       *
*									       *
*	Calling Sequence:	readphysectid(unit, trakid, bp)		       *
*									       *
*	Called by:		rfioctl					       *
*									       *
*	Calling Parameters:						       *
*		unit - the device unit number				       *
*		trakid - the return id structure			       *
*		bp - UNIX buffer pointer    			               *
*									       *
*	Local Variables:	    					       *
*		device - drive information       			       *
*		conf - parameter block for return configuration commands       *
*									       *
*	Calls Subroutines:	putpb                              	       *
*									       *
*	Public/Global Variables:None					       *
*									       *
*	Description:							       *
*		Process a user's return configuration request.                 *
*		                                                               *
*******************************************************************************/

readphysectid(unit, u_trakid, bp)
int unit;
TRAKID *u_trakid;
struct buf *bp;
{
	long index;
	PARMBLK pb;
	REGISTER struct mb_device *device = rfdrive_info[unit];
	REGISTER struct dk_geom *gp;
	TRAKID *tid;
	int trksize;

	/* geometry point used to compute disk address */
	gp = &geometry[device->md_unit];

	trksize = sizeof(TRAKID)*(gp->dkg_nsect + gp->dkg_apc/gp->dkg_nhead +
			(gp->dkg_extra[XTCFF] & CONFSSP));

	/* get some memory to DMA the IDs from the track into */
	tid = (TRAKID *)safealloc(trksize);
	if (tid == NULL) {
		printf("rf: Cannot rmalloc track id buffer!\n");
		return(ENOMEM);
	}

	/* set up read track of ID's command for controller */
	pb.id = (dword)bp;
	pb.command = C_READTID;
	pb.unit = (device->md_slave << 1) + VOL(device->md_flags);
	pb.control = 0;
	pb.addrmod = VME_ADD_MOD;
	pb.vmeaddr = RF_ADDR(tid);
	pb.diskaddr = (u_trakid->cyl * gp->dkg_nhead + u_trakid->head)
			 * gp->dkg_nsect;
	pb.blkcount = 0;
	pb.reserved = 0;
        putpb(&pb, device);

        /* wait for the command to finish */
	iowait(bp);

	/* index into the array the location of the bad spot */
	index=u_trakid->sec;
	/* copy the correct block where the user wants it */
	*(TRAKID *)u_trakid = tid[index];

	/* free DVMA memory */
	safefree((caddr_t)tid);

	/* get error info */
	return(bp->b_error ? EIO : 0);
}
 
/*******************************************************************************
*									       *
*	Subroutine:		readtrkid                                      *
*									       *
*	Calling Sequence:	readtrkid(unit, trakid, bp)	               *
*									       *
*	Called by:		rfioctl                                        *
*									       *
*	Calling Parameters:	                                               *
*		unit - the device unit number                                  *
*		trakid - the return id structure			       *
*		bp - UNIX buffer pointer    			               *
*									       *
*	Local Variables:	    					       *
*		device - drive information       			       *
*		tid - pointer to DVMA memory for controller to writ into       *
*									       *
*	Calls Subroutines:	putpb                              	       *
*									       *
*	Public/Global Variables:None					       *
*									       *
*	Description:							       *
*		Process a user's return configuration request.                 *
*		                                                               *
*******************************************************************************/

readtrkid(unit, u_trakid, bp)
int unit;
TRAKID **u_trakid;
struct buf *bp;
{
	PARMBLK pb;
	REGISTER struct mb_device *device = rfdrive_info[unit];
	REGISTER struct dk_geom *gp;
	TRAKID *tid;
	int trksize;
	TRAKID inspec;
	int retval;

	if (retval = copyin(*u_trakid, &inspec, sizeof (TRAKID))) {
		return (retval);
		}

	/* geometry pointer used to compute disk address */
	gp = &geometry[device->md_unit];

	trksize = sizeof(TRAKID)*(gp->dkg_nsect + gp->dkg_apc/gp->dkg_nhead +
			(gp->dkg_extra[XTCFF] & CONFSSP));

	/* get some memory to DMA the ID's into */
	tid = (TRAKID *)safealloc(trksize);
	if (tid == NULL) {
		printf("rf: Cannot allocate track id buffer!\n");
		return(ENOMEM);
	}

	/* set up read track of ID's command for controller */
	pb.id = (dword)bp;
	pb.command = C_READTID;
	pb.unit = (device->md_slave << 1) + VOL(device->md_flags);
	pb.control = 0;
	pb.addrmod = VME_ADD_MOD;
	pb.vmeaddr = RF_ADDR(tid);
	pb.diskaddr = (inspec.cyl * gp->dkg_nhead + inspec.head)
			 * gp->dkg_nsect;
	pb.blkcount = 0;
	pb.reserved = 0;
        putpb(&pb, device);

        /* wait for the command to finish */
	iowait(bp);

	/* copy the information where the user wants it */
	if (retval = copyout(tid, *u_trakid, trksize)) {
		safefree((caddr_t)tid);
	    	return(retval);
	}

	/* free DVMA memory */
	safefree((caddr_t)tid);

	/* get error info */
	return(bp->b_error ? EIO : 0);
}

/*******************************************************************************
*									       *
*	Subroutine:		readmap                                        *
*									       *
*	Calling Sequence:	readmap(unit, u_defectmap, bp)	               *
*									       *
*	Called by:		rfioctl                                        *
*									       *
*	Calling Parameters:	                                               *
*		unit - the device unit number                                  *
*		u_defectmap - the defect map for the specified unit	       *
*		bp - UNIX buffer pointer    			               *
*									       *
*	Local Variables:	    					       *
*		pb - parameter block for this request			       *
*		device - drive information       			       *
*		conf - parameter block for return configuration commands       *
*									       *
*	Calls Subroutines:	putpb                              	       *
*									       *
*	Public/Global Variables:None					       *
*									       *
*	Description:							       *
*		Process a user's return configuration request.                 *
*		                                                               *
*******************************************************************************/

readmap(unit,u_defectmap, bp)
int unit;
DEFECTMAP  *u_defectmap;
struct buf *bp;
{
	PARMBLK pb;
	REGISTER struct mb_device *device = rfdrive_info[unit];
	DEFECTMAP *conf;

	/* get some memory to DMA the configuration into */
	conf = (DEFECTMAP *)safealloc(sizeof(DEFECTMAP));
	if (conf == NULL) {
		printf("defectmap: Cannot rmalloc Defect buffer!\n");
		return(ENOMEM);
	}

	/* set up return configuration  command for controller */
	pb.id = (dword)bp;
	pb.command = C_READMAP;
	pb.unit = (device->md_slave << 1) + VOL(device->md_flags);
	pb.control = 0;
	pb.addrmod = VME_ADD_MOD;
	pb.vmeaddr = RF_ADDR(conf);
	pb.diskaddr = 0;
	pb.blkcount = 0;
	pb.reserved = 0;
/*	printf(" about to put onto pb list\n"); */
        putpb(&pb, device);

        /* wait for the command to finish */
	iowait(bp);

	/* copy the correct block where the user wants it */
	*u_defectmap = *conf;

	/* free DVMA memory */
	safefree((caddr_t)conf);

	/* get error info */
	return(bp->b_error ? EIO : 0);
} /* readmap */

/*******************************************************************************
*									       *
*	Subroutine:		readddat                                       *
*									       *
*	Calling Sequence:	readddat(unit, u_defectdat, bp)	               *
*									       *
*	Called by:		rfioctl                                        *
*									       *
*	Calling Parameters:	                                               *
*		unit - the device unit number                                  *
*		u_defectdat - the defect data for the specified track	       *
*		bp - UNIX buffer pointer    			               *
*									       *
*	Local Variables:	    					       *
*		pb - parameter block for this request			       *
*		device - drive information       			       *
*		conf - parameter block for return configuration commands       *
*									       *
*	Calls Subroutines:	putpb                              	       *
*									       *
*	Public/Global Variables:None					       *
*									       *
*	Description:							       *
*		Process a user's return configuration request.                 *
*		                                                               *
*******************************************************************************/

readddat(unit,u_defectdat, bp)
int unit;
struct dk_rdb  **u_defectdat;
struct buf *bp;
{
	FMTVERPB pb;
	REGISTER struct mb_device *device = rfdrive_info[unit];
	struct dk_geom *gp;
	struct dk_rdb *rddb;
	int retval;

	gp = &geometry[device->md_unit];

	/* get some memory to DMA the defect data block into */
	rddb = (struct dk_rdb *)safealloc(sizeof(struct dk_rdb));
	if (rddb == NULL) {
		printf("defectlist: Cannot rmalloc Defect buffer!\n");
		return(ENOMEM);
	}

	/* copy the user io parameter into the iobpmap area */
	if (retval = copyin(*u_defectdat, rddb, sizeof (struct dk_rdb))) {
		safefree((caddr_t)rddb);
	    	return(retval);
	}

	/* set up return defect data command for controller */
	pb.id = (dword)bp;
	pb.command = C_READDEFECT;
	pb.unit = (device->md_slave << 1) + VOL(device->md_flags);
	pb.control = 0;
	pb.addrmod = VME_ADD_MOD;
	pb.vmeaddr = RF_ADDR(&rddb->db);
	if (rfctype[device->md_ctlr] == STATRF3200)
		pb.diskaddr = (rddb->db.smd.cylinder *gp->dkg_nhead +
			rddb->db.smd.head) * gp->dkg_nsect;
	else
		pb.diskaddr = rddb->db.esdi.header.cyllow << 16 |
			rddb->db.esdi.header.head;
	pb.blkcount = 0;

	/* put the command in the queue */
        putpb(&pb, device);

        /* wait for the command to finish */
	iowait(bp);

	/* Pass the error code back to the applications program. */
	rddb->dkd_error = bp->b_error;  /* 1.9c */

	/* copy the iobpmap area back into the user io parameter */
	if (retval = copyout(rddb, *u_defectdat, sizeof (struct dk_rdb))) {
		safefree((caddr_t)rddb);
	    	return(retval);
	}

	/* Free DVMA memory */
	safefree((caddr_t)rddb);

	/* We always return the error code in the ioctl structure, so there
	 * is no point in returning an error here.
	 */
	return(0);
}

/*******************************************************************************
*									       *
*	Subroutine:		retident                                       *
*									       *
*	Calling Sequence:	retident(unit, retident, bp)	               *
*									       *
*	Called by:		rfioctl                                        *
*									       *
*	Calling Parameters:	                                               *
*		unit - the device unit number                                  *
*		retident - the return identification structure		       *
*		bp - UNIX buffer pointer    			               *
*									       *
*	Local Variables:	    					       *
*		device - drive information       			       *
*		pb - parameter block for identification request		       *
*		ident - identification status block for return information     *
*									       *
*	Calls Subroutines:	putpb                              	       *
*									       *
*	Public/Global Variables:None					       *
*									       *
*	Description:							       *
*		Process a user's return identification request.                *
*		                                                               *
*******************************************************************************/

static
/*ARGSUSED*/
retident(device, u_retident)
REGISTER struct mb_device *device;
RETID *u_retident;
{
	REGISTER EXTPB *xpb;
	REGISTER PARMBLK *pb;
	REGISTER RETID *ident;

	/* get an extended parameter block */
	xpb = (EXTPB *)safealloc(sizeof(EXTPB));

	/* initialize extended parameter block - disable interrupts */
	xpb->intr = 0;
	xpb->errintr = 0;
	xpb->resv = 0;

	/* build the Identify command */
	pb = (PARMBLK *)&xpb->pb;
	pb->id = 0;
	pb->command = C_IDENTIFY;
	pb->control = 0;
	pb->unit = 0;
	pb->addrmod = VME_ADD_MOD;
	pb->diskaddr = 0;
	pb->vmeaddr = 0;
	pb->blkcount = 0;
	pb->reserved = 0;

	/* issue identify command to see if a drive is there */
	if (sglcmd(xpb, device->md_mc)) {
		/* error occured, assume no drive there */
#ifdef DSKBUG
		printf("retident: Cannot identify controller %d\n",
			device->md_mc);
#endif DSKBUG
		safefree((caddr_t)xpb);
		return(1);
	}

	/* success */

	/* copy identification into the user structure */
	*(STATBLK *)u_retident = xpb->sb;
	safefree((caddr_t)xpb);
	return(0);
}


/*******************************************************************************
*									       *
*	Subroutine:		putpb					       *
*									       *
*	Calling Sequence:	putpb(pb, rf)				       *
*									       *
*	Called by:		lots            			       *
*									       *
*	Calling Parameters:	pb - the parameter block to send               *
*									       *
*	Local Variables:	rf - pointer to RF 3200 hardware ports         *
*									       *
*	Calls Subroutines:	None                               	       *
*									       *
*	Public/Global Variables:					       *
*		iopb[].io_start - set to current time			       *
*									       *
*	Description:							       *
*		Place a parameter block in the command list and notify the     *
*		controller.                                                    *
*		                                                               *
*******************************************************************************/

putpb(pb, device)
REGISTER PARMBLK *pb;
REGISTER struct mb_device *device;
{
	REGISTER CMDLIST *cmdptr;
	register int i;
#ifdef STANDALONE	/* 1.15d */
        static EXTPB *xpb = NULL;     /* extended paramter block for
					 executing commands */
 
	if (xpb == NULL)
		xpb = (EXTPB *)safealloc(sizeof(EXTPB));
        bzero(xpb, sizeof(EXTPB));
        bcopy(pb, xpb, sizeof(PARMBLK));
        return(sglcmd(xpb, device->md_mc));
#else

	cmdptr = cmdq[device->md_ctlr];
#ifdef DSKBUG
	if (debuglevel & 4) {
		printf("putpb: pbin=%d, pbout=%d\n", cmdptr->pbin, cmdptr->pbout);
		printf("putpb: sbin=%d, sbout=%d\n", cmdptr->sbin, cmdptr->sbout);
		dumppb((caddr_t)pb);
	}
#endif DSKBUG

#ifndef NPBMASK
	while ((cmdptr->pbin + 1) % RF32NPB == cmdptr->pbout) {
#else NPBMASK
	while (((cmdptr->pbin + 1) & NPBMASK) == cmdptr->pbout) {
#endif NPBMASK
		/* 1.9e */
		/* the list is full - sleep until a status block returns.  */
#ifdef NOHANG
		waitsb = TRUE;
		sleep((caddr_t)&waitsb, PRIBIO);
#endif NOHANG
		}

	i = cmdptr->pbin;
	cmdptr->pblist[i] = *pb;

#ifndef NPBMASK
	cmdptr->pbin = (i + 1) % RF32NPB;
#else NPBMASK
	cmdptr->pbin = (i + 1) & NPBMASK;
#endif NPBMASK

	/* Tell the controller we've added a pb */
#ifdef DSKBUG
	if (debuglevel & 4) {
		printf("putpb: channel attention\n");
	}
#endif DSKBUG
	/* we think for perfmeter	*/
	if(device->md_dk >= 0)
		dk_xfer[device->md_dk]++;
#ifdef GNB
	if ((i = device->md_dk) >= 0) {
		dk_busy |= 1 << i;
		dk_seek[i]++;
		dk_wds[i] += pb->blkcount << 3;	/* 64 byte chunks? */
	}
#endif

	((RF32REG *) device->md_mc->mc_addr)->attention = SWAB(HOST_ID);
#ifdef DSKBUG
	if (debuglevel & 1) {
		printf("Started controller %d\n", device->md_ctlr);
		printf("Cmdlist at %x\n", cmdptr);
	}
#endif

#ifdef DSKBUG
	if ((debuglevel & 0xFF) == 0xFF) {
		for (i = 0; i < 100000; i++);
		for (i = 0; i < 100000; i++);
		printf("pb in: %d, out: %d\n", cmdptr->pbin, cmdptr->pbout);
		printf("sb in: %d, out: %d\n", cmdptr->sbin, cmdptr->sbout);
	}
#endif DSKBUG
#endif STANDALONE
}

/*******************************************************************************
*									       *
*	Subroutine:		getsb					       *
*									       *
*	Calling Sequence:	getsb()					       *
*									       *
*	Called by:		rfintr	        			       *
*									       *
*	Calling Parameters:						       *
*		ctlr	- The mb_ctlr structure that we should get a status    *
*			  block from.					       *
*									       *
*	Local Variables:	tsb - temporary (holding) status block         *
*									       *
*	Calls Subroutines:	None                               	       *
*									       *
*	Public/Global Variables:None					       *
*									       *
*	Description:							       *
*		Copy the status block from the command list and return a       *
*		pointer to it.  If there is no status block return NULL.       *
*		                                                               *
*******************************************************************************/

STATBLK *
getsb(ctlr)
REGISTER struct mb_ctlr *ctlr;
{
	static STATBLK tsb[NRFC];
	REGISTER CMDLIST *cmdptr;

	cmdptr = cmdq[ctlr->mc_ctlr];
#ifdef DSKBUG
	if (debuglevel & 8)
		printf("getsb: in=%d, out=%d\n", cmdptr->sbin, cmdptr->sbout);
#endif DSKBUG
	if (cmdptr->sbin >= RF32NSB) {
		/* 1.8c */
		dumpsb((caddr_t)&cmdptr->sblist[cmdptr->sbout]);
		panic("getsb: sbin pointer out of range!\n");
	}
	if (cmdptr->sbout >= RF32NSB) {
		/* 1.8c */
		dumpsb((caddr_t)&cmdptr->sblist[cmdptr->sbout]);
		panic("getsb: sbout pointer out of range!\n");
	}

	if (cmdptr->sbout == cmdptr->sbin)

		/* List empty, this should never have been called */
		return ((STATBLK *)NULL);

	tsb[ctlr->mc_ctlr] = cmdptr->sblist[cmdptr->sbout];
	cmdptr->sbout++;

	/* 1.9e */
	/* check to see if anyone was sleeping until a sb returns */
#ifdef NOHANG
	if (waitsb) {
	    waitsb = FALSE;
	    wakeup((caddr_t)(&waitsb));
	    }
#endif NOHANG

#ifndef NSBMASK
	cmdptr->sbout %= RF32NSB;
#else NSBMASK
	cmdptr->sbout &= NSBMASK;
#endif NSBMASK
	return (&tsb[ctlr->mc_ctlr]);
}

/*******************************************************************************
*									       *
*	Subroutine:		rferr                                          *
*									       *
*	Calling Sequence:	rferr(sb,printit)                              *
*									       *
*	Called by:		this driver                                    *
*									       *
*	Calling Parameters:	sb - pointer to status block from controller   *
*				printit - force printing of SB		       *
*				                                               *
*	Local Variables:	None                                           *
*				                                               *
*	Calls Subroutines:	printf                             	       *
*				                                               *
*	Public/Global Variables:None					       *
*				                                               *
*	Description:							       *
*		Displays the status block which caused an error.               *
*		                                                               *
*******************************************************************************/

rferr(sb,printit)
STATBLK *sb;
int printit;
{
        REGISTER struct buf *bp;

	bp = (struct buf *)sb->id;
	/* When the status block id is null, we are booting or dumping core.
	 * Otherwise it will always be set to a buffer header. */
	if (bp != (struct buf *) NULL || printit) {
		printf("id=%x, status=%x, error=%x, extra=%x %x\n",
			sb->id, sb->status, sb->errcode, sb->errextra, sb->errdata);
#ifdef ERRFILE
		if (sb->errcode <= MAXERRCODE)
			printf("Error was: %s\n", rfmsg[sb->errcode]);
		else
			printf("Unexpected error code %x!\n",sb->errcode);
#endif ERRFILE
	}
} /* rferr */

/*******************************************************************************
*									       *
*	Subroutine:		rfdump                                         *
*									       *
*	Calling Sequence:	rfdump(dev, addr, blkno, nblk)                 *
*									       *
*	Called by:		The UNIX panic/dump routines                   *
*									       *
*	Calling Parameters:	dev - major and minor device of swap           *
*				addr - starting address of click to dump       *
*				blkno - starting block at which to dump it     *
*				nblk - number of disk blocks to dump           *
*				                                               *
*	Local Variables:	xpb - extended paramter block for write        *
*				pb - point to paramters in above               *
*				unit - unit number (drive number)              *
*				device - system device description             *
*				                                               *
*	Calls Subroutines:	None                             	       *
*				                                               *
*	Public/Global Variables:DVMA					       *
*				                                               *
*	Description:							       *
*		Called for each core click of the system which paniced.        *
*		Writes out the mapped page to the device at the block          *
*		specified.                                                     *
*		                                                               *
*******************************************************************************/

rfdump(dev, addr, blkno, nblk)
dev_t dev;
caddr_t addr;
daddr_t blkno, nblk;
{
	EXTPB xpb;	/* extended paramter block for executing commands */
	REGISTER PARMBLK *pb;
	register int unit = UNIT(dev);
	register int part = PARTITION(dev);
	REGISTER struct mb_device *device;

	/* does the unit make sense and is the drive alive */
	if (unit >= NRF || (device = rfdrive_info[unit]) == NULL ||
			!device->md_alive) { /* 1.16k */
		return(ENXIO);
	}

	/* check to see if we are asking for a blk past the end	*/
	if(blkno > partitions[unit][part].dkl_nblk ||
	    (blkno + nblk) > partitions[unit][part].dkl_nblk) {
		return(EINVAL);
	}

	/* initialize extended parameter block - disable interrupts */
	xpb.intr = 0;
	xpb.errintr = 0;
	xpb.resv = 0;

	/* Setup and Interrogate the drive command */
	pb = (PARMBLK *)&xpb.pb;
	pb->id = 0;
	pb->command = C_WRITE;
	pb->control = 0;
	pb->unit = (device->md_slave << 1) + VOL(device->md_flags);
	pb->addrmod = VME_ADD_MOD;
	pb->diskaddr = blkno + (partitions[unit][part].dkl_cylno *
		geometry[unit].dkg_nsect * geometry[unit].dkg_nhead);
	pb->vmeaddr = RF_ADDR(addr-DVMA);
	pb->blkcount = nblk;
	pb->reserved = 0;

	/* issue the write command and check for errors */
	if (sglcmd(&xpb, device->md_mc)) {
		printf("rfdump: write failed from block %d for %d blocks\n",
			blkno, nblk);
		return(EIO);
	}
        return(0);
}

#ifdef DSKBUG
/* dump the extended parameter block */
static
dumpepb(c)
caddr_t c;
{
	int i, j;

	printf("Extended parameter block dump:\n");
	for (j = 0; j < 2; j++) {
		for (i = 0; i < 20; i++)
			printf(" %x", (0xff & *c++));
		printf("\n");
	}
}

static
dumppb(c)
caddr_t c;
{
	int i;

	printf("Standard Parameter block:\n");
	for (i = 0; i < 20; i++)
		printf(" %x", 0xff & (*c++) );
	printf("\n");
}
#endif

static
dumpsb(c)
caddr_t c;
{
	int i;

	printf("Status block:\n");
	for (i = 0; i < 12; i++)
		printf(" %x", *c++);
	printf("\n");
}

/*******************************************************************************
*									       *
*	Subroutine:		safealloc,safefree                             *
*									       *
*	Calling Sequence:	safealloc(size)				       *
*				safefree(ptr)				       *
*									       *
*	Called by:		rfprobe, rfslave, rfattach, rfstrategy and     *
*				rfintr just for scatter/gather header, and     *
*				various rfioctl support routines	       *
*									       *
*	Calling Parameters:	size - the size of DVMA needed                 *
*				ptr - pointer to a DVMA chunk to release       *
*				                                               *
*	Calls Subroutines:	safealloc: rmalloc                    	       *
*				safefree: rmfree			       *
*				                                               *
*	Public/Global Variables:None					       *
*				                                               *
*	Description:							       *
*		Called to allocate and return iopbmap resources to insulate us *
*		from unaligned iopbmap allocations.  To just protect ourselves,*
*		we could ask for (size + 3) bytes, but that might cause	       *
*		problems for other drivers, so we will ask for (size + 4)      *
*		bytes, thereby leaving iopbmap aligned or unaligned as we      *
*		found it.						       *
*		                                                               *
*******************************************************************************/

struct sapool {
	caddr_t start;
	long   len;
	};

static caddr_t 
safealloc(size)
int size;
{
    REGISTER char *chunk;
    REGISTER struct sapool *alchunk;
    register chsize;

    chsize = size + sizeof (struct sapool) + sizeof (int);
    if ((chunk = (caddr_t)rmalloc(iopbmap,chsize)) == NULL)
		return(chunk);
    alchunk = (struct sapool *)((int)(chunk+sizeof (int)) & ~(sizeof (int)-1));
    alchunk->start = chunk;
    alchunk->len = chsize;
    return ((char *)(alchunk + 1));
    } /* safealloc */

static void
safefree(chunk)
caddr_t chunk;
{
    REGISTER struct sapool *alchunk;

    alchunk = ((struct sapool *)chunk) - 1;
    rmfree (iopbmap, alchunk->len, alchunk->start);
    }

#endif NRF
